# Class for reading and visualizing the contents of a distance matrix in PHYLIP format.
#
# Usage:
# import dataHelper
# locData = dataHelper.getLocationSetData()
#
#  import DistMatrix
#  dm = DistMatrix.DistMatrix('myDistMatrix.phylip', locData)
#  dm.SetActiveTriangle('upper')
#  dm.DrawLines(['GBR'],['*'])

import GenGisAPIv1 as ggAPI
import dataHelper as dh

class DistMatrix:
	def __init__(self, filename, locData):
		"""
		Class for reading and visualizing the contents of a distance matrix in PHYLIP format. 
				
		@param filename: name of file containing distance matrix
		@param locData: location data for sites specified in distance matrix		
		"""
		
		# read in distance matrix
		fin = open(filename)
		matrix = fin.readlines()
		fin.close()
		
		# parse distance matrix
		self.index = {}
		self.matrix = []
		self.minDist = 1e10
		self.maxDist = -1e10
		for i in xrange(1, len(matrix)):
			data = matrix[i].split()
			self.index[data[0]] = i-1
			
			row = []
			for j in xrange(1,len(data)):
				dist = float(data[j])
				if dist < self.minDist:
					self.minDist = dist
					
				if dist > self.maxDist:
					self.maxDist = dist
				row.append(dist)
				
			self.matrix.append(row)

		# get dictionary indicating the geographic coordinates of each location site
		locData = dh.getLocationSetData()
		self.locDict = {}
		for loc in locData:
			self.locDict[loc.id] = [loc.northing, loc.easting]
			
		self.minSize = 1
		self.maxSize = 6
		
		self.startColour = ggAPI.Colour(1.0, 0.0, 0.0, 1.0)
		self.endColour = ggAPI.Colour(1.0, 1.0, 1.0, 1.0)
			
	def SetLineSizes(self, minSize, maxSize):
		"""
		Set min and max size of lines.
		"""
		self.minSize = minSize
		self.maxSize = maxSize
			
	def SetLineColours(self, startColour, endColour):
		"""
		Set start and end colour of lines.
		"""
		self.startColour = startColour
		self.endColour = endColour
			
	def __InterpolateSize(self, dist):
		"""
		Interpolate size of line.
		
		@param dist: distance value
		@return Size of line with a distance value of dist.
		"""
		offset = (dist - self.minDist) / (self.maxDist - self.minDist)
		return self.minSize + offset*(self.maxSize-self.minSize)
	
	def __InterpolateColour(self, dist):
		"""
		Interpolate size of line.
		
		@param dist: distance value
		@return Colour of line with a distance of dist.
		"""
		offset = (dist - self.minDist) /  (self.maxDist - self.minDist)
		return ggAPI.Colour(self.startColour.GetRed() + offset*(self.endColour.GetRed()-self.startColour.GetRed()), 
													self.startColour.GetGreen() + offset*(self.endColour.GetGreen() - self.startColour.GetGreen()), 
													self.startColour.GetBlue() + offset*(self.endColour.GetBlue() - self.startColour.GetBlue()), 
													self.startColour.GetAlpha() + offset*(self.endColour.GetAlpha() - self.startColour.GetAlpha()))

	def DrawLines(self, lines):
		"""
		Draw lines from between locations specified in distance matrix. For example,
		lines=[['A', 'B']] will draw a line from A to B. The distance for this line 
		will be taken from row A and column B. Multiple lines	can be draw using 
		lines = [['A','B'],['C','D']]. 
		
		@param lines: specifies lines to draw as a list of lists
		"""
					
		# add graph lines to GenGIS Viewport
		for line in lines:
				r = self.index[line[0]]
				c = self.index[line[1]]
				
				dist = self.matrix[r][c]
		
				startPt = ggAPI.Point3D()
				ggAPI.convertGeoCoord(self.locDict[line[0]][0], self.locDict[line[0]][1], startPt)
			
				endPt = ggAPI.Point3D()
				ggAPI.convertGeoCoord(self.locDict[line[1]][0], self.locDict[line[1]][1], endPt)
			
				size = self.__InterpolateSize(dist)
				colour = self.__InterpolateColour(dist)
			
				line = ggAPI.VisualLine(colour, size, ggAPI.LINE_STYLE.SOLID, ggAPI.Line3D(startPt, endPt))
				ggAPI.addLine(line)
			
				ggAPI.refresh()
				
	def DrawTopEntries(self, locationId, num, bMax, bRow):
		'''
		Draw lines indicating top \'num\' minimum or maximum entries for a given location.
		
		@param locationId: location of interest
		@param num: number of lines to draw
		@param bMax: indicates if lines should be drawn for the maximum (True) or minimum (False) entries
		@param bRow: indicates if distance are read along a row (True) or column (False)
		'''
		
		# get import or export distance from given location to all other locations
		locIndex = self.index[locationId]
		dist = {}
		if bRow:
			for key in self.index:
				dist[key] = self.matrix[locIndex][self.index[key]]
		else:
			for key in self.index:
				dist[key] = self.matrix[self.index[key]][locIndex]
				
		# sort dictionary by value
		items = [(v, k) for k, v in dist.items()]
		items.sort()
		
		if bMax:
			items.reverse()
		items = [(k, v) for v, k in items]
    
    # draw top 'num' lines
		for i in xrange(0,num):
			curLocId = items[i][0]
			dist = items[i][1]
			print dist
						
			startPt = ggAPI.Point3D()
			ggAPI.convertGeoCoord(self.locDict[locationId][0], self.locDict[locationId][1], startPt)
			
			endPt = ggAPI.Point3D()
			ggAPI.convertGeoCoord(self.locDict[curLocId][0], self.locDict[curLocId][1], endPt)
			
			size = self.__InterpolateSize(dist)
			colour = self.__InterpolateColour(dist)
			
			line = ggAPI.VisualLine(colour, size, ggAPI.LINE_STYLE.SOLID, ggAPI.Line3D(startPt, endPt))
			ggAPI.addLine(line)
			
			ggAPI.refresh()

		
					