# TYPE INTO PYTHON CONSOLE FIRST:
#
# import dataHelper
# 
# NOW USE FUNCTION AS (for example):
#
# locData = dataHelper.getLocationSetData()
# dataHelper.printData(locData, 'Temperature')
# filteredData = dataHelper.filterData(locData, 'Temperature', 20, dataHelper.filterFunc.greater)
# dataHelper.printData(filteredData, 'Temperature')

import GenGisAPIv1 as ggAPI
import filterFunc

def getChartSetView(locationSetIndex = 0, mapIndex = 0):
	"""
	Get specified chart set.

	@param locationSetIndex: index of location set of interest
	@param mapIndex: index of map of interest
	@return: chart set object which controls all charts for a given location set
	"""
	map = ggAPI.study.GetMap(mapIndex)
	locationSet = map.GetLocationSet(locationSetIndex) 
	return locationSet.GetChartSet()
	
def getLocationSetData(locationSetIndex = 0, mapIndex = 0):
	"""
	Get specified location set data.

	@param locationSetIndex: index of location set of interest
	@param mapIndex: index of map of interest
	@return: location set data for indicated location set as a python list
	"""
	map = ggAPI.study.GetMap(mapIndex)
	locationSet = map.GetLocationSet(locationSetIndex) 
	return createList(locationSet.locations)
	
def getSequenceData(locationIndex = 0, locationSetIndex = 0, mapIndex = 0):
	"""
	Get specified sequence data.

	@param locationIndex: index of location of interest
	@param locationSetIndex: index of location set of interest
	@param mapIndex: index of map of interest
	@return: sequence data from indicated location as a python list
	"""
	map = ggAPI.study.GetMap(mapIndex)
	locationSet = map.GetLocationSet(locationSetIndex) 
	location = locationSet.GetLocation(locationIndex)
	return createList(location.sequences)
		
def createList(dataVec):
	"""
	Simple helper function that converts a C++ std::vector into a python list.

	@return: python list which is equivalent in content to the provided C++ std::vector
	"""
	dataAsList = []
	for item in dataVec:
		dataAsList.append(item)
		
	return dataAsList
	
def createVec(dataList):
	"""
	Simple helper function that converts a Python list into a  C++ std::vector.

	@return: C++ std::vector which is equivalent in content to the provided Python list
	"""
	dataVec = ggAPI.wstringVector()
	for item in dataList:
		dataVec.append(item)
		
	return dataVec
	
def printData(dataList, field):
	"""
	Print value of specified field for all elements in a list.

	@param dataList: list with location set or sequence data objects
	@param field: data field of interest
	"""
	for item in dataList:
		print item.data[field]
		
def filterData(dataList, field, filterValue, filterFunc):
	"""
	Filter a list of data objects.
	
	Example: dataHelper.filterData(myData, 'Temperature', 10, dataHelper.filterFunc.greater)
	will return only data objects whose Temperature field is strictly greater than 10.

	@param dataList: list with location set or sequence data objects
	@param field: data field to filter on
	@param filterValue: value to filter on
	@param filterFunc: function to filter on (see filterFunc.py)
	@return: List containing only data objects that pass the filter
	"""
	filteredList = []
	for item in dataList:
		if filterFunc(item.data[field], filterValue):
			filteredList.append(item)
			
	return filteredList

def selectData(sequences, field, filterValue, filterFunc):
	"""
	Select sequences meeting the specified criteria.
	
	Example: dataHelper.selectData('Temperature', 10, dataHelper.filterFunc.greater)
	will select sequences whose Temperature field is strictly greater than 10.

	@param sequences: list of sequence data objects
	@param field: data field to filter on
	@param filterValue: value to filter on
	@param filterFunc: function to filter on (see filterFunc.py)
	"""
	for seq in sequences:
		if filterFunc(seq.data[field], filterValue):
			seq.SetActive(True)
		else:
			seq.SetActive(False)
	
	getChartSetView().UpdateCharts(field)
	ggAPI.refresh()
	
def selectAllSeqs(sequences):
	"""
	Select all sequences.
	
	@param sequences: list of sequences to select
	"""
	for seq in sequences:
		seq.SetActive(True)
		