//=======================================================================
// Author: Mike Porter
//
// Copyright 2009 Mike Porter
//
// This file is part of GenGIS.
//
// GenGIS is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// GenGIS is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with GenGIS.  If not, see <http://www.gnu.org/licenses/>.
//=======================================================================
#include "../core/Precompiled.hpp"

#include "../core/API.hpp"

#include "../core/App.hpp"
#include "../core/Viewport.hpp"
#include "../core/TerrainController.hpp"
#include "../core/TerrainView.hpp"
#include "../core/LayerTreeController.hpp"
#include "../core/SequenceController.hpp"
#include "../core/LocationLayer.hpp"
#include "../core/LocationSetLayer.hpp"
#include "../core/MapLayer.hpp"
#include "../core/SequenceLayer.hpp"
#include "../core/StudyLayer.hpp"
#include "../core/Camera.hpp"
#include "../core/UserElements.hpp"
#include "../core/LabelController.hpp"

#include "../widgets/WidgetController.hpp"
#include "../gui/GenGisFrame.hpp"
#include "../utils/Log.hpp"
#include "../utils/StringTools.hpp"

using namespace GenGIS;

// ~mikep. Is this "extern "C"" even needed anymore?
extern "C"    // Don't mangle the method names in the API... plugins need them un-mangled.
{
  void APIv1::write(std::string text) { error::Log::Inst().Write(text); }
  
  void APIv1::warning(std::string text) { error::Log::Inst().Warning(text); }
  
  void APIv1::error(std::string text) { error::Log::Inst().Error(text); }
  
  void APIv1::commands(bool detail)
  {
    // if(detail)
    // {
    //   PyRun_SimpleString("print 'API COMMANDS:\\n'");
    // }
    // else
    // {
    //   PyRun_SimpleString("print 'API COMMANDS:\\n'");
    // }
    return;
  }
  
  //==============================
  //   GenGIS control functions
  //==============================
  
  void APIv1::quit() { ((GenGisFrame*)GenGIS::App::Inst().GetMainWindow())->Close(TRUE); }
  
  void APIv1::openMap(std::string file) { ((GenGisFrame*)GenGIS::App::Inst().GetMainWindow())->LayerOpenMap(file); }
  
  void APIv1::openLocations(std::string file) { ((GenGisFrame*)GenGIS::App::Inst().GetMainWindow())->LayerOpenLocations(file); }

	void APIv1::openTree(std::string file) { ((GenGisFrame*)GenGIS::App::Inst().GetMainWindow())->LayerOpenTree(file); }
  
  void APIv1::openSequenceData(std::string file) { ((GenGisFrame*)GenGIS::App::Inst().GetMainWindow())->LayerOpenSequenceData(file); }
  
  void APIv1::saveImage(std::string path) { App::Inst().GetViewport()->SaveImage(path); }
  
  // void closeFile(std::string DEM=FALSE, std::string samples=FALSE, std::string genomeData=FALSE)
  // {
  //   "Closes the (already opened) DEM file, samples file, and/or genome file.";
  // }
  
  //=================================
  //   GenGIS appearance functions
  //=================================
  
  /*** General functions ***/
	
  // Yields control to pending messages in the windowing system. This can be useful, for example, 
  // when a time-consuming process writes to a text window. Without an occasional yield, the text 
  // window will not be updated properly
  void APIv1::safeYield() { ((GenGisFrame*)GenGIS::App::Inst().GetMainWindow())->SafeYield(); }

	int APIv1::viewportWidth() { return GenGIS::App::Inst().GetViewport()->GetWidth(); }

	int APIv1::viewportHeight() { return GenGIS::App::Inst().GetViewport()->GetHeight(); }

  /***   Side Panel   ***/
  
  bool APIv1::sidePanelVisible() { return ((GenGisFrame*)GenGIS::App::Inst().GetMainWindow())->m_splitterVertical->IsSplit(); }
  
  // Toggles side panel on/off.
  void APIv1::toggleSidePanel() { ((GenGisFrame*)GenGIS::App::Inst().GetMainWindow())->ViewSidePanel(); }
  
  // If `display` is TRUE, show the side panel. Otherwise hide it.
  void APIv1::displaySidePanel(bool display)
  {
    if(display && !sidePanelVisible()) { ((GenGisFrame*)GenGIS::App::Inst().GetMainWindow())->ViewSidePanel(); }
    else if(!display && sidePanelVisible()) { ((GenGisFrame*)GenGIS::App::Inst().GetMainWindow())->ViewSidePanel(); }
  }
  
  /***   Console   ***/
  
  // Returns TRUE if console is visible, FALSE otherwise.
  bool APIv1::consoleVisible() { return ((GenGisFrame*)GenGIS::App::Inst().GetMainWindow())->m_splitterHorizontal->IsSplit(); }
  
  // Toggles console on/off.
  void APIv1::toggleConsole() { ((GenGisFrame*)GenGIS::App::Inst().GetMainWindow())->ViewConsole(); }
  
  // If `display` is TRUE, show the console. Otherwise hide it.
  void APIv1::displayConsole(bool display)
  {
    if(display && !consoleVisible()) { ((GenGisFrame*)GenGIS::App::Inst().GetMainWindow())->ViewConsole(); }
    else if(!display && consoleVisible()) { ((GenGisFrame*)GenGIS::App::Inst().GetMainWindow())->ViewConsole(); }
  }

  //==========================
  //   Navigation functions
  //==========================
  
  // Pans the map up.
  void APIv1::up(float number)
  {
    App::Inst().GetViewport()->GetCamera()->TranslateDelta(0, (float)(number)*-0.003f);
  }
  
  // Pans the map down.
  void APIv1::down(float number)
  {
    App::Inst().GetViewport()->GetCamera()->TranslateDelta(0, (float)(number)*0.003f);
  }
  
  // Pans the map left.
  void APIv1::left(float number)
  {
    App::Inst().GetViewport()->GetCamera()->TranslateDelta((float)(number)*-0.003f, 0);
  }
  
  // Pans the map right.
  void APIv1::right(float number)
  {
    App::Inst().GetViewport()->GetCamera()->TranslateDelta((float)(number)*0.003f, 0);
  }
  
  // Rotates the map.
  void APIv1::rotate(float number)
  {
    App::Inst().GetViewport()->GetCamera()->YawDelta((float)number);
    ((GenGisFrame*)GenGIS::App::Inst().GetMainWindow())->GetWidgetController()->SetCompassAngle(App::Inst().GetViewport()->GetCamera()->GetYaw());
  }

	// Get yaw rotation of camera.
	float APIv1::getYaw()
	{
		return App::Inst().GetViewport()->GetCamera()->GetYaw();
	}

	// Set yaw rotation of camera.
	void APIv1::setYaw(float number)
	{
		App::Inst().GetViewport()->GetCamera()->SetYaw(number);
		((GenGisFrame*)GenGIS::App::Inst().GetMainWindow())->GetWidgetController()->SetCompassAngle(number);
	}
  
  // Adjusts the map's pitch.
  void APIv1::pitch(float number)
  {
    App::Inst().GetViewport()->GetCamera()->PitchDelta(-(float)number);
  }

	// Get pitch of camera.
	float APIv1::getPitch()
	{
		return App::Inst().GetViewport()->GetCamera()->GetPitch();
	}

	// Set pitch of camera.
	void APIv1::setPitch(float number)
	{
		App::Inst().GetViewport()->GetCamera()->SetPitch(number);
	}
  
  // Zoomes the map.
  void APIv1::zoom(float number)
  {
    App::Inst().GetViewport()->GetCamera()->ZoomDelta((float)(number)*-0.025f);
  }

	// Get camera zoom.
	 float APIv1::getZoom()
  {
    return App::Inst().GetViewport()->GetCamera()->GetZoom();
  }

	 // Get position of camera.
	const Point3D& APIv1::getPosition()
	{
		return App::Inst().GetViewport()->GetCamera()->GetPosition();
	}

	// Set position of camera.
	void APIv1::setPosition(float x, float y, float z)
	{
		App::Inst().GetViewport()->GetCamera()->SetPosition(x, y, z);
	}
  
  //==========================
  //   3D Terrain functions
  //==========================
  
  // Sets the 3D terrain to a top-down view.
  void APIv1::viewTop() { ((GenGisFrame*)GenGIS::App::Inst().GetMainWindow())->ViewTop(); }
  
  // Reset terrain view.
  void APIv1::viewReset() { ((GenGisFrame*)GenGIS::App::Inst().GetMainWindow())->ViewReset(); }
  
  /***   Wireframe   ***/
  
  bool APIv1::isWireframe() { return ((GenGisFrame*)GenGIS::App::Inst().GetMainWindow())->m_terrainController->GetTerrainView()->IsWireFrame(); }
  
  void APIv1::displayWireframe(bool display)
  {
    if(display) { ((GenGisFrame*)GenGIS::App::Inst().GetMainWindow())->m_terrainController->GetTerrainView()->SetWireFrame(TRUE);  }
    else        { ((GenGisFrame*)GenGIS::App::Inst().GetMainWindow())->m_terrainController->GetTerrainView()->SetWireFrame(FALSE); }
    refresh();
  }
  
  void APIv1::toggleWireframe() { displayWireframe(!isWireframe()); }
  
  // Adjust the detail of the terrain.
  void APIv1::detail(int level)
  {
    if(level < 0)      { for(int i = 0; i < level; i++) { ((GenGisFrame*)GenGIS::App::Inst().GetMainWindow())->ViewCoarse(); } }
    else if(level > 0) { for(int i = 0; i < level; i++) { ((GenGisFrame*)GenGIS::App::Inst().GetMainWindow())->ViewFine();   } }
  }
  
  void APIv1::refresh() 
  { 
    App::Inst().GetViewport()->Render();
  }

	//==========================
  //   User graphical element functions
  //==========================
	uint APIv1::addLine(VisualLinePtr line)
	{
		return UserElements::Inst().AddLine(line);
	}

	uint APIv1::removeLine(uint lineId)
	{
		return UserElements::Inst().RemoveLine(lineId);
	}

	void APIv1::clearLines()
	{
		UserElements::Inst().ClearLines();
	}

	uint APIv1::addMarker(VisualMarkerPtr marker)
	{
		return UserElements::Inst().AddMarker(marker);
	}

	uint APIv1::removeMarker(uint markerId)
	{
		return UserElements::Inst().RemoveMarker(markerId);
	}

	void APIv1::clearMarkers()
	{
		UserElements::Inst().ClearMarkers();
	}

	uint APIv1::addLabel(VisualLabelPtr label)
	{
		return LabelController::Inst().AddLabel(label);
	}

	uint APIv1::removeLabel(uint labelId)
	{
		return LabelController::Inst().RemoveLabel(labelId);
	}

	void APIv1::clearLabels()
	{
		LabelController::Inst().ClearLabels();
	}

	void APIv1::convertGeoCoord(float northing, float easting, Point3D& objectCoord)
	{
		GeoCoord geoCoord(easting, northing);
		((GenGisFrame*)GenGIS::App::Inst().GetMainWindow())->GetTerrainController()->GeoToGrid(geoCoord, objectCoord);
	}

	void APIv1::convertLatLong(float latitude, float longitude, Point3D& objectCoord)
	{
		GeoCoord geoCoord(longitude, latitude);
		((GenGisFrame*)GenGIS::App::Inst().GetMainWindow())->GetTerrainController()->LatLongToGrid(geoCoord, objectCoord);
	}
}