//=======================================================================
// Author: Mike Porter
//
// Copyright 2009 Mike Porter
//
// This file is part of GenGIS.
//
// GenGIS is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// GenGIS is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with GenGIS.  If not, see <http://www.gnu.org/licenses/>.
//=======================================================================

#ifndef _GENGIS_API_
#define _GENGIS_API_

#include <string>
#include <stdio.h>

namespace GenGIS
{
  extern "C"    // Don't mangle the method names in the API... plugins need them un-mangled.
  {
    struct APIv1
    {
      static APIv1& Inst()
    	{
    		static APIv1 apiv1;
    		return apiv1;
    	}
      
      APIv1(){}
      
      // API methods are declared here:
      
      static void write(std::string text);
      
      static void warning(std::string text);
      
      static void error(std::string text);
      
      static void commands(bool detail=false);
      
      //==============================
      //   GenGIS control functions
      //==============================
      
      static void quit();
      
      static void openMap(std::string file);
      
      static void openLocations(std::string file);

			static void openTree(std::string file);
      
      static void openSequenceData(std::string file);
      
      static void saveImage(std::string path);
      
      // static void closeFile(std::string DEM=FALSE, std::string samples=FALSE, std::string genomeData=FALSE)
      // {
      //   "Closes the (already opened) DEM file, samples file, and/or genome file.";
      // }
      
      //=================================
      //   GenGIS appearance functions
      //=================================

      /*** General functions ***/
			
      // Yields control to pending messages in the windowing system. This can be useful, for example, 
      // when a time-consuming process writes to a text window. Without an occasional yield, the text 
      // window will not be updated properly
      static void safeYield();
			
			// Return width of Viewport.
			static int viewportWidth();

			// Return height of Viewport.
			static int viewportHeight();
      
      /***   Side Panel   ***/
      
      static bool sidePanelVisible();
      
      // Toggles side panel on/off.
      static void toggleSidePanel();
      
      // If `display` is TRUE, show the side panel. Otherwise hide it.
      static void displaySidePanel(bool display);
      
      /***   Console   ***/
      
      // Returns TRUE if console is visible, FALSE otherwise.
      static bool consoleVisible();
      
      // Toggles console on/off.
      static void toggleConsole();
      
      // If `display` is TRUE, show the console. Otherwise hide it.
      static void displayConsole(bool display);

      //==========================
      //   Navigation functions
      //==========================
      
      // Pans the map up.
      static void up(float number=1.0f);
      
      // Pans the map down.
      static void down(float number=1.0f);
      
      // Pans the map left.
      static void left(float number=1.0f);
      
      // Pans the map right.
      static void right(float number=1.0f);
      
      // Rotates the map.
      static void rotate(float number=1.0f);

			// Get yaw rotation of camera.
			static float getYaw();

			// Set yaw rotation of camera.
			static void setYaw(float number);
      
      // Adjusts the map's pitch.
      static void pitch(float number=1.0f);

			// Get pitch of camera.
			static float getPitch();

			// Set pitch of camera.
			static void setPitch(float number);
      
      // Zoomes the map.
      static void zoom(float number=1.0f);

			// Get camera zoom.
      static float getZoom();

			// Get position of camera.
			static const Point3D& getPosition();

			// Set position of camera.
			static void setPosition(float x, float y, float z);
      
      //==========================
      //   3D Terrain functions
      //==========================
      
      // Sets the 3D terrain to a top-down view.
      static void viewTop();
      
      // Reset terrain view.
      static void viewReset();
      
      /***   Wireframe   ***/
      
      static bool isWireframe();
      
      static void displayWireframe(bool display);
      
      static void toggleWireframe();
      
      // Adjust the detail of the terrain.
      static void detail(int level=1);
      
      static void refresh();

			//==========================
      //   User Graphical Elements
      //==========================
			static uint addLine(VisualLinePtr line);
			static uint removeLine(uint lineId);
			static void clearLines();
			static uint addMarker(VisualMarkerPtr line);
			static uint removeMarker(uint markerId);
			static void clearMarkers();
			static uint addLabel(VisualLabelPtr label);
			static uint removeLabel(uint labelId);
			static void clearLabels();

			static void convertGeoCoord(float easting, float northing, Point3D& objectCoord);
			static void convertLatLong(float easting, float northing, Point3D& objectCoord);
    };
  }
}

#endif