//=======================================================================
// Author: Donovan Parks
//
// Copyright 2009 Donovan Parks
//
// This file is part of GenGIS.
//
// GenGIS is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// GenGIS is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with GenGIS.  If not, see <http://www.gnu.org/licenses/>.
//=======================================================================

#ifndef _GENGIS_APP_
#define _GENGIS_APP_

#include "../core/Precompiled.hpp"
#include "../core/LayoutObjectProperties.hpp"
#include "../core/View.hpp"
#include "../core/LayerTreeController.hpp"
#include "../core/StudyLayer.hpp"
#include "../core/StudyModel.hpp"

#include "../utils/Font.hpp"

#include <wx/dir.h>

using namespace boost::filesystem;

namespace GenGIS
{

/**
 * @class App
 * @brief Singleton class for holding project specific settings and widely access system level objects.
 */
class App
{
public:
	/** Program operating modes. */
	enum MODE { STANDARD, LAYOUT_LINE, LAYOUT_ELLIPSE, GEO_AXIS_POLYLINE, USER_LABEL };

public:
	/** Destructor */
	~App()
	{

	}

	/** Get the one and only instance of this class. */
	static App& Inst()
	{
		static App app;
		return app;
	}

	/** 
	 * @brief Set main application window. 
	 * @param window Pointer to main application window.
	 */
	void SetMainWindow(wxFrame* window) { m_mainWindow = window; }

	/** Get main application window. */
	wxFrame* GetMainWindow() { return m_mainWindow; }

	/** Get terrain controller. Right now we are assuming there is only one terrain loaded. */
	TerrainControllerPtr GetTerrainController();

	/** 
	 * @brief Set OpenGL viewport.
	 * @param viewport Smart pointer to OpenGL viewport
	 */
	void SetViewport(ViewportPtr viewport) { m_viewport = viewport; }

	/** Get OpenGL viewport. */
	ViewportPtr GetViewport() { return m_viewport; }

	/** Set controller to layer tree which gives access to all loaded layers. */

	/** 
	 * @brief Set controller to layer tree which gives access to all loaded layers.
	 * @param controller Smart pointer to layer tree controller.
	 */
	void SetLayerTreeController(LayerTreeControllerPtr controller) { m_layerTreeController = controller; }

	/** Get controller to layer tree which gives access to all loaded layers. */
	LayerTreeControllerPtr GetLayerTreeController() { return m_layerTreeController; }

	/** Get study controller. */
	StudyControllerPtr GetStudyController() { return m_layerTreeController->GetStudyLayer(0)->GetStudyController(); }

	/** 
	 * @brief Set the current program mode.
	 * @param mode Desired program mode.
	 */
	void SetMode(MODE mode) { m_mode = mode; }

	/** Get the current program mode. */
	MODE GetMode() { return m_mode; }

	/** 
	 * @brief Set the 3D position of the mouse in world (OpenGL) coordinates.
	 * @param pt Position of mouse in world coordinates.
	 */
	void SetMouseWorldPos(const Point3D& pt) { m_mouseWorldPos = pt; }

	/** Get the 3D position of the mouse in world (OpenGL) coordinates. */
	Point3D GetMouseWorldPos() { return m_mouseWorldPos; }

	/** Get the 2D position of the mouse in the plane of the terrain (y=0). */
	Point3D GetMouseTerrainPlanePos() { return Point3D(m_mouseWorldPos.x, 0, m_mouseWorldPos.z); }

	/** 
	 * @brief Set the 2D position of the mouse in screen coordinates.
	 * @param pt Position of mouse in screen coordinates.
	 */
	void SetMouseScreenPos(const Point2D& pt) { m_mouseScreenPos = pt; }

	/** Get the 2D position of the mouse in screen coordinates. */
	Point2D GetMouseScreenPos() { return m_mouseScreenPos; }

	/** Set flag indicating if the mouse is being dragged. */
	void SetMouseDragging(bool state) { m_mouseDragging = state; }

	/** Get flag indicating if the mouse is being dragged. */
	bool GetMouseDragging() { return m_mouseDragging; }

	/** Get layout object properties. */
	LayoutObjectProperties::Ptr GetLayoutObjProp() { return m_layoutObjProp; }

	/** Set currently selected view. */

	/** 
	 * @brief Set currently selected view.
	 * @param view Smart pointer to currently selected view.
	 */
	void SetSelectedView(ViewPtr view) 
	{ 
		// unselect current view
		if(m_selectedView)
			m_selectedView->SetSelected(false);

		// set new selected view
		m_selectedView = view; 
		if(m_selectedView)
			m_selectedView->SetSelected(true);
	}

	/** Get currently selected view. */
	ViewPtr GetSelectedView() { return m_selectedView; }

	/** 
	 * @brief Set currently active right-click view.
	 * @param view Smart pointer to currently active right-click view.
	 */
	void SetRightClickView(ViewPtr view) { m_rightClickView = view; }

	/** Get currently active right-click view. */
	ViewPtr GetRightClickView() { return m_rightClickView; }

	/** Get selection colour. */
	Colour GetSelectionColour() { return Colour(255, 0, 0); }

	/** Get selection thickness. */
	float GetSelectionThickness() { return 2; }

	/** Get directory containing program executable. */
	wxString GetExeDir()
	{
		wxStandardPaths standardPaths;
    wxString exeDir;
		#ifdef WIN32
		  // Windows version: if you need to change these lines please contact Donovan first
			exeDir = standardPaths.GetExecutablePath();
		  exeDir = exeDir.SubString(0, exeDir.find_last_of('\\'));
    #else
      // Mac version
      exeDir = standardPaths.GetExecutablePath();
      exeDir = exeDir.SubString(0, exeDir.length()-13);
      exeDir = exeDir + _T("Resources/");
    #endif
		return exeDir;
	}

	/** Get directory containing user data for program. */
	wxString GetUserDataDir()
	{
		wxStandardPaths standardPaths;
		if(!wxDir::Exists(standardPaths.GetUserLocalDataDir()))
		{
			create_directory(standardPaths.GetUserLocalDataDir().c_str());
		}

		return standardPaths.GetUserLocalDataDir();
	}

	/** Set colour map manager. */
	void SetColourMapManager(ColourMapManagerPtr colourMapManager) { m_colourMapManager = colourMapManager; }

	/** Get colour map manager. */
	ColourMapManagerPtr GetColourMapManager() const { return m_colourMapManager; }

	/** Set shape map manager. */
	void SetShapeMapManager(ShapeMapManagerPtr shapeMapManager) { m_shapeMapManager = shapeMapManager; }

	/** Get shape map manager. */
	ShapeMapManagerPtr GetShapeMapManager() const { return m_shapeMapManager; }

	/** Initialize OpenGL font. */
	void InitFont(const std::wstring& font)
	{
		wxString fontFile = App::Inst().GetExeDir() + wxT("fonts/") + font.c_str();
		m_font.reset(new Font(std::string(fontFile.begin(), fontFile.end()), 12));
	}

	/** Get OpenGL font. */
	FontPtr GetFont() { return m_font; }

	/** Get database connection string. */
	std::wstring GetDatabaseConnectionString() { return m_dbConnectionString; }

	/** Set database connection string. */
	void SetDatabaseConnectionString(const std::wstring& str) { m_dbConnectionString = str; }

	/** Set resolution factor for saving high resolution images. */
	void SetResolutionFactor(uint factor) { m_resolutionFactor = factor; }

	/** Get resolution factor for saving high resolution images. */
	uint GetResolutionFactor() const { return m_resolutionFactor; }

	/* Set flag indicating if high resolution image is being saved. */
	void SetSavingHiResImage(bool state) { m_bRenderingHiResImage = state; }

	/* Get flag indicating if high resolution image is being saved. */
	bool IsSavingHiResImage() const { return m_bRenderingHiResImage; }

protected:
	/** Private singleton constructor (must instantiate this class through Inst()). */
	App(): m_layoutObjProp(new LayoutObjectProperties()), 
		m_dbConnectionString(_T("http://ratite.cs.dal.ca/mg_web/fetch_metagenome_samples.cgi?type=metagenome"))
	{

	}

	/** Private copy constructor (must instantiate this class through Inst()). */
  App(const App&);

	/** Private assignment constructor (must instantiate this class through Inst()). */
  App& operator= (const App&);


private:
	/** Pointer to main application window. */
	wxFrame* m_mainWindow;

	/** Pointer to OpenGL viewport. */
	ViewportPtr m_viewport;

	/** Pointer to layer tree controller which holds all objects loaded into program. */
	LayerTreeControllerPtr m_layerTreeController;

	/** Current program mode. */
	MODE m_mode;

	/** The 3D position of the mouse in world (OpenGL) coordinates. */
	Point3D m_mouseWorldPos;

	/** The 2D position of the mouse in screen coordinates. */
	Point2D m_mouseScreenPos;

	/** Flag indicating if the mouse is currently being dragged. */
	bool m_mouseDragging;

	/** Visual properties for layout objects (e.g., line, ellipse). */
	LayoutObjectProperties::Ptr m_layoutObjProp;

	/** Currently selected view in OpenGL viewport. */
	ViewPtr m_selectedView;

	/** Currently active right-click view. */
	ViewPtr m_rightClickView;

	/** Colour map manager. */
	ColourMapManagerPtr m_colourMapManager;

	/** Shape map manager. */
	ShapeMapManagerPtr m_shapeMapManager;

	/** Font class used to render TrueType fonts on an OpenGL canvas. */
	FontPtr m_font;

	/** Database connection string. */
	std::wstring m_dbConnectionString;

	/** Sizing factor for saving high resolution image. */
	uint m_resolutionFactor;

	/** Flag indicating high resolution image is being rendered. */
	bool m_bRenderingHiResImage;
};

}

#endif
