//=======================================================================
// Author: Donovan Parks
//
// Copyright 2009 Donovan Parks
//
// This file is part of GenGIS.
//
// GenGIS is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// GenGIS is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with GenGIS.  If not, see <http://www.gnu.org/licenses/>.
//=======================================================================

#ifndef _GENGIS_GEO_TREE_CONTROLLER_
#define _GENGIS_GEO_TREE_CONTROLLER_

#include "../core/Precompiled.hpp"

#include "../utils/Tree.hpp"
#include "../core/NodeGeoTree.hpp"
#include "../core/GeoTreeView.hpp"

namespace GenGIS 
{

/**
 * @brief Controller for geographic tree model. 
 */
class GeoTreeController
{
public:
	/** 
	 * @brief Constuctor.
	 * @param treeView Tree view to associate with this controller.
	 */
	GeoTreeController(GeoTreeViewPtr treeView) : m_treeView(treeView) {}

	/** Destructor. */
	virtual ~GeoTreeController() {}

	/** Render tree model. */
	void Render();

	/** Get tree view associated with this controller. */
	GeoTreeViewPtr GetTreeView();

	/** Get number of nodes in tree. */
	uint GetNumberOfNodes();

	/** Get number of leaf nodes in tree. */
	uint GetNumberOfLeaves();

	/** Get depth of tree. */
	uint GetDepthOfTree();

	/** Indicate if tree is multifurcating. */
	bool IsMultifurcating();

	/** 
	 * @brief Project tree onto a set of leaf nodes.
	 * @param names Names of leaf nodes to project tree onto.
	 */
	void ProjectTree(std::vector<std::wstring>& names) { m_treeView->ProjectTree(names); }

	/** 
	 * @brief Restore original tree.
	 */
	void RestoreTree() { m_treeView->RestoreTree(); }

	/** 
	 * @brief Get names of all leaf nodes in tree.
	 */
	std::vector<std::wstring> GetLeafNames() { return m_treeView->GetTree()->GetLeafNames(); }

protected:
	/** Tree view associated with this controller. */
	GeoTreeViewPtr  m_treeView;
};

}

#endif
