//=======================================================================
// Author: Donovan Parks
//
// Copyright 2009 Donovan Parks
//
// This file is part of GenGIS.
//
// GenGIS is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// GenGIS is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with GenGIS.  If not, see <http://www.gnu.org/licenses/>.
//=======================================================================

#ifndef _GENGIS_GEO_TREE_ELLIPTICAL_LAYOUT
#define _GENGIS_GEO_TREE_ELLIPTICAL_LAYOUT

#include "../core/Precompiled.hpp"

#include "../core/NodeGeoTree.hpp"
#include "../utils/Tree.hpp"

namespace GenGIS 
{

class GeoTreeEllipticalLayout
{
public:
	/** Node and associated graphical elements for an elliptical layout. */
	typedef struct sELLIPTICAL_NODE
	{
		typedef boost::shared_array<sELLIPTICAL_NODE> ArrayPtr;
		typedef boost::shared_ptr<sELLIPTICAL_NODE> Ptr;

		/** Constructor. */
		sELLIPTICAL_NODE(NodeGeoTree* _node, Line3D _line = Line3D(), PolyLine3D _arc = PolyLine3D()): node(_node), line(_line), arc(_arc) {}

		/** Node in elliptical layout. */
		NodeGeoTree* node;

		/** Arcs connecting node to its children. */
		PolyLine3D arc;

		/** Line connecting node to its parent arc.*/
		Line3D line;
	} EllipticalNode;

public:
	/** 
	 * @brief Constructor. 
	 * @param terrainController Controller for terrain tree is to be laid out on.
	 */
	GeoTreeEllipticalLayout(TerrainControllerPtr terrainController): m_terrainController(terrainController) {}

	/** Destructor. */
	~GeoTreeEllipticalLayout() { }

	/** Layout tree around the layout ellipse with points positioned independently in each quadrant. */
	void LayoutEllipseQuadrant(Tree<NodeGeoTree>::Ptr tree, LayoutEllipsePtr layoutEllipse, float height);

	/** Get elliptical nodes that specify how to render this layout. */
	const std::vector<EllipticalNode>& GetEllipticalNodes() const { return m_ellipticalNodes; }

protected:
	/**
	 * @brief Calculate angle to point between 0 and 2PI in standard ellipse coordinate frame.			
	 * @param pt Point on ellipse.
	 * @return Angle to point from x-axis in standard ellipse coordinate frame.
	 */
	double GetAngleOnStandardEllipse(LayoutEllipsePtr layoutEllipse,const Point3D& pt);

	/**
	 * @brief Calculate position of node in elliptical layout.
	 * @param angle Angle of node from x-axis in CCW direction (in radians).
	 * @param childOffset Distance of child node from layout ellipse. 
	 * @param parentOffset Distance of parent node from layout ellipse.
	 * @return Position of child node on the arc which connects the node to its parent.
	 */
	Point3D GetEllipseLayoutPt(LayoutEllipsePtr layoutEllipse, double angle, double childOffset, double parentOffset);

protected:
	/** Controller for terrain tree is to be laid out on */
	TerrainControllerPtr m_terrainController;

	/** Nodes and associated graphical elements for an elliptical layout. */
	std::vector<EllipticalNode> m_ellipticalNodes;
};

}

#endif
