//=======================================================================
// Author: Donovan Parks
//
// Copyright 2009 Donovan Parks
//
// This file is part of GenGIS.
//
// GenGIS is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// GenGIS is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with GenGIS.  If not, see <http://www.gnu.org/licenses/>.
//=======================================================================

#ifndef _GENGIS_LAYER_
#define _GENGIS_LAYER_

#include "../core/Precompiled.hpp"

namespace GenGIS
{

/**
 * @brief Abstract base class representing a generic GIS layer.
 */
class Layer 
{
public:
	/** All supported layer types. */
	enum LAYER_TYPE { STUDY, MAP, LOCATION_SET, TREE, LOCATION, SEQUENCE };

	/**
	 * @brief Get string which indicates the layer type.
	 * @param type Flag indicating layer type.
	 * @return String indicating layer type.
	 */
	static std::wstring GetLayerTypeString(LAYER_TYPE type)
	{
		if(type == STUDY) return _T("Study");
		else if(type == MAP) return _T("Map");
		else if(type == LOCATION_SET) return _T("Location Set");
		else if(type == TREE) return _T("Tree");
		else if(type == LOCATION) return _T("Location");
		else if(type == SEQUENCE) return _T("Sequence");

		return _T("Unknown layer type! Problem in Layer::GetLayerTypeString()!");
	}

public:
	/** 
	 * @brief Constuctor.
	 * @param id Unique string to identify layer.
	 * @param desc Description of layer.
	 * @param authours Authours of layer.
	 * @param path Path to file associated with this layer.
	 * @param filename File associated with this layer.
	 * @param layer Type of layer.
	 * @param parent Parent of this layer.
	 */
	Layer(uint id, LAYER_TYPE type, LayerPtr parent)
		: m_id(id), m_type(type), m_parent(parent), m_visible(true), m_name(_T("")), 
			m_description(_T("")), m_authours(_T("")), m_path(_T("")), m_filename(_T("")) {};

	/** Destructor. */
	virtual ~Layer() {}

	/** Get id of layer. */
	virtual const uint& GetId() const { return m_id; }

	/** Set id of layer. */
	virtual const void SetId(uint id) { m_id = id; }

	/** Get name of layer. */
	virtual const std::wstring& GetName() const { return m_name; }

	/** Set name of layer. */
	virtual const void SetName(const std::wstring& name) { m_name = name; }

	/** Get description of layer. */
	virtual const std::wstring& GetDescription() const { return m_description; }

	/** Set description of layer. */
	virtual const void SetDescription(const std::wstring& description) { m_description = description; }

	/** Get authours of layer. */
	virtual const std::wstring& GetAuthours() const { return m_authours; }

	/** Set authours of layer. */
	virtual const void SetAuthours(const std::wstring& authours) { m_authours = authours; }

	/** Get path of layer. */
	virtual const std::wstring& GetPath() const { return m_path; }

	/** Set path of layer. */
	virtual const void SetPath(const std::wstring& path) { m_path = path; }

	/** Get filename of layer. */
	virtual const std::wstring& GetFilename() const { return m_filename; }

	/** Set filename of layer. */
	virtual const void SetFilename(const std::wstring& filename) { m_filename = filename; }

	/** Get type of layer. */
	virtual const LAYER_TYPE& GetType() const { return m_type; }

	/** 
	 * @brief Set visibility state of layer.
	 * @param state Desired visibility of layer.
	 */
	virtual void SetVisible(bool state) { m_visible = state; }

	/** Get visibility of layer. */
	virtual bool IsVisible() const { return m_visible; }

	/** Toggle visibility of layer. */
	virtual bool ToggleVisibility() { m_visible = !m_visible; return m_visible; }

	/**
	 * @brief Get layer from tree below this layer. 
	 * @param id Unique id of desired layer.
	 * @return layer at given id.
	 */
	virtual LayerPtr GetLayer(uint id) const = 0;

	/**
	 * @brief Remove layer from tree below this layer. 
	 * @param id Unique id of layer to remove.
	 * @return True if layer removed, false if layer not found.
	 */
	virtual bool RemoveLayer(uint id) = 0;

	/** Get parent layer. */
	LayerPtr GetParent() { return m_parent; }

	/** Render contents of layer. */
	virtual void Render() = 0;

protected:
	/** Unique number identifying this layer. */
	uint m_id;

	/** Name of layer. */
	std::wstring m_name;

	/** Description of layer. */
	std::wstring m_description;

	/** Authours of layer. */
	std::wstring m_authours;

	/** Path to file associated with this layer. */
	std::wstring m_path;

	/** File associated with this layer. */
	std::wstring m_filename;

	/** Pointer to parent of this layer. */
	LayerPtr m_parent;

	/** Type of layer (i.e., study, map, location set, ...). */
	LAYER_TYPE m_type;

	/** Indicate is layer is currently visible. */
	bool m_visible;
};

}

#endif
