//=======================================================================
// Author: Donovan Parks
//
// Copyright 2009 Donovan Parks
//
// This file is part of GenGIS.
//
// GenGIS is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// GenGIS is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with GenGIS.  If not, see <http://www.gnu.org/licenses/>.
//=======================================================================

#include "../core/Precompiled.hpp"

#include "../core/App.hpp"
#include "../core/LayerTreeController.hpp"
#include "../core/LayerTreeView.hpp"
#include "../core/LayerTreeItem.hpp"
#include "../core/GeoTreeController.hpp"
#include "../core/LocationView.hpp"
#include "../core/SequenceModel.hpp"
#include "../core/LayerTree.hpp"
#include "../core/MapLayer.hpp"
#include "../core/LocationSetLayer.hpp"
#include "../core/LocationLayer.hpp"
#include "../core/SequenceLayer.hpp"
#include "../core/TreeLayer.hpp"
#include "../core/Viewport.hpp"

#include "../gui/StudyPropertiesDlg.hpp"
#include "../gui/MapPropertiesDlg.hpp"
#include "../gui/LocationSetPropertiesDlg.hpp"
#include "../gui/LocationPropertiesDlg.hpp"
#include "../gui/SequencePropertiesDlg.hpp"
#include "../gui/TreePropertiesDlg.hpp"

using namespace GenGIS;

LayerTreeController::LayerTreeController(LayerTreePtr tree, LayerTreeViewPtr treeView)
		: m_layerTreeModel(tree), m_treeView(treeView)
{

}

const std::wstring& LayerTreeController::GetName(uint id) const 
{ 
	return m_layerTreeModel->GetLayer(id)->GetName(); 
}

const std::wstring& LayerTreeController::GetDescription(uint id) const 
{ 
	return m_layerTreeModel->GetLayer(id)->GetDescription(); 
}

void LayerTreeController::SetDescription(uint id, const std::wstring& description) 
{ 
	m_layerTreeModel->GetLayer(id)->SetDescription(description); 
}

const std::wstring& LayerTreeController::GetAuthours(uint id) const 
{ 
	return m_layerTreeModel->GetLayer(id)->GetAuthours(); 
}

void LayerTreeController::SetAuthours(uint id, const std::wstring& authours) 
{ 
	m_layerTreeModel->GetLayer(id)->SetAuthours(authours); 
}

const std::wstring& LayerTreeController::GetPath(uint id) const 
{ 
	return m_layerTreeModel->GetLayer(id)->GetPath(); 
}

const std::wstring& LayerTreeController::GetFilename(uint id) const 
{ 
	return m_layerTreeModel->GetLayer(id)->GetFilename(); 
}

const bool LayerTreeController::IsVisible(uint id) const 
{ 
	return m_layerTreeModel->GetLayer(id)->IsVisible(); 
}

void LayerTreeController::OnRender() 
{ 
	m_layerTreeModel->Render(); 
}

uint LayerTreeController::GetNumStudyLayers() 
{ 
	return 1; 
}

StudyLayerPtr LayerTreeController::GetStudyLayer(uint index) 
{ 
	return m_layerTreeModel->GetStudy(index); 
}

uint LayerTreeController::GetNumMapLayers() 
{ 
	return m_maps.size(); 
}

MapLayerPtr LayerTreeController::GetMapLayer(uint index) 
{ 
	return m_maps.at(index); 
}

uint LayerTreeController::GetNumLocationSetLayers() 
{ 
	return m_locationSets.size(); 
}

LocationSetLayerPtr LayerTreeController::GetLocationSetLayer(uint index) 
{ 
	return m_locationSets.at(index); 
}

uint LayerTreeController::GetNumLocationLayers() 
{ 
	return m_locations.size(); 
}

LocationLayerPtr LayerTreeController::GetLocationLayer(uint index) 
{ 
	return m_locations.at(index); 
}

uint LayerTreeController::GetNumSequenceLayers() 
{ 
	return m_sequences.size(); 
}

SequenceLayerPtr LayerTreeController::GetSequenceLayer(uint index) 
{ 
	return m_sequences.at(index); 
}

uint LayerTreeController::GetNumTreeLayers() 
{ 
	return m_trees.size(); 
}

TreeLayerPtr LayerTreeController::GetTreeLayer(uint index) 
{ 
	return m_trees.at(index); 
}

wxTreeCtrl* LayerTreeController::GetTreeCtrl() 
{ 
	return m_treeView->GetTreeCtrl(); 
}

LayerTreeViewPtr LayerTreeController::GetTreeView() 
{ 
	return m_treeView; 
}

LayerPtr LayerTreeController::GetSelectedLayer()
{
	wxTreeItemId selectionId = m_treeView->GetTreeCtrl()->GetSelection();
	if(!selectionId.IsOk())
		return LayerPtr();

	LayerTreeItem* selectionItem = (LayerTreeItem*)m_treeView->GetTreeCtrl()->GetItemData(selectionId);

	return selectionItem->GetLayer();
}

void LayerTreeController::SetName(uint id, const std::wstring& name)
{
	m_layerTreeModel->GetLayer(id)->SetName(name);
	m_treeView->SetName(m_treeView->GetTreeCtrl()->GetSelection(), name);
}

void LayerTreeController::SetVisible(uint id, bool visibility) 
{ 
	m_layerTreeModel->GetLayer(id)->SetVisible(visibility); 

	wxTreeItemId selectionId = m_treeView->GetTreeCtrl()->GetSelection();
	if(!selectionId.IsOk())
		return;

	m_treeView->SetChecked(selectionId, m_layerTreeModel->GetLayer(id)->IsVisible());

	App::Inst().GetViewport()->Refresh(false);
}

LocationControllerPtr LayerTreeController::GetLocationController(const std::wstring& id)
{
	for(uint i = 0; i < GetNumLocationLayers(); ++i)
	{
		LocationLayerPtr locationLayer = GetLocationLayer(i);
		if(locationLayer->GetLocationController()->GetId() == id)
		{
			return locationLayer->GetLocationController();
		}
	}

	return LocationControllerPtr();
}

bool LayerTreeController::AddMap(MapLayerPtr map)
{
	wxTreeItemId selectionId = m_treeView->GetTreeCtrl()->GetSelection();
	
	LayerPtr selectedLayer = GetSelectedLayer();
	if(selectedLayer != LayerPtr() && selectedLayer->GetType() == Layer::STUDY)
	{
		// add map to tree controller so it can be easily access
		m_maps.push_back(map);

		// add map to tree model
		StudyLayerPtr studyPtr = boost::dynamic_pointer_cast<StudyLayer>(selectedLayer);
		studyPtr->AddMap(map);		

		// add map to tree control
		LayerTreeItem* layerItem = new LayerTreeItem(map);
		wxString idStr(std::wstring(map->GetName() + _T(" : Map")).c_str());
		wxTreeItemId newId = m_treeView->GetTreeCtrl()->AppendItem(selectionId, idStr, 
																						LayerTreeView::CHECKED, -1, layerItem);

		// select the newly added layer
		m_treeView->GetTreeCtrl()->Expand(selectionId);
		m_treeView->GetTreeCtrl()->SelectItem(newId);

		App::Inst().GetViewport()->Refresh(false);
	}
	else
	{
		wxMessageBox(wxT("Please select a study node before adding a new map."),
												wxT("Select study node"), wxOK | wxICON_INFORMATION);

		return false;
	}

	return true;
}

bool LayerTreeController::AddTree(TreeLayerPtr tree)
{
	wxTreeItemId selectionId = m_treeView->GetTreeCtrl()->GetSelection();
	LayerPtr selectedLayer = GetSelectedLayer();

	if(selectedLayer != LayerPtr() && selectedLayer->GetType() == Layer::MAP)
	{
		// add tree to tree controller so it can be easily access
		m_trees.push_back(tree);

		// add tree layer to tree model
		MapLayerPtr mapPtr = boost::dynamic_pointer_cast<MapLayer>(selectedLayer);
		mapPtr->AddTree(tree);

		// add tree layer to tree control
		LayerTreeItem* layerItem = new LayerTreeItem(tree);
		wxString idStr(std::wstring(tree->GetName() + _T(" : Tree")).c_str());
		wxTreeItemId treeId = m_treeView->GetTreeCtrl()->AppendItem(selectionId, idStr, 
																																	LayerTreeView::CHECKED, -1, layerItem);

		// select the newly added layer
		m_treeView->GetTreeCtrl()->Expand(selectionId);
		m_treeView->GetTreeCtrl()->SelectItem(treeId);

		App::Inst().GetViewport()->Refresh(false);
	}
	else
	{
		wxMessageBox(wxT("Please select a map node before adding a new tree."),
												wxT("Select map node"), wxOK | wxICON_INFORMATION);

		return false;
	}

	return true;
}

bool LayerTreeController::AddLocationSet(LocationSetLayerPtr locationSet)
{
	wxTreeItemId selectionId = m_treeView->GetTreeCtrl()->GetSelection();
	
	LayerPtr selectedLayer = GetSelectedLayer();
	if(selectedLayer != LayerPtr() && selectedLayer->GetType() == Layer::MAP)
	{
		// add location set to tree controller so it can be easily access
		m_locationSets.push_back(locationSet);

		// add location set to tree model
		MapLayerPtr mapPtr = boost::dynamic_pointer_cast<MapLayer>(selectedLayer);
		mapPtr->AddLocationSet(locationSet);

		// add location set to tree control
		LayerTreeItem* layerItem = new LayerTreeItem(locationSet);
		wxString idStr(std::wstring(locationSet->GetName() + _T(" : Location Set")).c_str());
		wxTreeItemId locationSetId = m_treeView->GetTreeCtrl()->AppendItem(selectionId, idStr, 
																																					LayerTreeView::CHECKED, -1, layerItem);

		// select the newly added layer
		m_treeView->GetTreeCtrl()->Expand(selectionId);
		m_treeView->GetTreeCtrl()->SelectItem(locationSetId);

		// add each location to tree control
		for(unsigned int i = 0; i < locationSet->GetNumLocations(); ++i)
		{
			LocationLayerPtr location = locationSet->GetLocation(i);

			m_locations.push_back(location);

			LayerTreeItem* locationItem = new LayerTreeItem(location);
			wxString idStr(std::wstring(location->GetName() + _T(" : Location")).c_str());
			wxTreeItemId locationId = m_treeView->GetTreeCtrl()->AppendItem(locationSetId, idStr, 
																																					LayerTreeView::CHECKED, -1, locationItem);
		}

		App::Inst().GetViewport()->Refresh(false);
	}
	else
	{
		wxMessageBox(wxT("Please select a map node before adding a new location set."),
												wxT("Select map node"), wxOK | wxICON_INFORMATION);

		return false;
	}

	return true;
}

bool LayerTreeController::AddSequence(SequenceLayerPtr sequence, SequenceModelPtr sequenceModel, std::set<std::wstring>& missingLocations)
{
	wxTreeItemId selectionId = m_treeView->GetTreeCtrl()->GetSelection();
	
	LayerPtr selectedLayer = GetSelectedLayer();

	if(selectedLayer != LayerPtr() && selectedLayer->GetType() == Layer::LOCATION_SET)
	{
		// add sequence layer to tree controller so it can be easily access
		m_sequences.push_back(sequence);

		// add sequence layer to appropriate location layer
		LocationSetLayerPtr locationSetPtr = boost::dynamic_pointer_cast<LocationSetLayer>(selectedLayer);
		bool bFoundLocationLayer = false;
		wxTreeItemIdValue cookie;
		wxTreeItemId locationTreeId = m_treeView->GetTreeCtrl()->GetFirstChild(selectionId, cookie);
		for(unsigned int j = 0; j < locationSetPtr->GetNumLocations(); ++j)
		{
			LocationLayerPtr locationLayer = locationSetPtr->GetLocation(j);
			if(locationLayer->GetLocationController()->GetId() == sequenceModel->GetSiteId())
			{
				locationLayer->AddSequence(sequence);
				locationLayer->GetLocationController()->AddSequence(sequenceModel);

				// add sequence to tree control
				LayerTreeItem* layerItem = new LayerTreeItem(sequence);
				wxString idStr(std::wstring(sequence->GetName() + _T(" : Sequence")).c_str());
				m_treeView->GetTreeCtrl()->AppendItem(locationTreeId, idStr, LayerTreeView::CHECKED, -1, layerItem);

				bFoundLocationLayer = true;
				break;
			}

			// move to next child in wxTreeCtrl
			locationTreeId = m_treeView->GetTreeCtrl()->GetNextChild(selectionId, cookie);
		}	

		if(!bFoundLocationLayer)
		{
			missingLocations.insert(sequenceModel->GetSiteId());

		}
	}
	else
	{
		wxMessageBox(wxT("Please select a location set node before adding new sequence data."),
												wxT("Select location set node"), wxOK | wxICON_INFORMATION);

		return false;
	}

	return true;
}

void LayerTreeController::OnSelect( wxTreeEvent& event )
{
	event.Skip();
}

void LayerTreeController::OnClick( wxMouseEvent& event )
{
	event.Skip();

	int hitPos;
	wxTreeItemId hitId = m_treeView->GetTreeCtrl()->HitTest(event.GetPosition(), hitPos);

	if(hitPos & wxTREE_HITTEST_ONITEMICON)
	{
		LayerTreeItem* treeItem = (LayerTreeItem*)m_treeView->GetTreeCtrl()->GetItemData(hitId);
		bool bVisible = treeItem->GetLayer()->ToggleVisibility();
		m_treeView->SetChecked(hitId, bVisible);

		App::Inst().GetViewport()->Refresh(false);
	}
}

void LayerTreeController::OnLeftDClick( wxMouseEvent& event )
{
	// cause the tree item under the cursor to become selected
	wxTreeItemId id = GetTreeCtrl()->GetSelection();

	if(id.IsOk())
	{
		LayerTreeItem* layerItem = (LayerTreeItem*)GetTreeCtrl()->GetItemData(id);
		LayerPtr layer = layerItem->GetLayer();

		wxMenu menu;
		if(layer->GetType() == Layer::STUDY)
		{
			StudyLayerPtr studyLayer = boost::dynamic_pointer_cast<StudyLayer>(layer);
			StudyPropertiesDlg* dlg = new StudyPropertiesDlg(App::Inst().GetMainWindow(), studyLayer);
			dlg->SetPage(StudyPropertiesDlg::PROJECTION);
			dlg->Show();
		}
		else if(layer->GetType() == Layer::MAP)
		{
			MapLayerPtr mapLayer = boost::dynamic_pointer_cast<MapLayer>(layer);
			MapPropertiesDlg* dlg = new MapPropertiesDlg(App::Inst().GetMainWindow(), mapLayer);
			dlg->SetPage(MapPropertiesDlg::SYMBOLOGY);
			dlg->Show();
		}
		else if(layer->GetType() == Layer::LOCATION_SET)
		{
			LocationSetLayerPtr locationSetLayer = boost::dynamic_pointer_cast<LocationSetLayer>(layer);
			LocationSetPropertiesDlg* dlg = new LocationSetPropertiesDlg(App::Inst().GetMainWindow(), locationSetLayer);
			dlg->SetPage(LocationSetPropertiesDlg::SYMBOLOGY);
			dlg->Show();
		}
		else if(layer->GetType() == Layer::LOCATION)
		{
			LocationLayerPtr locationLayer = boost::dynamic_pointer_cast<LocationLayer>(layer);
			LocationPropertiesDlg* dlg = new LocationPropertiesDlg(App::Inst().GetMainWindow(), locationLayer);
			dlg->SetPage(LocationPropertiesDlg::SYMBOLOGY);
			dlg->Show();
		}
		else if(layer->GetType() == Layer::SEQUENCE)
		{
			SequenceLayerPtr sequenceLayer = boost::dynamic_pointer_cast<SequenceLayer>(layer);
			SequencePropertiesDlg* dlg = new SequencePropertiesDlg(App::Inst().GetMainWindow(), sequenceLayer);
			dlg->SetPage(SequencePropertiesDlg::GENERAL);
			dlg->Show();
		}
		else if(layer->GetType() == Layer::TREE)
		{
			TreeLayerPtr treeLayer = boost::dynamic_pointer_cast<TreeLayer>(layer);
			TreePropertiesDlg* dlg = new TreePropertiesDlg(App::Inst().GetMainWindow(), treeLayer);
			dlg->SetPage(TreePropertiesDlg::SYMBOLOGY);
			dlg->Show();
		}
	}
}

void LayerTreeController::OnLayerRemove( wxCommandEvent& event, LayerPtr& layer )
{
	wxTreeItemId selectionId = m_treeView->GetTreeCtrl()->GetSelection();

	// delete selected item and all of its children from the tree layer model
	LayerTreeItem* selectionItem = (LayerTreeItem*)m_treeView->GetTreeCtrl()->GetItemData(selectionId);
	layer = selectionItem->GetLayer();
	
	// remove layer from quick access array
	if(layer->GetType() == Layer::STUDY)
	{
		// Note:: studies can't be deleted at this point in time
		wxMessageBox(wxT("GenGIS currently supports only a single, persistent study layer."));
		return;
	}
	else
	{
		if(!EraseMap(layer->GetId()))
		{			
			if(!EraseTree(layer->GetId()))
			{
				if(!EraseLocationSet(layer->GetId()))
				{
					if(!EraseLocation(layer->GetId()))
					{
						if(!EraseSequence(layer->GetId()))
						{
							// should have erased something!
							error::Log::Inst().Error("OnLayerRemove(): Unhandle case encountered.");
						}
					}
				}
			}
		}
	}

	// determine layer from layer tree
	m_layerTreeModel->RemoveLayer(layer->GetId());

	// delete selected item and all of its children from the tree control
	m_treeView->GetTreeCtrl()->DeleteChildren(selectionId);
	m_treeView->GetTreeCtrl()->Delete(selectionId);

	App::Inst().GetViewport()->Refresh();
}	

bool LayerTreeController::EraseMap(uint layerId)
{
	for(uint i = 0; i < m_maps.size(); ++i)
	{
		if(m_maps.at(i)->GetId() == layerId)
		{
			EraseLocationSets(m_maps.at(i));
			EraseTrees(m_maps.at(i));
			m_maps.erase(m_maps.begin()+i);

			return true;
		}
	}

	return false;
}

void LayerTreeController::EraseTrees(MapLayerPtr mapLayer)
{
	for(uint i = 0; i < mapLayer->GetNumLocationSets(); ++i)
	{
		EraseTree(mapLayer->GetLocationSet(i)->GetId());
	}
}

bool LayerTreeController::EraseTree(uint layerId)
{
	for(uint i = 0; i < m_trees.size(); ++i)
	{
		if(m_trees.at(i)->GetId() == layerId)
		{
			m_trees.erase(m_trees.begin()+i);
			return true;
		}
	}

	return false;
}

void LayerTreeController::EraseLocationSets(MapLayerPtr mapLayer)
{
	for(uint i = 0; i < mapLayer->GetNumLocationSets(); ++i)
	{
		EraseLocationSet(mapLayer->GetLocationSet(i)->GetId());
	}
}

bool LayerTreeController::EraseLocationSet(uint layerId)
{
	for(uint i = 0; i < m_locationSets.size(); ++i)
	{
		if(m_locationSets.at(i)->GetId() == layerId)
		{
			EraseLocations(m_locationSets.at(i));
			m_locationSets.erase(m_locationSets.begin()+i);
			return true;
		}
	}

	return false;
}

void LayerTreeController::EraseLocations(LocationSetLayerPtr locationSetLayer)
{
	for(uint i = 0; i < locationSetLayer->GetNumLocations(); ++i)
	{
		EraseLocation(locationSetLayer->GetLocation(i)->GetId());
	}
}

bool LayerTreeController::EraseLocation(uint layerId)
{
	for(uint i = 0; i < m_locations.size(); ++i)
	{
		if(m_locations.at(i)->GetId() == layerId)
		{
			EraseSequences(m_locations.at(i));
			m_locations.erase(m_locations.begin()+i);
			return true;
		}
	}

	return false;
}

void LayerTreeController::EraseSequences(LocationLayerPtr locationLayer)
{
	for(uint i = 0; i < locationLayer->GetNumSequences(); ++i)
	{
		EraseSequence(locationLayer->GetSequence(i)->GetId());
	}
}

bool LayerTreeController::EraseSequence(uint layerId)
{
	for(uint i = 0; i < m_sequences.size(); ++i)
	{
		if(m_sequences.at(i)->GetId() == layerId)
		{
			m_sequences.erase(m_sequences.begin()+i);
			return true;
		}
	}

	return false;
}

void LayerTreeController::OnHideAll( wxCommandEvent& event )
{
	wxTreeItemIdValue cookie;
	wxTreeItemId rootId = m_treeView->GetTreeCtrl()->GetFirstChild(m_treeView->GetTreeCtrl()->GetRootItem(), cookie);
	SetVisibilityAllChildren(rootId, cookie, false);

	App::Inst().GetViewport()->Refresh(false);
}

void LayerTreeController::OnShowAll( wxCommandEvent& event )
{
	wxTreeItemIdValue cookie;
	wxTreeItemId rootId = m_treeView->GetTreeCtrl()->GetFirstChild(m_treeView->GetTreeCtrl()->GetRootItem(), cookie);
	SetVisibilityAllChildren(rootId, cookie, true);

	App::Inst().GetViewport()->Refresh(false);
}

void LayerTreeController::SetVisibilityAllChildren(wxTreeItemId id, wxTreeItemIdValue cookie, bool visibility)
{
	if(id.IsOk())
	{
		// modify current node
		if(visibility)
		{
			m_treeView->GetTreeCtrl()->SetItemTextColour(id, *wxBLACK);
		}
		else
		{
			m_treeView->GetTreeCtrl()->SetItemTextColour(id, *wxLIGHT_GREY);
		}

		LayerTreeItem* layerItem = (LayerTreeItem*)m_treeView->GetTreeCtrl()->GetItemData(id);
		layerItem->GetLayer()->SetVisible(visibility);
		m_treeView->SetChecked(id, visibility);

		// depth first traversal of tree
		wxTreeItemIdValue childCookie;
		wxTreeItemId childId = m_treeView->GetTreeCtrl()->GetFirstChild(id, childCookie);
		SetVisibilityAllChildren(childId, childCookie, visibility);
		
		wxTreeItemId siblingId = m_treeView->GetTreeCtrl()->GetNextSibling(id);
		SetVisibilityAllChildren(siblingId, cookie, visibility);
	}

	// no more children
	return;	
}

bool LayerTreeController::OnLayoutTree(GeoTreeView::LAYOUT layout, LayoutLinePtr layoutLine)
{
	LayerPtr selectedLayer = GetSelectedLayer();
	if(selectedLayer != LayerPtr() && selectedLayer->GetType() == Layer::TREE)
	{
		// change layout style of tree
		TreeLayerPtr treePtr = boost::dynamic_pointer_cast<TreeLayer>(selectedLayer);

		GeoTreeViewPtr geoTreeView = treePtr->GetGeoTreeController()->GetTreeView();

		if(layout == GeoTreeView::SLANTED_CLADOGRAM_3D || layout == GeoTreeView::CLADOGRAM_3D)
		{
			geoTreeView->SetLayout(layout);
			App::Inst().GetViewport()->Refresh(false);
			return false;
		}

		// check if a layout line has already been assigned to the tree or not
		if(geoTreeView->IsLayoutLine())
		{
			geoTreeView->SetLayout(layout);
			App::Inst().GetViewport()->Refresh(false);
			return false;
		}
		else
		{
			// a layout line is needed so make sure that there is a valid one to assign to the tree
			if(layoutLine)
			{
				geoTreeView->SetLayoutLine(layoutLine);
				geoTreeView->SetLayout(layout);
				App::Inst().GetViewport()->Refresh(false);
				return true;
			}
			else
			{
				wxMessageBox(wxT("Please draw a layout line."),
										wxT("Missing layout line"), wxOK | wxICON_INFORMATION);
			}
		}
	}
	else
	{
		// it should not be possible to call this function for a non-tree layer
		error::Log::Inst().Error("OnLayoutTree(): Unhandle case encountered.");
	}

	return false;
}

bool LayerTreeController::OnLayoutTree(GeoTreeView::LAYOUT layout, LayoutEllipsePtr layoutEllipse)
{
	LayerPtr selectedLayer = GetSelectedLayer();
	if(selectedLayer != LayerPtr() && selectedLayer->GetType() == Layer::TREE)
	{
		// change layout style of tree
		TreeLayerPtr treePtr = boost::dynamic_pointer_cast<TreeLayer>(selectedLayer);

		GeoTreeViewPtr geoTreeView = treePtr->GetGeoTreeController()->GetTreeView();

		if(layout == GeoTreeView::SLANTED_CLADOGRAM_3D || layout == GeoTreeView::CLADOGRAM_3D)
		{
			geoTreeView->SetLayout(layout);
			App::Inst().GetViewport()->Refresh(false);
			return false;
		}

		// check if a layout ellipse has already been assigned to the tree or not
		if(geoTreeView->IsLayoutEllipse())
		{
			geoTreeView->SetLayout(layout);
			App::Inst().GetViewport()->Refresh(false);
			return false;
		}
		else
		{
			// a layout ellipse is needed so make sure that there is a valid one to assign to the tree
			if(layoutEllipse)
			{
				geoTreeView->SetLayoutEllipse(layoutEllipse);
				geoTreeView->SetLayout(layout);
				App::Inst().GetViewport()->Refresh(false);
				return true;
			}
			else
			{
				wxMessageBox(wxT("Please draw a layout ellipse."),
										wxT("Missing layout ellipse"), wxOK | wxICON_INFORMATION);
			}
		}
	}
	else
	{
		// it should not be possible to call this function for a non-tree layer
		error::Log::Inst().Error("OnLayoutTree(): Unhandle case encountered.");
	}

	return false;
}

bool LayerTreeController::OnAssignGeoAxisToTree(GeoAxisPolylinePtr geoAxisPolyline)
{
	LayerPtr selectedLayer = GetSelectedLayer();
	if(selectedLayer != LayerPtr() && selectedLayer->GetType() == Layer::TREE)
	{
		// assign geographical axis to tree
		TreeLayerPtr treePtr = boost::dynamic_pointer_cast<TreeLayer>(selectedLayer);

		GeoTreeViewPtr geoTreeView = treePtr->GetGeoTreeController()->GetTreeView();

		// check that the geographical axis has been defined before assigning it to the tree
		if(geoAxisPolyline)
		{
			geoTreeView->SetGeographicAxis(geoAxisPolyline);
			App::Inst().GetViewport()->Refresh(false);
			return true;
		}
		else
		{
			wxMessageBox(wxT("A geographic axis polyline must be constructed before it can be assigned to a tree."),
									wxT("Missing geographical axis"), wxOK | wxICON_INFORMATION);
		}
	}
	else
	{
		// it should not be possible to call this function for a non-tree layer
		error::Log::Inst().Error("OnAssignGeoAxisToTree(): Unhandle case encountered.");
	}

	return false;
}

void LayerTreeController::OnReverseGeoAxisOrder(bool state)
{
	LayerPtr selectedLayer = GetSelectedLayer();
	if(selectedLayer != LayerPtr() && selectedLayer->GetType() == Layer::TREE)
	{
		// assign geographical axis to tree
		TreeLayerPtr treePtr = boost::dynamic_pointer_cast<TreeLayer>(selectedLayer);

		GeoTreeViewPtr geoTreeView = treePtr->GetGeoTreeController()->GetTreeView();

		geoTreeView->SetReverseOrderGeographicPts(state);
	}
	else
	{
		// it should not be possible to call this function for a non-tree layer
		error::Log::Inst().Error("OnReverseGeoAxisOrder(): Unhandle case encountered.");
	}
}

bool LayerTreeController::OnLayoutCharts(ChartSetView::LAYOUT layout, LayoutLinePtr layoutLine)
{
	LayerPtr selectedLayer = GetSelectedLayer();
	if(selectedLayer != LayerPtr() && selectedLayer->GetType() == Layer::LOCATION_SET)
	{
		// assign layout line to chart set
		LocationSetLayerPtr locationSetLayer = boost::dynamic_pointer_cast<LocationSetLayer>(selectedLayer);

		ChartSetViewPtr chartSetView = locationSetLayer->GetChartSetView();

		// check if a layout line has already been assigned 
		if(chartSetView->IsLayoutLine())
		{
			chartSetView->SetLayout(layout);
			App::Inst().GetViewport()->Refresh(false);
			return false;
		}
		else
		{
			// a layout line is needed so make sure that there is a valid one to assign
			if(layoutLine)
			{
				chartSetView->SetLayoutLine(layoutLine);
				chartSetView->SetLayout(layout);
				App::Inst().GetViewport()->Refresh(false);
				return true;
			}
			else
			{
				wxMessageBox(wxT("Please draw a layout line."),
										wxT("Missing layout line"), wxOK | wxICON_INFORMATION);
			}
		}
	}
	else
	{
		// it should not be possible to call this function for a location set layer
		error::Log::Inst().Error("OnLayoutCharts(): Unhandle case encountered.");
	}

	return false;
}

bool LayerTreeController::OnLayoutCharts(ChartSetView::LAYOUT layout, LayoutEllipsePtr layoutEllipse)
{
	LayerPtr selectedLayer = GetSelectedLayer();
	if(selectedLayer != LayerPtr() && selectedLayer->GetType() == Layer::LOCATION_SET)
	{
		// assign layout ellipse to chart set
		LocationSetLayerPtr locationSetLayer = boost::dynamic_pointer_cast<LocationSetLayer>(selectedLayer);

		ChartSetViewPtr chartSetView = locationSetLayer->GetChartSetView();

		// check if a layout ellipse has already been assigned 
		if(chartSetView->IsLayoutEllipse())
		{
			chartSetView->SetLayout(layout);
			App::Inst().GetViewport()->Refresh(false);
			return false;
		}
		else
		{
			// a layout ellipse is needed so make sure that there is a valid one to assign
			if(layoutEllipse)
			{
				chartSetView->SetLayoutEllipse(layoutEllipse);
				chartSetView->SetLayout(layout);
				App::Inst().GetViewport()->Refresh(false);
				return true;
			}
			else
			{
				wxMessageBox(wxT("Please draw a layout ellipse."),
										wxT("Missing layout ellipse"), wxOK | wxICON_INFORMATION);
			}
		}
	}
	else
	{
		// it should not be possible to call this function for a location set layer
		error::Log::Inst().Error("OnLayoutCharts(): Unhandle case encountered.");
	}

	return false;
}

void LayerTreeController::OnTreeLayoutLineVisibility()
{
	LayerPtr selectedLayer = GetSelectedLayer();
	if(selectedLayer != LayerPtr() && selectedLayer->GetType() == Layer::TREE)
	{
		// change visibility of layout line
		TreeLayerPtr treePtr = boost::dynamic_pointer_cast<TreeLayer>(selectedLayer);
		GeoTreeViewPtr geoTreeView = treePtr->GetGeoTreeController()->GetTreeView();
		geoTreeView->SetLayoutLineVisibility(!geoTreeView->GetLayoutLineVisibility());		

		App::Inst().GetViewport()->Refresh(false);
	}
	else
	{
		// it should not be possible to call this function for a non-tree layer
		error::Log::Inst().Error("OnTreeLayoutLineVisibility(): Unhandle case encountered.");
	}
}

void LayerTreeController::OnTreeLayoutEllipseVisibility()
{
	LayerPtr selectedLayer = GetSelectedLayer();
	if(selectedLayer != LayerPtr() && selectedLayer->GetType() == Layer::TREE)
	{
		// change visibility of layout ellipse
		TreeLayerPtr treePtr = boost::dynamic_pointer_cast<TreeLayer>(selectedLayer);
		GeoTreeViewPtr geoTreeView = treePtr->GetGeoTreeController()->GetTreeView();
		geoTreeView->SetLayoutEllipseVisibility(!geoTreeView->GetLayoutEllipseVisibility());		

		App::Inst().GetViewport()->Refresh(false);
	}
	else
	{
		// it should not be possible to call this function for a non-tree layer
		error::Log::Inst().Error("OnTreeLayoutEllipseVisibility(): Unhandle case encountered.");
	}
}

void LayerTreeController::OnTreeGeographicAxisVisibility()
{
	LayerPtr selectedLayer = GetSelectedLayer();
	if(selectedLayer != LayerPtr() && selectedLayer->GetType() == Layer::TREE)
	{
		// change visibility of geographic axis
		TreeLayerPtr treePtr = boost::dynamic_pointer_cast<TreeLayer>(selectedLayer);
		GeoTreeViewPtr geoTreeView = treePtr->GetGeoTreeController()->GetTreeView();
		geoTreeView->SetGeographicAxisVisibility(!geoTreeView->GetGeographicAxisVisibility());		

		App::Inst().GetViewport()->Refresh(false);
	}
	else
	{
		// it should not be possible to call this function for a non-tree layer
		error::Log::Inst().Error("OnTreeGeographicAxisVisibility(): Unhandle case encountered.");
	}
}

void LayerTreeController::OnChartSetLayoutLine()
{
	LayerPtr selectedLayer = GetSelectedLayer();
	if(selectedLayer != LayerPtr() && selectedLayer->GetType() == Layer::LOCATION_SET)
	{
		// change layout style of tree
		LocationSetLayerPtr locationSet = boost::dynamic_pointer_cast<LocationSetLayer>(selectedLayer);
		ChartSetViewPtr chartSetView = locationSet->GetChartSetView();
		chartSetView->SetLayoutLineVisibility(!chartSetView->GetLayoutLineVisibility());		

		App::Inst().GetViewport()->Refresh(false);
	}
	else
	{
		// it should not be possible to call this function for a non-tree layer
		error::Log::Inst().Error("OnChartSetLayoutLine(): Unhandle case encountered.");
	}
}

void LayerTreeController::OnChartSetLayoutEllipse()
{
	LayerPtr selectedLayer = GetSelectedLayer();
	if(selectedLayer != LayerPtr() && selectedLayer->GetType() == Layer::LOCATION_SET)
	{
		// change layout style of tree
		LocationSetLayerPtr locationSet = boost::dynamic_pointer_cast<LocationSetLayer>(selectedLayer);
		ChartSetViewPtr chartSetView = locationSet->GetChartSetView();
		chartSetView->SetLayoutEllipseVisibility(!chartSetView->GetLayoutEllipseVisibility());		

		App::Inst().GetViewport()->Refresh(false);
	}
	else
	{
		// it should not be possible to call this function for a non-tree layer
		error::Log::Inst().Error("OnChartSetLayoutEllipse(): Unhandle case encountered.");
	}
}

void LayerTreeController::OnLeafNodeLabels()
{
	LayerPtr selectedLayer = GetSelectedLayer();
	if(selectedLayer != LayerPtr() && selectedLayer->GetType() == Layer::TREE)
	{
		// change layout style of tree
		TreeLayerPtr treePtr = boost::dynamic_pointer_cast<TreeLayer>(selectedLayer);
		GeoTreeViewPtr geoTreeView = treePtr->GetGeoTreeController()->GetTreeView();
		geoTreeView->SetFontVisibility(!geoTreeView->GetFontVisibility());		

		App::Inst().GetViewport()->Refresh(false);
	}
	else
	{
		// it should not be possible to call this function for a non-tree layer
		error::Log::Inst().Error("OnLeafNodeLabels(): Unhandle case encountered.");
	}
}

bool LayerTreeController::MouseLeftDown(const Point2D& mousePt)
{
	// Note: Although it might seem this function could return once if finds a valid hit,
	// this is not the case. Many visual objects need to be informed when a new object is
	// selected (i.e., that they have become unselected).
	bool bHit = false;

	App::Inst().SetSelectedView(ViewPtr());

	// iterate through all views in location set layers
	for(uint i = 0; i < GetNumLocationSetLayers(); ++i)
	{
		LocationSetLayerPtr locationSetLayer = GetLocationSetLayer(i);

		if(locationSetLayer->IsVisible())
		{
			if(locationSetLayer->GetChartSetView()->MouseLeftDown(mousePt))
			{
				App::Inst().SetSelectedView(locationSetLayer->GetChartSetView());
				bHit = true;
			}
		}
	}

	// iterate through all views in location layers
	for(uint i = 0; i < GetNumLocationLayers(); ++i)
	{
		LocationLayerPtr locationLayer = GetLocationLayer(i);

		if(locationLayer->IsVisible())
		{
			LocationControllerPtr locationController = locationLayer->GetLocationController();

			// check chart view
			if(locationController->GetChartView()->MouseLeftDown(mousePt))
			{
				App::Inst().SetSelectedView(locationController->GetChartView());
				bHit = true;
				break;
			}

			// check location view
			if(locationController->GetLocationView()->MouseLeftDown(mousePt))
			{
				App::Inst().SetSelectedView(locationController->GetLocationView());
				bHit = true;
				break;
			}
		}
	}

	// iterate through all views in tree layers
	for(uint i = 0; i < GetNumTreeLayers(); ++i)
	{
		TreeLayerPtr treeLayer = GetTreeLayer(i);

		if(treeLayer->IsVisible())
		{
			// check tree view
			if(treeLayer->GetGeoTreeController()->GetTreeView()->MouseLeftDown(mousePt))
			{
				App::Inst().SetSelectedView(treeLayer->GetGeoTreeController()->GetTreeView());
				bHit = true;
				break;
			}
		}
	}

	return bHit;
}

bool LayerTreeController::MouseRightDown(const Point2D& mousePt, wxMenu& popupMenu)
{
	// Note: Although it would seem this function could return once if finds a valid hit,
	// this is not the case. Many visual objects need to be informed when a new object is
	// selected. 
	bool bHit = false;

	// iterate through all views in tree layers
	for(uint i = 0; i < GetNumTreeLayers(); ++i)
	{
		TreeLayerPtr treeLayer = GetTreeLayer(i);

		if(treeLayer->IsVisible())
		{
			// check tree view
			if(treeLayer->GetGeoTreeController()->GetTreeView()->MouseRightDown(mousePt, popupMenu))
			{
				if(!bHit)
					App::Inst().SetRightClickView(treeLayer->GetGeoTreeController()->GetTreeView());

				bHit = true;
			}
		}
	}

	// iterate through all views in location set layers
	for(uint i = 0; i < GetNumLocationSetLayers(); ++i)
	{
		LocationSetLayerPtr locationSetLayer = GetLocationSetLayer(i);

		if(locationSetLayer->IsVisible())
		{
			if(locationSetLayer->GetChartSetView()->MouseRightDown(mousePt, popupMenu))
			{
				if(!bHit)
					App::Inst().SetRightClickView(locationSetLayer->GetChartSetView());

				bHit = true;
			}
		}
	}

	// iterate through all views in location layers
	for(uint i = 0; i < GetNumLocationLayers(); ++i)
	{
		LocationLayerPtr locationLayer = GetLocationLayer(i);

		if(locationLayer->IsVisible())
		{
			// check location view
			if(locationLayer->GetLocationController()->GetLocationView()->MouseRightDown(mousePt, popupMenu))
			{
				if(!bHit)
					App::Inst().SetRightClickView(locationLayer->GetLocationController()->GetLocationView());

				bHit = true;
			}

			// check chart view
			if(locationLayer->GetLocationController()->GetChartView()->MouseRightDown(mousePt, popupMenu))
			{
				if(!bHit)
					App::Inst().SetRightClickView(locationLayer->GetLocationController()->GetChartView());

				bHit = true;
			}
		}
	}

	return bHit;
}
