//=======================================================================
// Author: Donovan Parks
//
// Copyright 2009 Donovan Parks
//
// This file is part of GenGIS.
//
// GenGIS is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// GenGIS is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with GenGIS.  If not, see <http://www.gnu.org/licenses/>.
//=======================================================================

#include "../core/Precompiled.hpp"

#include "../core/LayerTreeView.hpp"
#include "../core/LayerTreeItem.hpp"
#include "../core/StudyController.hpp"
#include "../core/StudyModel.hpp"
#include "../core/StudyLayer.hpp"
#include "../core/LayerTree.hpp"

#include "../utils/UniqueId.hpp"

#include "../../bin/images/checked.xpm"
#include "../../bin/images/unchecked.xpm"

using namespace GenGIS;

LayerTreeView::LayerTreeView(LayerTreePtr tree, wxTreeCtrl* treeCtrl): m_tree(tree), m_treeCtrl(treeCtrl)
{
	// setup tree control
	const int IMAGE_SIZE = 16;
	wxImageList *images = new wxImageList(IMAGE_SIZE, IMAGE_SIZE, true);

	wxIcon image[2];
	image[0] = wxIcon(unchecked_xpm);
	image[1] = wxIcon(checked_xpm);

	for(size_t i = 0; i < WXSIZEOF(image); i++)
	{
		images->Add(image[i]);
	}

  m_treeCtrl->AssignImageList(images);
	m_treeCtrl->AddRoot(wxT("Root"), 0);

	// add initial study to tree model
	StudyModelPtr studyModel(new StudyModel());
	StudyControllerPtr studyController(new StudyController(studyModel));
	StudyLayerPtr study(new StudyLayer(UniqueId::Inst().GenerateId(), LayerPtr(), studyController));
	study->SetName(_T("New Study"));
	m_tree->AddStudy(study);

	// add initial study to tree view
	LayerTreeItem* layerItem = new LayerTreeItem(study);
	wxString idStr(std::wstring(study->GetName() + _T(" : Study")).c_str() );
	m_treeCtrl->AppendItem(m_treeCtrl->GetRootItem(), idStr, LayerTreeView::CHECKED, -1, layerItem);
}

void LayerTreeView::SetChecked(const wxTreeItemId& treeId, bool bVisible)
{
	if(bVisible)
	{
		m_treeCtrl->SetItemImage(treeId, CHECKED);
	}
	else
	{	
		m_treeCtrl->SetItemImage(treeId, UNCHECKED);
	}

	// set to black or light grey all items below checked/unchecked node
	wxColour colour;
	bool stopAtUnchecked;
	if(bVisible)
	{
		colour = *wxBLACK;
		stopAtUnchecked = true;
	}
	else
	{
		colour = *wxLIGHT_GREY;
		stopAtUnchecked = false;
	}

	// traverse all nodes below the hit node
	wxTreeItemIdValue cookie;
	wxTreeItemId childId = m_treeCtrl->GetFirstChild(treeId, cookie);
	SetColourAllChildren(childId, cookie, stopAtUnchecked, colour);		
}

void LayerTreeView::SetColourAllChildren(wxTreeItemId id, wxTreeItemIdValue cookie, bool stopAtUnchecked, wxColour colour)
{
	if(id.IsOk())
	{
		// modify current node
		m_treeCtrl->SetItemTextColour(id, colour);

		if(!stopAtUnchecked || m_treeCtrl->GetItemImage(id) == LayerTreeView::CHECKED)
		{
			// do not travere children that are currently unchecked when we have just checked a parent node
			wxTreeItemIdValue childCookie;
			wxTreeItemId childId = m_treeCtrl->GetFirstChild(id, childCookie);
			SetColourAllChildren(childId, childCookie, stopAtUnchecked, colour);
		}
		
		wxTreeItemId siblingId = m_treeCtrl->GetNextSibling(id);
		SetColourAllChildren(siblingId, cookie, stopAtUnchecked, colour);
	}

	// no more children
	return;
}

void LayerTreeView::SetName(const wxTreeItemId& treeId, const std::wstring& name) 
{ 
	LayerTreeItem* layerItem = (LayerTreeItem*)m_treeCtrl->GetItemData(treeId);
	Layer::LAYER_TYPE layerType = layerItem->GetLayer()->GetType();

	std::wstring typeStr = Layer::GetLayerTypeString(layerType);
	std::wstring treeName = name + std::wstring(_T(" : ")) + typeStr;

	m_treeCtrl->SetItemText(treeId, wxString(treeName.c_str()));
}

