//=======================================================================
// Author: Donovan Parks
//
// Copyright 2009 Donovan Parks
//
// This file is part of GenGIS.
//
// GenGIS is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// GenGIS is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with GenGIS.  If not, see <http://www.gnu.org/licenses/>.
//=======================================================================

#ifndef _GENGIS_LOCATION_CONTROLLER
#define _GENGIS_LOCATION_CONTROLLER

#include "../core/Precompiled.hpp"

#include "../core/LocationModel.hpp"
#include "../core/LocationView.hpp"
#include "../core/ChartView.hpp"

namespace GenGIS
{

/**
 * @brief Handle events related to a given location.
 */
class LocationController 
{
public:
	/** Types of visual objects (views) which can be selected. */
	enum SELECTED_OBJ { LOCATION_MARKER, CHART };

public:
	/** 
	 * @brief Constuctor.
	 * @param locationModel Location model to associate with this controller.
	 * @param locationView Location view to associate with this controller.
	 * @param chartView Pie chart view to associate with this controller.
	 */
	LocationController(LocationModelPtr locationModel, LocationViewPtr locationView, ChartViewPtr chartView)
		: m_locationModel(locationModel), m_locationView(locationView), m_chartView(chartView), m_bRenderChart(true) {}

	/** Destructor. */
	virtual ~LocationController() {}

	/** Get name of location site. */
	virtual const std::wstring GetName() { return m_locationModel->GetName(); }

	/** Get id of location site. */
	virtual const std::wstring GetId() { return m_locationModel->GetId(); }

	/** Get northing position of location. */
	virtual const float& GetNorthing() const { return m_locationModel->GetNorthing(); }

	/** Get easting position of location. */
	virtual const float& GetEasting() const { return m_locationModel->GetEasting(); }

	/** 
	 * @brief Get data for specified field. 
	 * @param field Field to get data for. 
	 * @return Data associated with the provided field.
	 */
	const std::wstring& GetData(const std::wstring& field) { return m_locationModel->GetData(field); }

	/** Get projected coordinates of location. */
	GeoCoord GetProjection() { return m_locationModel->GetProjection(); }

	/** Render location model. */
	void Render() 
	{ 
		if(m_locationView)
			m_locationView->Render(); 

		if(m_chartView && m_bRenderChart)
			m_chartView->Render();
	}

	/**
	 * @brief Add sequence to location.
	 * @param sequence Sequence to add.
	 */
	void AddSequence(SequenceModelPtr sequence) { m_locationModel->AddSequence(sequence); }

	/** Get number of sequence at this location. */
	unsigned int GetNumSequence() const { return m_locationModel->GetNumSequence(); }

	/**
	 * @brief Get sequence from location.
	 * @param index Index of desired sequence.
	 * @return Sample at given index.
	 */
	SequenceModelPtr GetSequence(unsigned int index) const { return m_locationModel->GetSequence(index); }

	/**
	 * @brief Remove sequence from location.
	 * @param index Sequence to remove.
	 */
	void RemoveSequence(unsigned int index) { m_locationModel->RemoveSequence(index); }

	/** Remove all sequence from location. */
	void RemoveAllSequence() { m_locationModel->RemoveAllSequence(); }

	/** Get 3D location marker associated with this controller. */
	LocationViewPtr GetLocationView() { return m_locationView; }

	/** Get chart view associated with this controller. */
	ChartViewPtr GetChartView() { return m_chartView; }

	/** Get data associate with location site. */
	std::map<std::wstring,std::wstring> GetData() { return m_locationModel->GetData(); }

	/** Set flag indicating if chart should be rendered. */
	void SetRenderChart(bool bRender) { m_bRenderChart = bRender; }

	/** Get flag indicating if chart will be rendered. */
	bool GetRenderChart() { return m_bRenderChart; }

	/** Set colour of location. */
	void SetColour(const Colour& colour) { m_locationView->SetColour(colour); }

	/** Set size of location. */
	void SetSize(float size) { m_locationView->SetSize(size); }

	/** Set visibility of location. */
	void SetVisibility(bool state) { m_locationView->SetVisibility(state); }

	/** Get rendering order. */
	float GetRenderingOrder() { return m_locationView->GetRenderingOrder(); }

	/** Set rendering order. */
	void SetRenderingOrder(float renderingOrder) { m_locationView->SetRenderingOrder(renderingOrder); }

protected:
	/** Location model associated with this controller. */
	LocationModelPtr m_locationModel;

	/** Location view associated with this controller. */
	LocationViewPtr m_locationView;

	/** Pie chart view associated with this controller. */
	ChartViewPtr m_chartView;

	/** Flag indicating if the chart associated with this location should be rendered. */
	bool m_bRenderChart;
};

}

#endif
