//=======================================================================
// Author: Donovan Parks
//
// Copyright 2009 Donovan Parks
//
// This file is part of GenGIS.
//
// GenGIS is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// GenGIS is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with GenGIS.  If not, see <http://www.gnu.org/licenses/>.
//=======================================================================

#ifndef _GENGIS_LOCATION_LAYER_
#define _GENGIS_LOCATION_LAYER_

#include "../core/Precompiled.hpp"

#include "../core/Layer.hpp"

namespace GenGIS 
{

/**
 * @brief Location of samples in projected coordinates. A single location may be associated with multiple samples.
 */
class LocationLayer : public Layer
{
public:
	/** 
	 * @brief Constuctor.
	 * @param id Unique identifier used to specify location.
	 * @param parent Parent of this layer.
	 * @param location Location controller to associate with this layer.
	 */
	LocationLayer(uint id, LayerPtr parent, LocationControllerPtr locationController)
		: Layer(id, Layer::LOCATION, parent), m_locationController(locationController) {}

	/** Destructor. */
	virtual ~LocationLayer() {}

	/**
	 * @brief Get layer from tree below this layer. 
	 * @param id Unique id of desired layer.
	 * @return layer at given id.
	 */
	virtual LayerPtr GetLayer(uint id) const;

	/**
	 * @brief Remove layer from tree below this layer. 
	 * @param id Unique id of layer to remove.
	 * @return True if layer removed, false if layer not found.
	 */
	virtual bool RemoveLayer(uint id);

	/** Render contents of layer. */
	virtual void Render();

	/**
	 * @brief Add sequence to location.
	 * @param sequence Sequence to add.
	 */
	void AddSequence(SequenceLayerPtr sequence) { m_sequence.push_back(sequence); }

	/** Get number of samples at this location. */
	unsigned int GetNumSequences() const { return m_sequence.size(); }

	/**
	 * @brief Get sample from location.
	 * @param index Index of desired sample.
	 * @return Sample at given index.
	 */
	SequenceLayerPtr GetSequence(uint index) const { return m_sequence.at(index); }

	/**
	 * @brief Remove sample from location.
	 * @param index Sample to remove.
	 */
	void RemoveSequence(uint index);

	/** Remove all samples. */
	void RemoveAllSequences() { m_sequence.clear(); }

	/** Get location controller associated with this layer. */
	LocationControllerPtr GetLocationController() { return m_locationController; }
  
	/** Get vector of all sequence layers. */
  const std::vector<SequenceLayerPtr> GetAllSequences() { return m_sequence; }
  
private:
	/** All samples collected at this location. */
	std::vector<SequenceLayerPtr> m_sequence;

	/** Controller for location model. */
	LocationControllerPtr m_locationController;
};

}

#endif
