//=======================================================================
// Author: Donovan Parks
//
// Copyright 2009 Donovan Parks
//
// This file is part of GenGIS.
//
// GenGIS is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// GenGIS is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with GenGIS.  If not, see <http://www.gnu.org/licenses/>.
//=======================================================================

#ifndef _GENGIS_LOCATION_MODEL_
#define _GENGIS_LOCATION_MODEL_

#include "../core/Precompiled.hpp"

namespace GenGIS 
{

/**
 * @brief Location of samples in projected coordinate space.
 */
class LocationModel
{
public:
	/** 
	 * @brief Constuctor.
	 * @param locationName Descriptive name of location site.
	 * @param siteId Unique id for location site.
	 * @param northing Northing coordinate of location.
	 * @param easting Easting coordinate of locaiton.
	 * @param environmentType Type of environment of location.
	 * @param data Data associated with this location.
	 */
	LocationModel(const std::wstring& locationName, const std::wstring& siteId, float northing, float easting, 
									const std::wstring& environmentType, std::map<std::wstring,std::wstring> data)
		: m_name(locationName), m_id(siteId), m_northing(northing), m_easting(easting), m_environmentType(environmentType),
			m_data(data) {} 

	/** Destructor. */
	virtual ~LocationModel() {}

	/** Get name of location site. */
	const std::wstring GetName() { return m_name; }

	/** Get id of location site. */
	const std::wstring GetId() { return m_id; }

	/** Get type of environment. */
	const std::wstring GetEnvironmentType() { return m_environmentType; }

	/**
	 * @brief Set northing coordinates of location.
	 * @param Northing coordinates.
	 */
  void SetNorthing( float northing ) { m_northing = northing;  }

	/** Get northing position of location. */
	const float& GetNorthing() const { return m_northing; }

	/**
	 * @brief Set easting coordinates of location.
	 * @param Easting coordinates.
	 */
  void SetEasting( float easting ) { m_easting = easting; }

	/** Get easting position of location. */
	const float& GetEasting() const { return m_easting; }

	/** Get projected coordinates of location. */
	GeoCoord GetProjection() { return GeoCoord(m_easting, m_northing); }

	/**
	 * @brief Add sequence to location.
	 * @param sequence Sequence to add.
	 */
	void AddSequence(SequenceModelPtr sequence) { m_sequences.push_back(sequence); }

	/** Get number of sequence at this location. */
	unsigned int GetNumSequence() const { return m_sequences.size(); }

	/**
	 * @brief Get sequence from location.
	 * @param index Index of desired sequence.
	 * @return Sample at given index.
	 */
	SequenceModelPtr GetSequence(unsigned int index) const { return m_sequences.at(index); }

	/**
	 * @brief Remove sequence from location.
	 * @param index Sequence to remove.
	 */
	void RemoveSequence(unsigned int index) { m_sequences.erase(m_sequences.begin() + index); }

	/** Remove all sequence from location. */
	void RemoveAllSequence() { m_sequences.clear(); }

	/** Get data associate with location site. */
	std::map<std::wstring,std::wstring> GetData() { return m_data; }

	/** 
	 * @brief Get data for specified field. 
	 * @param field Field to get data for. 
	 * @return Data associated with the provided field.
	 */
	const std::wstring& GetData(const std::wstring& field) { return m_data[field]; }

private:
	/** Northing coordinate of location. */
  float m_northing;

	/** Easting coordinate of location. */
  float m_easting;

	/** Name of location site. */
	std::wstring m_name;

	/** Unique id for location site. */
	std::wstring m_id;

	/** Type of environment of location. */
	std::wstring m_environmentType;

	/** All samples taken at this location. */
	std::vector<SequenceModelPtr> m_sequences;

	/** Data associated with location site. */
	std::map<std::wstring,std::wstring> m_data;
};

}

#endif


