//=======================================================================
// Author: Donovan Parks
//
// Copyright 2009 Donovan Parks
//
// This file is part of GenGIS.
//
// GenGIS is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// GenGIS is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with GenGIS.  If not, see <http://www.gnu.org/licenses/>.
//=======================================================================

#include "../core/Precompiled.hpp"

#include "../core/App.hpp"
#include "../core/LocationModel.hpp"
#include "../core/LocationSetIO.hpp"
#include "../core/StudyController.hpp"

#include "../utils/Log.hpp"
#include "../utils/StringTools.hpp"

#include "../gui/ProgressDlg.hpp"

using namespace boost;
using namespace std;
using namespace GenGIS;

bool LocationSetIO::Read(const std::wstring& path, const std::wstring& filename, 
																				StudyControllerPtr studyController, std::vector<LocationModelPtr>& locationModels) 
{
	const int ROWS_TO_SHOW_PROGRESS_DLG = 2000;

	wstring fullPath;
	if(path != _T(""))
		fullPath = path + _T("/") + filename;
	else
		fullPath = filename;

	string temp = util::StringTools::ToString(fullPath);
	wifstream infile(temp.c_str());
	if(!infile.is_open()) 
	{
		error::Log::Inst().Warning("(Warning) Failed to open location data file.");	
		return false;
	}

	// read in column names
	wstring columnHeaders;
	getline(infile, columnHeaders);
	size_t endofdata = columnHeaders.find_last_not_of(L"\r\n\v\t\f");
	if(endofdata != string::npos)
    columnHeaders.erase(endofdata+1);

	// tokenize columns
	tokenizer< escaped_list_separator<wchar_t>, wstring::const_iterator, wstring > columnTokens(columnHeaders);
	tokenizer< escaped_list_separator<wchar_t>, wstring::const_iterator, wstring >::iterator colIter;   

	// parse all rows
	wstring line;
	vector<wstring> rows;
	while(getline(infile, line))
	{
		endofdata = line.find_last_not_of(L"\r\n\v\t\f");
		if(endofdata != string::npos)
      line.erase(endofdata+1);
		rows.push_back(line);
  }
  
	ProgressDlgPtr progressDlg;
	if(rows.size() > ROWS_TO_SHOW_PROGRESS_DLG)
		progressDlg.reset(new ProgressDlg(_T("Loading location set"), _T("Reading locations..."), rows.size(), App::Inst().GetMainWindow()));

	uint rowNum = 0;
	vector<wstring>::const_iterator it;
	for(it = rows.begin(); it != rows.end(); ++it)
	{
		if(progressDlg && rowNum % 25 == 0)
			progressDlg->Update(rowNum);		

		rowNum++;

		tokenizer< escaped_list_separator<wchar_t>, wstring::const_iterator, wstring > dataTokens(*it);
		tokenizer< escaped_list_separator<wchar_t>, wstring::const_iterator, wstring >::iterator dataIter = dataTokens.begin();                

		wstring siteId = _T("");
		wstring siteName = _T("");
		wstring environmentType = _T("");
		double northing = -std::numeric_limits<float>::max();
		double easting = -std::numeric_limits<float>::max();
		bool bIsLatLong = false;
		double latitude = -std::numeric_limits<float>::max();
		double longitude = -std::numeric_limits<float>::max();

		std::map<std::wstring,std::wstring> data;
		for(colIter = columnTokens.begin(); colIter != columnTokens.end(); ++colIter)
		{
			std::wstring columnHeader = *colIter;
			std::wstring columnValue = *dataIter;
			data.insert(pair<wstring, wstring>(columnHeader, columnValue));

			// determine which column we are looking at (this allows user to put columns in any order)
			if(util::StringTools::ToLower(*colIter) == _T("site id"))
			{
				siteId = *dataIter;
			}
			else if(util::StringTools::ToLower(*colIter) == _T("site name"))
			{
				siteName = *dataIter;
			}
			else if(util::StringTools::ToLower(*colIter) == _T("environment type"))
			{
				environmentType = *dataIter;
			}
			else if(util::StringTools::ToLower(*colIter) == _T("northing"))
			{
				northing = util::StringTools::ToLong(*dataIter);
			}
			else if(util::StringTools::ToLower(*colIter) == _T("easting"))
			{
				easting = util::StringTools::ToLong(*dataIter);
			}
			else if(util::StringTools::ToLower(*colIter) == _T("latitude"))
			{
				latitude = util::StringTools::ToDouble(*dataIter);
				bIsLatLong = true;
			}
			else if(util::StringTools::ToLower(*colIter) == _T("longitude"))
			{
				longitude = util::StringTools::ToDouble(*dataIter);
				bIsLatLong = true;
			}

			// move to next column
			dataIter++;
		}

		// transform lat/lon coordinates to projected coordinates
		if(bIsLatLong)
		{
			if(studyController->GetProject())
			{
				ProjectionToolPtr projTool = studyController->GetProjectionTool();
				if(!projTool->Transform(1, &longitude, &latitude))
				{
					error::Log::Inst().Warning("(Warning) Failed to project data. Possibly missing 'Latitude' and 'Longitude' fields.");
					return false;
				}
			}

			northing = latitude;
			easting = longitude;
		}

		// check for required data
		if(siteId == _T(""))
		{
			error::Log::Inst().Warning("(Warning) Missing 'Site Id' field.");
			return false;
		}

		if(northing == -std::numeric_limits<float>::max())
		{
			error::Log::Inst().Warning("(Warning) Missing 'Latitude or Northing' field.");
			return false;
		}

		if(easting == -std::numeric_limits<float>::max())
		{
			error::Log::Inst().Warning("(Warning) Missing 'Longitude or Easting' field.");
			return false;
		}
			              
		// create a new location model from this data
		LocationModelPtr locationModel(new LocationModel(siteName, siteId, northing, easting, 
																												environmentType, data));

		locationModels.push_back(locationModel);
	}

	// closing the file stream.
	infile.close();

	return true;
}

