//=======================================================================
// Author: Donovan Parks
//
// Copyright 2009 Donovan Parks
//
// This file is part of GenGIS.
//
// GenGIS is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// GenGIS is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with GenGIS.  If not, see <http://www.gnu.org/licenses/>.
//=======================================================================

#include "../core/Precompiled.hpp"

#include "../core/LocationSetLayer.hpp"
#include "../core/App.hpp"
#include "../core/TerrainView.hpp"
#include "../core/MapLayer.hpp"
#include "../core/ChartSetView.hpp"
#include "../core/LocationLayer.hpp"
#include "../core/SequenceLayer.hpp"
#include "../core/SequenceController.hpp"
#include "../core/LocationSetController.hpp"

#include "../utils/ColourMapManager.hpp"
#include "../utils/ColourMap.hpp"
#include "../utils/ColourMapDiscrete.hpp"
#include "../utils/ShapeMapManager.hpp"
#include "../utils/ShapeMap.hpp"
#include "../utils/StringTools.hpp"

using namespace GenGIS;

LocationSetLayer::LocationSetLayer(uint id, LayerPtr parent, ChartSetViewPtr chartSetView)
		: Layer(id, Layer::LOCATION_SET, parent), m_chartSetView(chartSetView),
			m_locationSetController(new LocationSetController())
{

}

void LocationSetLayer::AddLocation(LocationLayerPtr locationLayer) 
{ 
	m_locationLayers.push_back(locationLayer);
}

LocationLayerPtr LocationSetLayer::GetLocation(const std::wstring& name) const
{
	std::vector<LocationLayerPtr>::const_iterator iter;
	for(iter = m_locationLayers.begin(); iter != m_locationLayers.end(); iter++)
	{
		if((*iter)->GetName() == name)
		{
			return *iter;
		}
	}

	return LocationLayerPtr();
}

void LocationSetLayer::RemoveLocation(unsigned int index)
{
	m_locationLayers.at(index)->RemoveAllSequences();
	m_locationLayers.erase(m_locationLayers.begin() + index);
}

void LocationSetLayer::RemoveLocation(const std::wstring& name)
{
	std::vector<LocationLayerPtr>::iterator iter;
	for(iter = m_locationLayers.begin(); iter != m_locationLayers.end(); iter++)
	{
		if((*iter)->GetName() == name)
		{
			(*iter)->RemoveAllSequences();
			m_locationLayers.erase(iter);
			break;
		}
	}
}

LayerPtr LocationSetLayer::GetLayer(uint id) const
{
	for(unsigned int index = 0; index < GetNumLocations(); index++)
	{
		LocationLayerPtr layer = GetLocation(index);
		if(layer->GetId() == id)
			return layer;

		LayerPtr returnLayer = layer->GetLayer(id);
		if(returnLayer != LayerPtr())
		{
			return returnLayer;
		}
	}

	return LayerPtr();
}

bool LocationSetLayer::RemoveLayer(uint id)
{
	bool bFound = false;
	for(unsigned int index = 0; index < GetNumLocations() && !bFound; index++)
	{
		LocationLayerPtr layer = GetLocation(index);
		if(layer->GetId() == id)
		{
			RemoveLocation(index);
			return true;
		}

		bFound = layer->RemoveLayer(id);
	}

	return bFound;
}

void LocationSetLayer::Render()
{
	if(!IsVisible())
		return;

	m_chartSetView->Render();

	foreach(LocationLayerPtr locationLayer, m_locationLayers)
	{
		locationLayer->Render();
	}
}

bool LocationSetLayer::IsSequencesData()
{
	foreach(LocationLayerPtr locationLayer, m_locationLayers)
	{
		if(locationLayer->GetNumSequences() > 0)
		{
			return true;
		}
	}

	return false;
}
