//=======================================================================
// Author: Donovan Parks
//
// Copyright 2009 Donovan Parks
//
// This file is part of GenGIS.
//
// GenGIS is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// GenGIS is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with GenGIS.  If not, see <http://www.gnu.org/licenses/>.
//=======================================================================

#ifndef _GENGIS_LOCATION_SET_LAYER_
#define _GENGIS_LOCATION_SET_LAYER_

#include "../core/Precompiled.hpp"

#include "../core/Layer.hpp"
#include "../core/LocationController.hpp"

#include "../utils/ShapeMap.hpp"

namespace GenGIS
{

/**
 * @brief A location set consists of 1 or more locations. 
 */
class LocationSetLayer : public Layer
{
public:
	/** 
	 * @brief Constuctor.
	 * @param id Unique identifier used to specify location set.
	 * @param parent Parent of this layer.
	 * @param chartSetView Chart set view for laying out chart views associated with this location set.
	 */
	LocationSetLayer(uint id, LayerPtr parent, ChartSetViewPtr chartSetView);

	/** Destructor. */
	virtual ~LocationSetLayer() {}

	/**
	 * @brief Get layer from tree below this layer. 
	 * @param id Unique id of desired layer.
	 * @return layer at given id.
	 */
	virtual LayerPtr GetLayer(uint id) const;

	/**
	 * @brief Remove layer from tree below this layer. 
	 * @param id Unique id of layer to remove.
	 * @return True if layer removed, false if layer not found.
	 */
	virtual bool RemoveLayer(uint id);

	/** Render contents of layer. */
	virtual void Render();

	/** Add location to location set. */
	void AddLocation(LocationLayerPtr location);
	
	/** Get number of locations contained in location set. */
	unsigned int GetNumLocations() const { return m_locationLayers.size(); }

	/**
	 * @brief Get location from location set.
	 * @param index Index of desired location.
	 * @return Location at given index.
	 */
	LocationLayerPtr GetLocation(unsigned int index) const { return m_locationLayers.at(index); }

	/**
	 * @brief Get location from location set.
	 * @param name Unique name of desired location.
	 * @return Location with given name.
	 */
	LocationLayerPtr GetLocation(const std::wstring& name) const;

	/**
	 * @brief Remove location from location set.
	 * @param index Location to remove.
	 */
	void RemoveLocation(unsigned int index);

	/**
	 * @brief Remove location set from study.
	 * @param name Name of location.
	 */
	void RemoveLocation(const std::wstring& name);

	/** Get chart set controller. */
	ChartSetViewPtr GetChartSetView() const { return m_chartSetView; }

	/** Get location set controller. */
	LocationSetControllerPtr GetLocationSetController() const { return m_locationSetController; }
  
	/** Get vector of all location layers. */
  const std::vector<LocationLayerPtr> GetAllLocations() { return m_locationLayers; }
	
	/** Check if there is sequence data associate with at least one location within this location set. */
	bool IsSequencesData();

protected:
	/** Location layers beneath this location set layer. */
	std::vector<LocationLayerPtr> m_locationLayers;

	/** Chart controller for laying out chart views associated with this location set. */
	ChartSetViewPtr m_chartSetView;

	/** Location set controller for setting visual properties to entire set of locations. */
	LocationSetControllerPtr m_locationSetController;
};

}

#endif
