//=======================================================================
// Author: Donovan Parks
//
// Copyright 2009 Donovan Parks
//
// This file is part of GenGIS.
//
// GenGIS is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// GenGIS is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with GenGIS.  If not, see <http://www.gnu.org/licenses/>.
//=======================================================================

#include "../core/Precompiled.hpp"

#include "../core/MapLayer.hpp"
#include "../core/TreeLayer.hpp"
#include "../core/LocationSetLayer.hpp"
#include "../core/TerrainController.hpp"
#include "../core/TerrainView.hpp"

using namespace GenGIS;

LocationSetLayerPtr MapLayer::GetLocationSet(const std::wstring& name) const
{
	std::vector<LocationSetLayerPtr>::const_iterator iter;
	for(iter = m_locationSets.begin(); iter != m_locationSets.end(); iter++)
	{
		if((*iter)->GetName() == name)
		{
			return *iter;
		}
	}

	return LocationSetLayerPtr();
}

void MapLayer::RemoveLocationSet(unsigned int index)
{
	// erase study and all children 
	for(unsigned int i = 0; i < m_locationSets.at(index)->GetNumLocations(); ++i)
	{
		m_locationSets.at(index)->RemoveLocation(0);	// keep removing the first element
	}

	m_locationSets.erase(m_locationSets.begin() + index);
}

void MapLayer::RemoveLocationSet(const std::wstring& name)
{
	std::vector<LocationSetLayerPtr>::iterator iter;
	for(iter = m_locationSets.begin(); iter != m_locationSets.end(); iter++)
	{
		if((*iter)->GetName() == name)
		{
			// erase study and all children 
			for(unsigned int i = 0; i < (*iter)->GetNumLocations(); ++i)
			{
				(*iter)->RemoveLocation(0);	// keep removing the first element
			}

			m_locationSets.erase(iter);
			break;
		}
	}
}

void MapLayer::RemoveTree(unsigned int index)
{
	m_trees.erase(m_trees.begin() + index);
}

void MapLayer::RemoveTree(const std::wstring& name)
{
	std::vector<TreeLayerPtr>::iterator iter;
	for(iter = m_trees.begin(); iter != m_trees.end(); iter++)
	{
		if((*iter)->GetName() == name)
		{
			m_trees.erase(iter);
			break;
		}
	}
}

LayerPtr MapLayer::GetLayer(uint id) const
{
	for(unsigned int index = 0; index < GetNumLocationSets(); index++)
	{
		LocationSetLayerPtr layer = GetLocationSet(index);
		if(layer->GetId() == id)
			return layer;

		LayerPtr returnLayer = layer->GetLayer(id);
		if(returnLayer != LayerPtr())
		{
			return returnLayer;
		}
	}

	for(unsigned int index = 0; index < GetNumTrees(); index++)
	{
		TreeLayerPtr layer = GetTree(index);
		if(layer->GetId() == id)
			return layer;

		LayerPtr returnLayer = layer->GetLayer(id);
		if(returnLayer != LayerPtr())
		{
			return returnLayer;
		}
	}

	return LayerPtr();
}

bool MapLayer::RemoveLayer(uint id)
{
	bool bFound = false;

	// Check if layer is a location set:
	for(unsigned int index = 0; index < GetNumLocationSets() && !bFound; index++)
	{
		LocationSetLayerPtr layer = GetLocationSet(index);
		if(layer->GetId() == id)
		{
			RemoveLocationSet(index);
			return true;
		}

		bFound = layer->RemoveLayer(id);
	}

	// Check if layer is a tree:
	for(unsigned int index = 0; index < GetNumTrees() && !bFound; index++)
	{
		TreeLayerPtr layer = GetTree(index);
		if(layer->GetId() == id)
		{
			RemoveTree(index);
			return true;
		}

		bFound = layer->RemoveLayer(id);
	}

	return bFound;
}

void MapLayer::Render()
{
	if(!IsVisible())
		return;
	
	m_terrainController->GetTerrainView()->Render();

	/*
	// Note: Please leave this code in as it is extremely useful for debugging projection problems.
	// Render a simple point-based map just to verify the actual rendering is correct
	if(GetNumLocationSets() > 0)
	{
		glDisable(GL_LIGHTING);
		
		GLUquadricObj* m_markerQuadric = gluNewQuadric();
		gluQuadricDrawStyle( m_markerQuadric, GLU_FILL );
		gluQuadricNormals( m_markerQuadric, GLU_SMOOTH );

		Point3DArrayPtr grid = m_terrainController->GetGrid();
		for ( int m = 0; m < m_terrainController->GetHeader()->nRows ; m += 25 ) 
		{
			for ( int n = 0; n < m_terrainController->GetHeader()->nCols; n += 25 ) 
			{
				glPushMatrix();
				int index = n + m*m_terrainController->GetHeader()->nCols;
				glTranslated(grid[index].x, grid[index].y, grid[index].z);

				if(grid[index].y > 0.0)
				{
					glColor3f(1.0f,0.1f,0.1f);
				}
				else
				{
					glColor3f(0.1f,0.1f,1.0f);
				}
				gluSphere(m_markerQuadric, 0.01, 10, 10);
				glPopMatrix();
			}
		}

		glEnable(GL_LIGHTING);
	}
	*/

	for(unsigned int index = 0; index < GetNumLocationSets(); index++)
	{
			unsigned int num = m_locationSets.size();
			LocationSetLayerPtr layer = m_locationSets.at(0);

		GetLocationSet(index)->Render();
	}

	for(unsigned int index = 0; index < GetNumTrees(); index++)
	{
		GetTree(index)->Render();
	}
}
