//=======================================================================
// Author: Donovan Parks
//
// Copyright 2009 Donovan Parks
//
// This file is part of GenGIS.
//
// GenGIS is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// GenGIS is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with GenGIS.  If not, see <http://www.gnu.org/licenses/>.
//=======================================================================

#ifndef _GENGIS_MAP_LAYER_
#define _GENGIS_MAP_LAYER_

#include "../core/Precompiled.hpp"

#include "../core/Layer.hpp"

namespace GenGIS
{

/**
 * @brief A map consists of a digital elevation model (DEM) and 0 or more location sets. 
 */
class MapLayer : public Layer
{
public:
	/** Allows convenient syntax for creating smart pointers to this class */
	typedef boost::shared_ptr<MapLayer> Ptr;	

public:
	/** 
	 * @brief Constuctor.
	 * @param id Unique identifier used to specify map.
	 * @param parent Parent of this layer.
	 * @param terrainController Controller for terrain model associated with this map.
	 */
	MapLayer(uint id, LayerPtr parent, TerrainControllerPtr terrainController)
		: Layer(id, Layer::MAP, parent), m_terrainController(terrainController) {}

	/** Destructor. */
	virtual ~MapLayer() {}

	/**
	 * @brief Get layer from tree below this layer. 
	 * @param id Unique id of desired layer.
	 * @return layer at given id.
	 */
	virtual LayerPtr GetLayer(uint id) const;

	/**
	 * @brief Remove layer from tree below this layer. 
	 * @param id Unique id of layer to remove.
	 * @return True if layer removed, false if layer not found.
	 */
	virtual bool RemoveLayer(uint id);

	/** Render contents of layer. */
	virtual void Render();

	/** Get terrain controller associated with this map. */
	TerrainControllerPtr GetTerrainController() { return m_terrainController; }

	/** 
	 * @brief Add location set to map. 
	 * @param locationSite Location site to add to map.
	 */
	void AddLocationSet(LocationSetLayerPtr locationSet) { m_locationSets.push_back(locationSet); }
	
	/** Get number of location set contained in map. */
	unsigned int GetNumLocationSets() const { return m_locationSets.size(); }

	/**
	 * @brief Get location set from map.
	 * @param index Index of desired location set.
	 * @return Location set at given index.
	 */
	LocationSetLayerPtr GetLocationSet(unsigned int index) const { return m_locationSets.at(index); }

	/**
	 * @brief Get location set from map.
	 * @param name Name of desired location set.
	 * @return Location set with given name.
	 */
	LocationSetLayerPtr GetLocationSet(const std::wstring& name) const;

	/**
	 * @brief Remove location set from map.
	 * @param index Location set to remove.
	 */
	void RemoveLocationSet(unsigned int index);

	/**
	 * @brief Remove location set from map.
	 * @param name Name of location set.
	 */
	void RemoveLocationSet(const std::wstring& name);

	/** 
	 * @brief Add tree to map. 
	 * @param tree Tree to add to map.
	 */
	void AddTree(TreeLayerPtr tree) { m_trees.push_back(tree); }
	
	/** Get number of trees contained in map. */
	unsigned int GetNumTrees() const { return m_trees.size(); }

	/**
	 * @brief Get tree from map.
	 * @param index Index of desired tree.
	 * @return Tree at given index.
	 */
	TreeLayerPtr GetTree(unsigned int index) const { return m_trees.at(index); }

	/**
	 * @brief Get tree from map.
	 * @param name Name of desired tree.
	 * @return Tree with given name.
	 */
	TreeLayerPtr GetTree(const std::wstring& name) const;

	/**
	 * @brief Remove tree from map.
	 * @param index Tree to remove.
	 */
	void RemoveTree(unsigned int index);

	/**
	 * @brief Remove tree from map.
	 * @param name Name of tree to remove.
	 */
	void RemoveTree(const std::wstring& name);
  
	/** Get all location set layers. */
  const std::vector<LocationSetLayerPtr> GetAllLocationSets() { return m_locationSets; }

	/** Get all tree layers. */
	const std::vector<TreeLayerPtr> GetAllTrees() { return m_trees; }
  
protected:
	/** Terrain model for this map. */
	TerrainControllerPtr m_terrainController;

	/** Location sets associated with this map. */
	std::vector<LocationSetLayerPtr> m_locationSets;

	/** Tree layers associated with this map. */
	std::vector<TreeLayerPtr> m_trees;
};

}

#endif
