//=======================================================================
// Author: Donovan Parks
//
// Copyright 2009 Donovan Parks
//
// This file is part of GenGIS.
//
// GenGIS is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// GenGIS is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with GenGIS.  If not, see <http://www.gnu.org/licenses/>.
//=======================================================================

#include "../core/Precompiled.hpp"

#include "../core/SequenceIO.hpp"

#include "../core/App.hpp"
#include "../utils/Log.hpp"
#include "../utils/StringTools.hpp"

#include "../gui/ProgressDlg.hpp"

using namespace std;
using namespace GenGIS;

bool SequenceIO::Read(const std::wstring& path, const std::wstring& filename, std::vector<SequenceModelPtr>& sequenceModels) 
{
	wstring fullPath;
	if(path != _T(""))
		fullPath = path + _T("/") + filename;
	else
		fullPath = filename;

	string temp = util::StringTools::ToString(fullPath);
	wifstream infile(temp.c_str());
	if(!infile.is_open()) 
	{
		error::Log::Inst().Warning("(Warning) Failed to open sequence data file.");	
		return false;
	}

	// count number of lines in file so we can track loading progress
	int numLines = 0;
	wstring line;
	while( getline(infile, line) )
	{
			numLines++;
	}
	infile.clear(); 
	infile.seekg(0);

	// track progress
	ProgressDlg progressDlg(wxT("Loading file"), wxT("Loading sequences..."), 100, App::Inst().GetMainWindow());
	
	// read in column names
	wstring columnHeaders;
	getline(infile, columnHeaders);

	// tokenize columns
	boost::tokenizer< boost::escaped_list_separator<wchar_t>, wstring::const_iterator, wstring  > columnTokens(columnHeaders);
	     
	// parse all rows
	wstring data;
	int progress = 0;
	while(getline(infile, data)) 
	{
		if(progress % 100 == 0)
			progressDlg.Update(float(progress)/numLines * 100);

		progress++;

		boost::tokenizer< boost::escaped_list_separator<wchar_t>, wstring::const_iterator, wstring  > dataTokens(data);
		boost::tokenizer< boost::escaped_list_separator<wchar_t>, wstring::const_iterator, wstring  >::iterator dataIter = dataTokens.begin();                

		wstring siteId = _T("");
		wstring sequenceId = _T("");

		boost::tokenizer< boost::escaped_list_separator<wchar_t>, wstring::const_iterator, wstring >::iterator colIter;   

		std::map<std::wstring,std::wstring> data;
		for(colIter = columnTokens.begin(); colIter != columnTokens.end(); ++colIter)
		{
			std::wstring columnHeader = *colIter;
			std::wstring columnValue = *dataIter;
			data.insert(pair<std::wstring, std::wstring>(columnHeader, columnValue));

			// determine which column we are looking at (this allows user to put columns in any order)
			if(util::StringTools::ToLower(*colIter) == _T("site id"))
			{
				siteId = *dataIter;
			}
			else if(util::StringTools::ToLower(*colIter) == _T("sequence id"))
			{
				sequenceId = *dataIter;
			}
			
			// move to next column
			dataIter++;
		}

		// check for required data
		if(siteId == _T(""))
		{
			error::Log::Inst().Warning("(Warning) Missing 'Site Id' field.");
			return false;
		}

		if(sequenceId == _T(""))
		{
			error::Log::Inst().Warning("(Warning) Missing 'Sequence Id' field.");
			return false;
		}
		              
		// create a new location model from this data
		SequenceModelPtr sequenceModel(new SequenceModel(sequenceId, siteId, data));

		sequenceModels.push_back(sequenceModel);
	}

	// closing the file stream.
	infile.close();

	return true;
}

