//=======================================================================
// Author: Donovan Parks
//
// Copyright 2009 Donovan Parks
//
// This file is part of GenGIS.
//
// GenGIS is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// GenGIS is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with GenGIS.  If not, see <http://www.gnu.org/licenses/>.
//=======================================================================

#ifndef _GENGIS_SAMPLE_MODEL_
#define _GENGIS_SAMPLE_MODEL_

#include "../core/Precompiled.hpp"

#include "../utils/Log.hpp"
#include "../utils/StringTools.hpp"

namespace GenGIS 
{

/**
 * @brief Sequence data.
 */
class SequenceModel
{
public:
	/** 
	 * @brief Constuctor.
	 * @param sequenceId Unique id for sequence.
	 * @param siteId Unique id of location site associated with sequence.
	 * @param data Data associated with this location.
	 */
	SequenceModel(const std::wstring& sequenceId, const std::wstring& siteId, std::map<std::wstring,std::wstring> data)
		: m_sequenceId(sequenceId), m_siteId(siteId), m_bActive(true), m_data(data) {} 

	/** Destructor. */
	virtual ~SequenceModel() {}

	/** Get id of sequence. */
	const std::wstring GetSequenceId() { return m_sequenceId; }

	/** Get id of location site associated with sequence. */
	const std::wstring GetSiteId() { return m_siteId; }

	/** Get flag indicating if sequence is in active set. */
	bool IsActive() { return m_bActive; }

	/** Set flag indicating if sequence is in active set. */
	void SetActive(bool state) { m_bActive = state; }

	/** Get data associate with sequence. */
	std::map<std::wstring,std::wstring> GetData() { return m_data; }

	/** 
	 * @brief Get data for specified field. 
	 * @param field Field to get data for. 
	 * @return Data associated with the provided field.
	 */
	const std::wstring& GetData(const std::wstring& field) 
	{ 
		std::map<std::wstring,std::wstring>::const_iterator it;
		it = m_data.find(field);
		if(it != m_data.end())
		{
			return it->second;
		}
		else
		{			
			error::Log::Inst().Error("(Error) Failed to find data with a field value = " + util::StringTools::ToString(field));
			return m_siteId;
		}
	}

private:
	/** Unique id for sequence. */
	std::wstring m_sequenceId;

	/** Unique id for location site associated with sequence. */
	std::wstring m_siteId;

	/** Flag indicating if this sequence is currently in the active set. */
	bool m_bActive;

	/** Data associated with sequence. */
	std::map<std::wstring,std::wstring> m_data;
};

}

#endif
