//=======================================================================
// Author: Donovan Parks
//
// Copyright 2009 Donovan Parks
//
// This file is part of GenGIS.
//
// GenGIS is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// GenGIS is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with GenGIS.  If not, see <http://www.gnu.org/licenses/>.
//=======================================================================

#ifndef _GENGIS_STUDY_LAYER_
#define _GENGIS_STUDY_LAYER_

#include "../core/Precompiled.hpp"

#include "../core/Layer.hpp"

namespace GenGIS 
{

/**
 * @brief A study consists of 1 or more maps. It is the highest level
 *				organizational concept. In a traditional GIS settings, a study 
 *				would be termed a project.
 */
class StudyLayer : public Layer
{
public:
	/** Allows convenient syntax for creating smart pointers to this class */
	typedef boost::shared_ptr<StudyLayer> Ptr;	

public:
	/** 
	 * @brief Constuctor.
	 * @param id Unique identifier used to specify study.
	 * @param parent Parent of this layer.
	 */
	StudyLayer(uint id, LayerPtr parent, StudyControllerPtr studyController)
		: Layer(id, Layer::STUDY, parent), m_studyController(studyController) {}

	/** Destructor. */
	virtual ~StudyLayer() {}

	/**
	 * @brief Get layer from tree below this layer. 
	 * @param id Unique id of desired layer.
	 * @return layer at given id.
	 */
	virtual LayerPtr GetLayer(uint id) const;

	/**
	 * @brief Remove layer from tree below this layer. 
	 * @param id Unique id of layer to remove.
	 * @return True if layer removed, false if layer not found.
	 */
	virtual bool RemoveLayer(uint id);

	/** Render contents of layer. */
	virtual void Render();

	/** 
	 * @brief Add map to study. 
	 * @param map Map to add to study.
	 */
	void AddMap(MapLayerPtr map) { m_maps.push_back(map); }
	
	/** Get number of maps contained in study. */
	unsigned int GetNumMaps() const { return m_maps.size(); }

	/**
	 * @brief Get map from study.
	 * @param index Index of desired map.
	 * @return Map at given index.
	 */
	MapLayerPtr GetMap(unsigned int index) const { return m_maps.at(index); }

	/**
	 * @brief Get map from study.
	 * @param name Name of desired map.
	 * @return Map with given name.
	 */
	MapLayerPtr GetMap(const std::wstring& name) const;

	/**
	 * @brief Remove map from study.
	 * @param index Map to remove.
	 */
	void RemoveMap(unsigned int index);

	/**
	 * @brief Remove map from study.
	 * @param name Name of map.
	 */
	void RemoveMap(const std::wstring& name);

	/** Get study controller associated with this layer. */
	StudyControllerPtr GetStudyController() { return m_studyController; }
  
  const std::vector<MapLayerPtr> GetAllMaps() { return m_maps; }
  
protected:
	/** Maps associated with this study. */
	std::vector<MapLayerPtr> m_maps;

	/** Controller for study model. */
	StudyControllerPtr m_studyController;
};

}

#endif
