//=======================================================================
// Author: Donovan Parks
//
// Copyright 2009 Donovan Parks
//
// This file is part of GenGIS.
//
// GenGIS is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// GenGIS is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with GenGIS.  If not, see <http://www.gnu.org/licenses/>.
//=======================================================================

#include "../core/Precompiled.hpp"

#include "../core/StudyModel.hpp"
#include "../utils/StringTools.hpp"
#include "../utils/Log.hpp"

using namespace GenGIS;

ProjectionToolPtr StudyModel::GetProjectionTool(OGRSpatialReference* oSource)
{
	if(m_bProject)
	{
		SetDestinationProjection();

		OGRCoordinateTransformation *poTransform;

		if(oSource != NULL)
		{
			poTransform = OGRCreateCoordinateTransformation(oSource, &m_oDestination);
		}
		else
		{
			// assume the source is in lat/lon with the same datum as specified for the destination
			OGRSpatialReference* poLatLong = m_oDestination.CloneGeogCS();
			poTransform = OGRCreateCoordinateTransformation(poLatLong, &m_oDestination);
		}

		if(poTransform == NULL)
		{
			error::Log::Inst().Warning("(Warning) Failed to create projection tool.");
			return ProjectionToolPtr();
		}

		return ProjectionToolPtr(poTransform);
	}

	return ProjectionToolPtr(); 
}

void StudyModel::SetDestinationProjection()
{
	std::string projInfoStr = "";

	// set datum
	if(m_datum == _T("WGS84"))
	{
		projInfoStr = "WGS84";
		m_oDestination.SetWellKnownGeogCS("WGS84");
	}
	else if(m_datum == _T("WGS72"))
	{
		projInfoStr = "WGS72";
		m_oDestination.SetWellKnownGeogCS("WGS72");
	}
	else if(m_datum == _T("NAD83"))
	{
		projInfoStr = "NAD83";
		m_oDestination.SetWellKnownGeogCS("NAD83");
	}
	else if(m_datum == _T("NAD27"))
	{
		projInfoStr = "NAD27";
		m_oDestination.SetWellKnownGeogCS("NAD27");
	}

	// set projection
	if(m_projection == _T("Albers Conic Equal Area"))
	{
		projInfoStr += " / Albers Conic Equal Area";
		m_oDestination.SetACEA(m_standardParallel1, m_standardParallel2, m_centreLat, m_centreLong, 0, 0); 	
	}
	else if(m_projection == _T("Azimuthal Equidistant"))
	{
		projInfoStr += " / Azimuthal Equidistant";
		m_oDestination.SetAE(m_centreLat, m_centreLong, 0, 0);
	}
	else if(m_projection == _T("Eckert IV"))
	{
		projInfoStr += " / Eckert IV";
		m_oDestination.SetEckertIV(m_centreLong, 0, 0);
	}
	else if(m_projection == _T("Equidistant Conic"))
	{
		projInfoStr += " / Equidistant Conic";
		m_oDestination.SetEC(m_standardParallel1, m_standardParallel2, m_centreLat, m_centreLong, 0, 0);
	}
	else if(m_projection == _T("Gall Stereograpic"))
	{
		projInfoStr += " / Gall Stereograpic";
		m_oDestination.SetGS(m_centreLong, 0, 0); 
	}
	else if(m_projection == _T("Lambert Azimuthal Equal-Area"))
	{
		projInfoStr += " / Lambert Azimuthal Equal-Area";
		m_oDestination.SetLAEA(m_centreLat, m_centreLong, 0, 0); 	
	}
	else if(m_projection == _T("Lambert Conformal Conic"))
	{
		projInfoStr += " / Lambert Conformal Conic";
		m_oDestination.SetLCC(m_standardParallel1, m_standardParallel2, m_centreLat, m_centreLong, 0, 0);
	}
	else if(m_projection == _T("Mercator"))
	{
		projInfoStr += " / Mercator";
		m_oDestination.SetMercator(m_centreLat, m_centreLong, 1, 0, 0);
	}
	else if(m_projection == _T("Mollweide"))
	{
		projInfoStr += " / Mollweide";
		m_oDestination.SetMollweide(m_centreLong, 0, 0);
	}
	else if(m_projection == _T("Orthographic"))
	{
		projInfoStr += " / Orthographic";
		m_oDestination.SetOrthographic(m_centreLat, m_centreLong, 0, 0); 
	}
	else if(m_projection == _T("Plate Carree (Equirectangular)"))
	{
		projInfoStr += " / Plate Carree (Equirectangular)";
		m_oDestination.SetEquirectangular(m_centreLat, m_centreLong, 0, 0);
	}
	else if(m_projection == _T("Robinson"))
	{
		projInfoStr += " / Robinson";
		m_oDestination.SetRobinson(m_centreLong, 0, 0);
	}
	else if(m_projection == _T("Stereographic"))
	{
		projInfoStr += " / Stereographic";
		m_oDestination.SetStereographic(m_centreLat, m_centreLong, 1, 0, 0); 
	}
	else if(m_projection == _T("Transverse Mercator"))
	{
		projInfoStr += " / Transverse Mercator";
		m_oDestination.SetTM(m_centreLat, m_centreLong, 1, 0, 0);
	}
	else if(m_projection == _T("Universal Transverse Mercator (UTM)"))
	{
		int utmZone = int(m_centreLong/6.0f) + 31;
		projInfoStr += " / UTM Zone " + util::StringTools::ToString(utmZone);
		m_oDestination.SetUTM(utmZone, m_centreLat > 0);
	}

	m_oDestination.SetProjCS(projInfoStr.c_str());
}

void StudyModel::SetCentreLatitude(float centre)
{
	m_centreLat = centre;
}


void StudyModel::SetCentreLongitude(float centre)
{
	m_centreLong = centre;
}

void StudyModel::SetFirstStandardParallel(float sp)
{
	m_standardParallel1 = sp;
}

void StudyModel::SetSecondStandardParallel(float sp)
{
	m_standardParallel2 = sp;
}
