//=======================================================================
// Author: Donovan Parks
//
// Copyright 2009 Donovan Parks
//
// This file is part of GenGIS.
//
// GenGIS is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// GenGIS is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with GenGIS.  If not, see <http://www.gnu.org/licenses/>.
//=======================================================================

#ifndef _GENGIS_STUDY_MODEL_
#define _GENGIS_STUDY_MODEL_

#include "../core/Precompiled.hpp"

namespace GenGIS
{

/**
 * @brief Study specific properties.
 */
class StudyModel 
{
public:
	/** Construction. */
	StudyModel(const std::wstring& datum = _T("WGS84"), const std::wstring& projection = _T("Plate Carree (Equirectangular)"), 
								uint utmZone = 30, const std::wstring& utmHemisphere = _T("North"))
		: m_bProject(true), m_datum(datum), m_projection(projection), m_centreLat(INT_MIN),	m_centreLong(INT_MIN),
		m_standardParallel1(INT_MIN), m_standardParallel2(INT_MIN) {}

	/** Destructor. */
	~StudyModel() {}

	/**
	 * @brief Get projection tool.
	 * @param oSource Coordinate system of source. If NULL, assume lat/lon with same datum as projected coordinate system.
	 * @return Projection tool.
	 */
	ProjectionToolPtr GetProjectionTool(OGRSpatialReference* oSource = NULL);

	/** Get specified destination coordinate system. */
	OGRSpatialReference GetDestinationProj() { SetDestinationProjection(); return m_oDestination; }

	/** Set flag indicating if data should be projected. */
	void SetProject(bool state) { m_bProject = state; }

	/** Get flag indicating if data should be projected. */
	bool GetProject() { return m_bProject; }


	/** Set datum. */
	void SetDatum(const std::wstring& datum) { m_datum = datum; }

	/** Get datum. */
	const std::wstring& GetDatum() const { return m_datum; }

	/** Set projection. */
	void SetProjection(const std::wstring& projection) { m_projection = projection; }

	/** Get projection. */
	const std::wstring& GetProjection() const { return m_projection; }

	/** Set centre latitude. */
	void SetCentreLatitude(float centre);

	/** Get centre Longitude. */
	float GetCentreLongitude() { return m_centreLat; }

	/** Set centre Longitude. */
	void SetCentreLongitude(float centre);

	/** Get centre latitude. */
	float GetCentreLatitude() { return m_centreLong; }

	/** Set first standard parallel. */
	void SetFirstStandardParallel(float sp);

	/** Get first standard parallel. */
	float GetFirstStandardParallel() { return m_standardParallel1; }

	/** Set second standard parallel. */
	void SetSecondStandardParallel(float sp);

	/** Get second standard parallel. */
	float GetSecondStandardParallel() { return m_standardParallel2; }

protected:
	/** Set desired projection. */
	void SetDestinationProjection();

protected:
	/** Desired geographic coordinate system. */
	OGRSpatialReference m_oDestination;

	/** Flag indicating if data should be projected. */
	bool m_bProject;

	/** Currently active datum. */
	std::wstring m_datum;

	/** Currently active projection. */
	std::wstring m_projection;

	/** Latitude at centre of map. */
	float m_centreLat;
	
	/** Longitude at centre of map. */
	float m_centreLong;

	/** First standard parallel to use for map projection. */
	float m_standardParallel1;
	
	/** Second standard parallel to use for map projection. */
	float m_standardParallel2;
};

}

#endif
