//=======================================================================
// Author: Donovan Parks
//
// Copyright 2009 Donovan Parks
//
// This file is part of GenGIS.
//
// GenGIS is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// GenGIS is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with GenGIS.  If not, see <http://www.gnu.org/licenses/>.
//=======================================================================

#ifndef _GENGIS_TERRAIN_CONTROLLER_
#define _GENGIS_TERRAIN_CONTROLLER_

#include "../core/Precompiled.hpp"
#include "../core/TerrainModel.hpp"

namespace GenGIS
{

	/**
 * @brief Controller for terrain model.
 */
class TerrainController
{
public:
	/** 
	 * @brief Construction. 
	 * @param terrainModel Terrain model to associate with controller.
	 * @param terrainView OpenGL (3D) view of terrain model.
	 */
	TerrainController(TerrainModelPtr terrainModel, TerrainViewPtr terrainView)
		: m_terrainModel(terrainModel), m_terrainView(terrainView) {}

	/** Destructor. */
	~TerrainController() {}

	/**
	 * @brief Get array of grid coordinates.
	 * @return Grid coordinate array.
	 */
	Point3DArrayPtr GetGrid()  { return m_terrainModel->GetGrid(); }

	/**
	 * @brief Convert geographic coordinates to grid coordinates.
	 * @param mapCoord Geographic coordinates.
	 * @param gridCoord Grid coordinate.
	 */
	void GeoToGrid(const GeoCoord &mapCoord, Point3D &gridCoord)  { return m_terrainModel->GeoToGrid(mapCoord, gridCoord); }

	/**
	 * @brief Convert lat/long coordinates to grid coordinates.
	 * @param mapCoord Lat/long coordinates.
	 * @param gridCoord Grid coordinate.
	 */
	void LatLongToGrid(const GeoCoord &mapCoord, Point3D &gridCoord) { return m_terrainModel->LatLongToGrid(mapCoord, gridCoord); }

	/**
	 * @brief Convert grid coordinates to geographic coordinates.
	 * @param gridCoord Grid coordinate.
	 * @param mapCoord Geographic coordinates.
	 */
	void GridToGeo(const Point3D &gridCoord, GeoCoord &mapCoord)  { return m_terrainModel->GridToGeo(gridCoord, mapCoord); }

	/**
	 * @brief Get DEM file header information.
	 * @return DEM header information.
	 */
	FileHeader* GetHeader() { return m_terrainModel->GetHeader(); }

	/**
	 * @brief Get DEM metadata information.
	 * @return DEM metadata information.
	 */
	FileMetaData* GetMetaData() { return m_terrainModel->GetMetaData(); }

	/**
	 * @brief Get transformation tool.
	 * @return Transformation tool.
	 */
	OGRCoordinateTransformation* GetTransformationTool() { return m_terrainModel->GetTransformationTool(); }

	/**
	 * @brief Width (x-dimension) of DEM map. 
	 * @return Width.
	 */
	float GetWidth() { return m_terrainModel->GetWidth(); }

	/**
	 * @brief Height (z-dimension) of DEM map. 
	 * @return Height.
	 */
	float GetHeight() { return m_terrainModel->GetHeight(); }

	/**
	 * @brief Elevation (y-dimension) of DEM map at specified point.
	 * @return Elevation.
	 */
	float GetElevation(float gridX, float gridZ) { return m_terrainModel->GetElevation(gridX, gridZ); }

	/**	 
	* @brief Elevation (y-dimension) of DEM map at specified point multiplied by the vertical exaggeration.	 
	* @return Exaggerated elevation.	 
	*/	
	float GetExaggeratedElevation(float gridX, float gridZ) { return m_terrainModel->GetExaggeratedElevation(gridX, gridZ); }	
	
	/** Set flag indicating if terrain model contains elevation information. */	
	void SetElevationMap(bool bElevationMap) { m_terrainModel->SetElevationMap(bElevationMap); }	
	
	/** Get flag indicating if terrain model contains elevation information. */	
	bool GetElevationMap() const { return m_terrainModel->GetElevationMap(); }	
	
	/** Get OpenGL (3D) terrain view. */
	TerrainViewPtr GetTerrainView() { return m_terrainView; }

	/** Set minimum elevation (in grid space). */
	void SetMinElevationGridSpace(float min) { m_terrainModel->SetMinElevationGridSpace(min); }

	/** Get minimum elevation (in grid space). */
	float GetMinElevationGridSpace() const { return m_terrainModel->GetMinElevationGridSpace(); }

	/** Set maximum elevation (in grid space). */
	void SetMaxElevationGridSpace(float max) { m_terrainModel->SetMaxElevationGridSpace(max); }

	/** Get maximum elevation (in grid space). */
	float GetMaxElevationGridSpace() const { return m_terrainModel->GetMaxElevationGridSpace(); }

	/** Set minimum elevation (in world space). */
	void SetMinElevation(float min) { m_terrainModel->SetMinElevation(min); }

	/** Get minimum elevation (in world space). */
	float GetMinElevation() const { return m_terrainModel->GetMinElevation(); }

	/** Set maximum elevation (in world space). */
	void SetMaxElevation(float max) { m_terrainModel->SetMaxElevation(max); }

	/** Get maximum elevation (in world space). */
	float GetMaxElevation() const { return m_terrainModel->GetMaxElevation(); }

	/** 
	 * @brief Project screen point onto terrain. 
	 * @param screenPt Point to project onto terrain.
	 * @param bInvertY Flag indicating if y coordinate should be inverted (i.e., should be true if Y is taken to be the top of the screen.)*/
	Point3D ProjectToTerrain(const Point2D& screenPt, bool bInvertY = false);

	/** Project 3D point to screen. */
	Point3D ProjectToScreen(const Point3D& worldPt);

protected:
	/** Terrain model. */
	TerrainModelPtr m_terrainModel;

	/** OpenGL (3D) view of terrain model. */
	TerrainViewPtr m_terrainView;
};

}

#endif
