//=======================================================================
// Author: Donovan Parks
//
// Copyright 2009 Donovan Parks
//
// This file is part of GenGIS.
//
// GenGIS is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// GenGIS is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with GenGIS.  If not, see <http://www.gnu.org/licenses/>.
//=======================================================================

#include "../core/Precompiled.hpp"

#include "../core/TerrainModel.hpp"
#include "../core/StudyController.hpp"
#include "../core/App.hpp"

#include "../utils/Log.hpp"
#include "../utils/StringTools.hpp"

using namespace GenGIS;

TerrainModel::TerrainModel( void ) : m_oCT(NULL), m_verticalExaggeration(1.0)
{

}

TerrainModel::~TerrainModel( void ) 
{

}

void TerrainModel::GeoToGrid(const GeoCoord &mapCoord, Point3D &gridCoord) 
{
	gridCoord.x = ((mapCoord.easting - GetHeader()->projExtents.x) / GetHeader()->projExtents.Width())
									* GetHeader()->width - GetHeader()->width*0.5f;
  gridCoord.z = ((GetHeader()->projExtents.dy - mapCoord.northing) / GetHeader()->projExtents.Height()) 
									* GetHeader()->height - GetHeader()->height*0.5f;

	gridCoord.y = GetExaggeratedElevation(gridCoord.x, gridCoord.z);
}

void TerrainModel::LatLongToGrid(const GeoCoord &mapCoord, Point3D &gridCoord) 
{
	double latitude = mapCoord.northing;
	double longitude = mapCoord.easting;

	// transform lat/lon coordinates to projected coordinates
	StudyControllerPtr studyController = App::Inst().GetStudyController();
	if(studyController->GetProject())
	{
		ProjectionToolPtr projTool = studyController->GetProjectionTool();
		if(!projTool->Transform(1, &longitude, &latitude))
		{
			error::Log::Inst().Warning("(Warning) Failed to project data.");
		}
	}

	GeoToGrid(GeoCoord(longitude, latitude), gridCoord);
}

void TerrainModel::GridToGeo(const Point3D &gridCoord, GeoCoord &mapCoord) 
{
	// x grid coordinates go from -1 to 1
	mapCoord.easting = GetHeader()->projExtents.x + 0.5*(gridCoord.x + 1)*GetHeader()->projExtents.Width();

	// z grid coordinates go from -height/2 to height/2
  mapCoord.northing = GetHeader()->projExtents.y + 0.5*(gridCoord.z + GetHeader()->height*0.5)*GetHeader()->projExtents.Height();
}

float TerrainModel::GetElevation(float gridX, float gridZ)
{
	float indexX = (gridX + GetHeader()->width*0.5) / GetHeader()->cellSizeX;
	float indexY = (gridZ + GetHeader()->height*0.5) / GetHeader()->cellSizeZ;
	float	iX = floor(indexX);
	float iY = floor(indexY);
	float dX = indexX - iX;
	float dY = indexY - iY;

	int tempCol = GetHeader()->nCols;
	int tempRows = GetHeader()->nRows;

	if(iX >= 0 && iX < (GetHeader()->nCols-1) && iY >= 0 && iY < (GetHeader()->nRows-1))
	{
		// find 2D interpolation of elevation
		float bottomStartX = m_grid[int(iX + iY*GetHeader()->nCols)].y;
		float bottomEndX = m_grid[int(iX+1 + iY*GetHeader()->nCols)].y;
		float elevBottom = bottomStartX + (bottomEndX - bottomStartX)*dX;

		float topStartX = m_grid[int(iX + (iY+1)*GetHeader()->nCols)].y;
		float topEndX = m_grid[int(iX+1 + (iY+1)*GetHeader()->nCols)].y;
		float elevTop = topStartX + (topEndX - topStartX)*dX;

		float elev = elevBottom + (elevTop - elevBottom)*dY;

		return elev;
	}
	else
	{
		// place any object that are beyond the map extents at an elevation of zero
		return 0;
	}
}
float TerrainModel::GetExaggeratedElevation(float gridX, float gridZ) 
{ 
	return m_verticalExaggeration*GetElevation(gridX, gridZ); 
}

