//=======================================================================
// Author: Donovan Parks
//
// Copyright 2009 Donovan Parks
//
// This file is part of GenGIS.
//
// GenGIS is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// GenGIS is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with GenGIS.  If not, see <http://www.gnu.org/licenses/>.
//=======================================================================

#ifndef _GENGIS_TERRAIN_MODEL_
#define _GENGIS_TERRAIN_MODEL_

#include "../core/Precompiled.hpp"

namespace GenGIS
{

/** Structure specifying DEM header information. */
typedef struct sFILE_HEADER 
{
	Box2D projExtents;	// lower, left and upper, right corners of map after projection

	int nCols;					// number of columns
	int nRows;					// number of rows
	float width;				// width of map in grid space
	float height;			// height of map in grid space
	float cellSizeX;		// in grid coordinates
	float cellSizeZ;		// in grid coordinates
	float scaleFactor;	// scale between projection space and grid (model) space 
											// (1 projection space unit = scaleFactor grid units)
	
	float noDataValue;	// value assigned to DEM cells that have no actual value
} FileHeader;

/** Structure specifying DEM metadata information. */
typedef struct sFILE_METADATA
{
	std::string driverDesc;							// driver description
	std::string driverMetaData;					// addition driver information
	std::string projection;							// string indicating projection parameters
	std::string colourInterpretation;		// not sure what this is?
	std::string dataType;								// type of data stored in DEM

	int xSize;					// number columns in DEM
	int ySize;					// number rows in DEM
	int bands;					// number of bands in DEM

	float originX;			// x-coordinate origin of DEM
	float originY;			// y-coordinate origin of DEM
	Box2D extents;			// extents of DEM (lower, left corner to upper, right)

	float pixelSizeX;	// size of cells in geographical coordinate space
	float pixelSizeY;	// size of cells in geographical coordinate space
} FileMetaData;

/**
 * @brief Model of digital elevation map (DEM).
 */
class TerrainModel 
{
public:
	/** Construction. */
	TerrainModel( void );

	/** Destructor. */
	~TerrainModel( void );

	/**
	 * @brief Get array of grid coordinates.
	 * @return Grid coordinate array.
	 */
	Point3DArrayPtr GetGrid() { return m_grid; }

	/**
	 * @brief Set array of grid coordinates.
	 * @param grid Grid coordinate array.
	 */
	void SetGrid( Point3DArrayPtr grid ) { m_grid = grid; }

	/**
	 * @brief Convert geographic coordinates to grid coordinates.
	 * @param mapCoord Geographic coordinates.
	 * @param gridCoord Grid coordinate.
	 */
	void GeoToGrid(const GeoCoord &mapCoord, Point3D &gridCoord);

	/**
	 * @brief Convert lat/long coordinates to grid coordinates.
	 * @param mapCoord Lat/long coordinates.
	 * @param gridCoord Grid coordinate.
	 */
	void LatLongToGrid(const GeoCoord &mapCoord, Point3D &gridCoord);

	/**
	 * @brief Convert grid coordinates to geographic coordinates.
	 * @param gridCoord Grid coordinate.
	 * @param mapCoord Geographic coordinates.
	 */
	void GridToGeo(const Point3D &gridCoord, GeoCoord &mapCoord);

	/**
	 * @brief Get DEM file header information.
	 * @return DEM header information.
	 */
	FileHeader* GetHeader() { return &m_fileHeader; }

	/**
	 * @brief Set transformation tool.
	 * @param tool New transformation tool.
	 */
	void SetTransformationTool(OGRCoordinateTransformation* tool) {	m_oCT = tool; }

	/**
	 * @brief Get transformation tool.
	 * @return Transformation tool.
	 */
	OGRCoordinateTransformation* GetTransformationTool()	{	return m_oCT;	}

	/**
	 * @brief Width (x-dimension) of DEM map. 
	 * @return Width.
	 */
	float GetWidth() {  return m_fileHeader.width; }

	/**
	 * @brief Height (z-dimension) of DEM map. 
	 * @return Height.
	 */
	float GetHeight() { return m_fileHeader.height; }

	/**
	 * @brief Elevation (y-dimension) of DEM map at specified point.
	 * @return Elevation.
	 */
	float GetElevation(float gridX, float gridZ);

	/**
	 * @brief Elevation (y-dimension) of DEM map at specified point multiplied by the vertical exaggeration.
	 * @return Exaggerated elevation.
	 */
	float GetExaggeratedElevation(float gridX, float gridZ);

	/**
	 * @brief Get DEM file metadata information.
	 * @return Metadata information.
	 */
	FileMetaData* GetMetaData() { return &m_metaData; }

	/** Set minimum elevation (in grid space). */
	void SetMinElevationGridSpace(float min) { m_minElevationGridSpace = min; }

	/** Get minimum elevation (in grid space). */
	float GetMinElevationGridSpace() const { return m_minElevationGridSpace; }

	/** Set maximum elevation (in grid space). */
	void SetMaxElevationGridSpace(float max) { m_maxElevationGridSpace = max; }

	/** Get maximum elevation (in grid space). */
	float GetMaxElevationGridSpace() const { return m_maxElevationGridSpace; }

	/** Set minimum elevation. */
	void SetMinElevation(float min) { m_minElevation = min; }

	/** Get minimum elevation. */
	float GetMinElevation() const { return m_minElevation; }

	/** Set maximum elevation (in grid coordinates). */
	void SetMaxElevation(float max) { m_maxElevation = max; }

	/** Get maximum elevation (in grid coordinates). */
	float GetMaxElevation() const { return m_maxElevation; }

	/** Set vertical exaggeration of elevation. */	
	void SetVerticalExaggeration(float exaggeration) { m_verticalExaggeration = exaggeration; }	
	
	/** Get vertical exaggeration of elevation. */	
	float GetVerticalExaggeration() { return m_verticalExaggeration; } 	
	
	/** Set flag indicating if terrain model contains elevation information. */	
	void SetElevationMap(bool bElevationMap) { m_bElevationMap = bElevationMap; }	
	
	/** Get flag indicating if terrain model contains elevation information. */	
	bool GetElevationMap() const { return m_bElevationMap; }	
	
	/** Set texture for terrain. */	
	void SetTexture(TerrainTexturePtr texture) { m_texture = texture; }	
	
	/** Get texture for terrain. */		
	TerrainTexturePtr GetTexture() { return m_texture; }
	
protected:
	/** DEM header information. */
	FileHeader m_fileHeader;

	/** DEM metadata information. */
	FileMetaData m_metaData;

	/** DEM grid specifying (x,y,z) coordinated. */
	Point3DArrayPtr m_grid;

	/** GDAL projection transformation utility. */
	OGRCoordinateTransformation *m_oCT;

	/** Minimum elevation (in grid space). */
	float m_minElevationGridSpace;

	/** Maximum elevation (in grid space). */
	float m_maxElevationGridSpace;

	/** Minimum elevation. */
	float m_minElevation;

	/** Maximum elevation. */
	float m_maxElevation;

	/** Amount to exaggerate elevation by. */	
	float m_verticalExaggeration;	
	
	/** Flag indicating if terrain model contains elevation information. */	
	bool m_bElevationMap;	
	
	/** Texture for terrain. */	
	TerrainTexturePtr m_texture;
};

}

#endif
