//=======================================================================
// Author: Donovan Parks
//
// Copyright 2009 Donovan Parks
//
// This file is part of GenGIS.
//
// GenGIS is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// GenGIS is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with GenGIS.  If not, see <http://www.gnu.org/licenses/>.
//=======================================================================

#ifndef _GENGIS_TERRAIN_VIEW_
#define _GENGIS_TERRAIN_VIEW_

#include "../core/Precompiled.hpp"

#include "../core/TerrainModel.hpp"
#include "../core/View.hpp"

namespace GenGIS
{

class TerrainView : public View
{
private:
	/** Holds all information required by a vertex which is part of the terrain. */
	typedef struct sVERTEX
	{
		typedef boost::shared_array<sVERTEX> ArrayPtr;
		typedef boost::shared_ptr<sVERTEX> Ptr;

		Point3D coordinate;		// 12
		bool	enabled;		// 1
		GLint	dependents[4];	// 4x4=16 (can't reduce size...) 
		bool	prob;			// 1
	} Vertex;					// 56??

	/** Quadtree structure used to build terrain model. */
	typedef struct sQUADTREE_NODE
	{
		GLint lodvertex[9];
		sQUADTREE_NODE *q1, *q2, *q3, *q4;
		bool enabled;
		float max_err;
	} QuadtreeNode;

	Point3D theNormal;

	void computNormal(int i, int j); 

public:
	/** Constructor. */
	TerrainView(TerrainModelPtr terrainModel, CameraPtr camera);

	/** Destructor. */
	~TerrainView();

	/** Build 3D terrain. */
	bool BuildTerrain(ProgressDlgPtr progressDlg = ProgressDlgPtr());
	
	/** Render terrain. */
	void Render();
	void Render(bool bSimplified);

	/** Check if terrain is being draw as a wireframe. */
	bool IsWireFrame() { return m_bWireFrame; }
	void SetWireFrame(bool state) { m_bWireFrame = state; }

	/** Set the amount of error allowed in determining the elevation. */
	void SetEpsilon(float error);

	/** Get the amount of error allowed in determining the elevation. */
	float GetEpsilon() { return m_epsilon; }

	/** Increment amount of error allowed in elevation. */
	void IncEpsilon(float amt);

	/** Set vertical exaggeration of elevation. */
	void SetVerticalExaggeration(float exaggeration);

	/** Get vertical exaggeration of elevation. */	
	float GetVerticalExaggeration() { return m_terrainModel->GetVerticalExaggeration(); }	

	/** Rebuild the LOD quadtree structure. */
	void RefreshQuadtree();

	/** Get terrain model associated with this view. */
	TerrainModelPtr GetTerrainModel() { return m_terrainModel; }

	/** Set max terrain resolution. */
	void SetMaxTerrainResolution(int resolution) { m_maxTerrainResolution = resolution; }

	/** Get max terrain resolution. */
	int GetMaxTerrainResolution() { return m_maxTerrainResolution; }

private:
	void SetVertices();
  void init_vertices();
  void getAverageNormal( int i, int j );
  void assignTextureCoordinate( int i , int j );

  // quadtree related functions
  QuadtreeNode* createQuadtree( int lodsize, int startindex );
  void delete_quadtree( QuadtreeNode *cur_node );
	bool inter_error( int center, int vertexA, int vertexB);
	bool box_test( QuadtreeNode *cur);
	void update_quadtree( QuadtreeNode *root_cur);
  void notify( int cur_idx );
  void update_anticrack( void );
  void clear_enabled_quadtree( QuadtreeNode *root_cur );
  void render_quadtree( QuadtreeNode *root_cur );
  void drawTerrainLeftover();
  void define_vertex( int index_vertex );

  //collapsing function
  void init_collapse();
  void enable_collapse_bounds();
  void collapse_geometry(Point3D &original, Point3D &transformed, bool& isScaled );
  void resort_geometry();

  // utility functions
  void crossProduct(Vertex &a, Vertex &b, Vertex &c, float* normal);

  QuadtreeNode* m_root;
  Vertex* m_vertices;

  // terrain state variables
  GLdouble m_epsilon; // for LOD
  bool m_bWireFrame;

  // Variables used for construct/rendering Quadtree
  // used by quadtree
  //----------------------------------
  int m_dimension;
  float m_curError;

  //
  // Collapsing control variables
  //----------------------------------
	std::pair<Point3D, Point3D> m_collapsePoints;
  int m_cbounderyCount;

  float m_collapseRatio, m_collapseRatioCur, m_collapseRatioInc;
  float m_uLast, m_vLast;

  float m_xAxis, m_zAxis;
  float m_k;
  float m_distTarget;
  bool m_select1, m_select2;
  float m_xCenter, m_zCenter;
  float m_collapseAngle;

	int m_maxTerrainResolution;

	GLuint m_displayList;

	TerrainModelPtr m_terrainModel;
	CameraPtr m_camera;

	int polyc;
};

}

#endif


