//=======================================================================
// Author: Donovan Parks
//
// Copyright 2009 Donovan Parks
//
// This file is part of GenGIS.
//
// GenGIS is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// GenGIS is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with GenGIS.  If not, see <http://www.gnu.org/licenses/>.
//=======================================================================

#include "../core/Precompiled.hpp"

#include "../gui/GenGisApp.hpp"
#include "../gui/GenGisFrame.hpp"

#include "../core/App.hpp"

#include "../utils/Log.hpp"
#include "../utils/ColourMapManager.hpp"
#include "../utils/ShapeMapManager.hpp"

#include "../core/API.hpp"

////////////////////////////////    ~mikep. Plugin work... ignore for now.
// #include "../core/PluginManager.hpp"
////////////////////////////////

// Don't use this code in Windows debug builds!
#ifndef _DEBUG
  #include "../python/PythonInterface.hpp"
#endif

// disable checked iterators as this is a large performance hit
// (see: http://msdn.microsoft.com/en-us/library/aa985896(VS.80).aspx and google search for SECURE_SCL)
#ifdef NDEBUG
	#define _SECURE_SCL 0
#endif

IMPLEMENT_APP(GenGisApp)

bool GenGisApp::OnInit()
{
	try
	{

		error::Log::Inst().SilentWrite("Program Started.");

		// initialize OpenGL fonts
		GenGIS::App::Inst().InitFont(_T("times.ttf"));

		GenGIS::App::Inst().SetResolutionFactor(1);
		GenGIS::App::Inst().SetSavingHiResImage(false);

		error::Log::Inst().SilentWrite("Fonts initialized.");
    
		// initialize image handlers
		wxImage::AddHandler(new wxPNGHandler);
		wxImage::AddHandler(new wxICOHandler);

		// initial file system handlers
		wxFileSystem::AddHandler(new wxInternetFSHandler);

		error::Log::Inst().SilentWrite("Handlers added.");

		// load main window
		GenGIS::GenGisFrame* frame = new GenGIS::GenGisFrame();
    GenGIS::App::Inst().SetMainWindow(frame);

		error::Log::Inst().SilentWrite("Main window set.");

		// initialize colour map manager
		GenGIS::ColourMapManagerPtr colourMapManager(new GenGIS::ColourMapManager());
		std::wstring colourMapPath = (GenGIS::App::Inst().GetExeDir() + wxT("colourMaps")).c_str();
		colourMapManager->LoadColourMaps(std::string(colourMapPath.begin(), colourMapPath.end()));
		GenGIS::App::Inst().SetColourMapManager(colourMapManager);

		error::Log::Inst().SilentWrite("Colour maps initialized.");

		// initialize shape map manager
		GenGIS::ShapeMapManagerPtr shapeMapManager(new GenGIS::ShapeMapManager());
		GenGIS::App::Inst().SetShapeMapManager(shapeMapManager);

		error::Log::Inst().SilentWrite("Shape maps initialized.");

		// set default database connection string
		//GenGIS::App::Inst().SetDatabaseConnectionString(_T("http://ratite.cs.dal.ca/mg_web/fetch_metagenome_samples.cgi?type=metagenome"));
    
    // Load the APIv1
    GenGIS::APIv1::Inst();

		error::Log::Inst().SilentWrite("API loaded.");
    
    // Don't use this code in Windows debug builds!
    #ifndef _DEBUG
      // Initialize python
      GenGIS::PythonInterface::Inst();
    #endif

		error::Log::Inst().SilentWrite("Python console initialized.");


    ///////////////////////////////     ~mikep. Plugins... Work in progress... please ignore for now.
    // Load Plugins
    // GenGIS::InternalPlugin& plugin = GenGIS::PluginManager::Inst().LoadPlugin("/Users/mike/GenGIS/mac/build/gcc/BUILD/DEBUG/OBJECTS","EnglishTextPlugin.so");
    // std::cout << "Name: " << plugin.GetFileName() << endl;
    // std::cout << plugin.GetExternalPlugin()->shortName;
    ///////////////////////////////
    // delete &text;
    // GenGIS::Plugin& text2 = GenGIS::PluginManager::Inst().GetPlugin("EnglishTextPlugin.so");
    // std::cout << "Name: " << text2.GetName() << endl;
    ////////////////////////////////
		
	  frame->InitializeTerrainController();
		
		error::Log::Inst().SilentWrite("Terrain controller initialized.");
    
	  SetTopWindow(frame);
	  wxIcon icon(GenGIS::App::Inst().GetExeDir() + wxT("images/CrazyEye.ico"), wxBITMAP_TYPE_ICO);
	  frame->SetIcon(icon);		
	  frame->Show();

		error::Log::Inst().SilentWrite("Frame shown.");
	}
	catch(...)
	{
		error::Log::Inst().Error("(Error) Unknown exception");
	}

	return true;
}
 
int GenGisApp::OnExit()
{
	return 0;
}
