//=======================================================================
// Author: Donovan Parks
//
// Copyright 2009 Donovan Parks
//
// This file is part of GenGIS.
//
// GenGIS is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// GenGIS is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with GenGIS.  If not, see <http://www.gnu.org/licenses/>.
//=======================================================================

#include "../core/Precompiled.hpp"

#include "../gui/GenGisFrame.hpp"

#include "../core/App.hpp"
#include "../core/TerrainView.hpp"
#include "../core/TerrainTexture.hpp"
#include "../core/GeoTreeController.hpp"
#include "../core/Camera.hpp"
#include "../core/LayerTreeView.hpp"
#include "../core/GeoTreeIO.hpp"
#include "../core/LocationSetIO.hpp"
#include "../core/SequenceIO.hpp"
#include "../core/SequenceController.hpp"
#include "../core/LayerTree.hpp"
#include "../core/LayoutLine.hpp"
#include "../core/LayoutEllipse.hpp"
#include "../core/GeoAxisPolyline.hpp"
#include "../core/LayerTreeController.hpp"
#include "../core/TerrainController.hpp"
#include "../core/PopupMenu.hpp"
#include "../core/Viewport.hpp"
#include "../core/LayerTreeItem.hpp"
#include "../core/Layer.hpp"
#include "../core/StudyLayer.hpp"
#include "../core/MapLayer.hpp"
#include "../core/LocationSetLayer.hpp"
#include "../core/LocationLayer.hpp"
#include "../core/TreeLayer.hpp"
#include "../core/SequenceLayer.hpp"
#include "../core/VisualLabel.hpp"
#include "../core/UserElements.hpp"
#include "../core/LabelController.hpp"
#include "../core/Lighting.hpp"

#include "../widgets/WidgetController.hpp"

#include "../gui/GenGisFrame.hpp"
#include "../gui/StudyPropertiesDlg.hpp"
#include "../gui/MapPropertiesDlg.hpp"
#include "../gui/LocationSetPropertiesDlg.hpp"
#include "../gui/LocationPropertiesDlg.hpp"
#include "../gui/SequencePropertiesDlg.hpp"
#include "../gui/TreePropertiesDlg.hpp"
#include "../gui/LayoutObjectPropertiesDlg.hpp"
#include "../gui/LightingPropertiesDlg.hpp"
#include "../gui/FontPropertiesDlg.hpp"
#include "../gui/ProgressDlg.hpp"
#include "../gui/LabelToolDlg.hpp"
#include "../gui/SeqSelectionDlg.hpp"
#include "../gui/DbSelectSampleDlg.hpp"
#include "../gui/ConfigureDatabaseDlg.hpp"
#include "../gui/SaveImageDlg.hpp"

#include "../utils/StringTools.hpp"
#include "../utils/UniqueId.hpp"
#include "../utils/ColourMap.hpp"
#include "../utils/ColourMapDiscrete.hpp"
#include "../utils/ColourMapManager.hpp"
#include "../utils/DemReader.hpp"
#include "../utils/Log.hpp"

using namespace GenGIS;

const int ID_MNU_LAYER_STUDY_PROPERTIES = 5000;
const int ID_MNU_LAYER_MAP_PROPERTIES = 5001;
const int ID_MNU_LAYER_LOCATION_SET_PROPERTIES = 5002;
const int ID_MNU_LAYER_LOCATION_PROPERTIES = 5003;
const int ID_MNU_LAYER_TREE_PROPERTIES = 5004;
const int ID_MNU_LAYER_SEQUENCE_PROPERTIES = 5005;

const int ID_RIGHT_CLICK_CLADOGRAM_3D = 5100;
const int ID_RIGHT_CLICK_SLANTED_CLADOGRAM_3D = 5101;
const int ID_RIGHT_CLICK_CLADOGRAM_2D = 5102;
const int ID_RIGHT_CLICK_PHYLOGRAM_2D = 5103;
const int ID_RIGHT_CLICK_SLANTED_CLADOGRAM_2D = 5104;
const int ID_RIGHT_CLICK_LAYOUT_ELLIPSE_QUADRANT = 5105;
const int ID_RIGHT_CLICK_TREE_LAYOUT_LINE = 5106;
const int ID_RIGHT_CLICK_TREE_LAYOUT_ELLIPSE = 5107;
const int ID_RIGHT_CLICK_TREE_LEAF_NODE_LABELS = 5108;
const int ID_RIGHT_CLICK_TREE_SIGNIFICANCE_TEST = 5109;
const int ID_RIGHT_CLICK_GEO_AXIS_TO_TREE = 5110;
const int ID_RIGHT_CLICK_GEO_AXIS_REVERSE_ORDER = 5112;
const int ID_RIGHT_CLICK_TREE_GEO_AXIS = 5011;

const int ID_RIGHT_CLICK_LAYOUT_CHARTS_EVENLY_ON_LINE = 5201;
const int ID_RIGHT_CLICK_LAYOUT_CHARTS_LINE_PROJECTED = 5202;
const int ID_RIGHT_CLICK_LAYOUT_CHARTS_ELLIPSE_QUADRANT = 5203;
const int ID_RIGHT_CLICK_LAYOUT_CHARTS_EVENLY_ON_ELLIPSE = 5204;
const int ID_RIGHT_CLICK_LAYOUT_CHARTS_ELLIPSE_PROJECTED = 5205;
const int ID_RIGHT_CLICK_CHART_SET_LAYOUT_LINE = 5206;
const int ID_RIGHT_CLICK_CHART_SET_LAYOUT_ELLIPSE = 5207;

const int ID_MNU_LAYOUT_LINE_DELETE = 6000;
const int ID_MNU_LAYOUT_ELLIPSE_DELETE = 6001;
const int ID_MNU_GEO_AXIS_POLYLINE_DELETE = 6002;

// Popup menu IDs are in the 7000's


GenGisFrame::GenGisFrame(wxWindow* parent): GenGisLayout(parent)
{
	error::Log::Inst().SetConsole(m_console);
	
	App::Inst().SetViewport(new Viewport((wxPanel*)m_pnlViewport, ID_VIEWPORT, wxPoint(0,0), wxSize(0,0)));

	// setup sizers so viewport automatically expand to fit panel
	wxBoxSizer* sizer = new wxBoxSizer( wxVERTICAL );
	sizer->Add(App::Inst().GetViewport(), 1, wxEXPAND);
	m_pnlViewport->SetSizer(sizer);

	int width, height;
	m_pnlViewport->GetClientSize(&width, &height);

	// initialize layer tree model, view, and controller
	LayerTreePtr layerTree(new LayerTree());
	LayerTreeViewPtr layerTreeView(new LayerTreeView(layerTree, m_treeCtrlLayer));
	LayerTreeControllerPtr layerTreeController(new LayerTreeController(layerTree, layerTreeView));
	App::Inst().SetLayerTreeController(layerTreeController);

	// select the first item (study) in the layer tree
	wxTreeItemId treeId = App::Inst().GetLayerTreeController()->GetTreeCtrl()->GetRootItem();
	wxTreeItemIdValue cookie;
	wxTreeItemId studyId = App::Inst().GetLayerTreeController()->GetTreeCtrl()->GetFirstChild(treeId, cookie);
	App::Inst().GetLayerTreeController()->GetTreeCtrl()->SelectItem(studyId);
	
	// attach viewport mouse and keyboard events to local functions
	App::Inst().GetViewport()->SignalRender(boost::bind(&GenGisFrame::SigRender, this));
	App::Inst().GetViewport()->SignalMouse(boost::bind(&GenGisFrame::SigViewportMouse, this, _1));
	App::Inst().GetViewport()->SignalMouseWheel(boost::bind(&GenGisFrame::SigViewportMouseWheel, this, _1));
	App::Inst().GetViewport()->SignalKeyboard(boost::bind(&GenGisFrame::SigViewportKeyboard, this, _1));

	// initialize OpenGL widget controller
	m_widgetController.reset(new WidgetController(App::Inst().GetViewport()));
	m_widgetController->SetLayerTree(layerTree);

	// seed random number generator
	srand((unsigned)time(0));

	ConnectEvents();
}

GenGisFrame::~GenGisFrame()
{
	DisconnectEvents();
}

void GenGisFrame::ConnectEvents()
{
	this->Connect( ID_MNU_LAYER_STUDY_PROPERTIES, wxEVT_COMMAND_MENU_SELECTED, wxCommandEventHandler( GenGisFrame::OnStudyProperties ) );
	this->Connect( ID_MNU_LAYER_MAP_PROPERTIES, wxEVT_COMMAND_MENU_SELECTED, wxCommandEventHandler( GenGisFrame::OnMapProperties ) );
	this->Connect( ID_MNU_LAYER_LOCATION_SET_PROPERTIES, wxEVT_COMMAND_MENU_SELECTED, wxCommandEventHandler( GenGisFrame::OnLocationSetProperties ) );
	this->Connect( ID_MNU_LAYER_LOCATION_PROPERTIES, wxEVT_COMMAND_MENU_SELECTED, wxCommandEventHandler( GenGisFrame::OnLocationProperties ) );
	this->Connect( ID_MNU_LAYER_SEQUENCE_PROPERTIES, wxEVT_COMMAND_MENU_SELECTED, wxCommandEventHandler( GenGisFrame::OnSequenceProperties ) );
	this->Connect( ID_MNU_LAYER_TREE_PROPERTIES, wxEVT_COMMAND_MENU_SELECTED, wxCommandEventHandler( GenGisFrame::OnTreeProperties ) );

	this->Connect( ID_RIGHT_CLICK_CLADOGRAM_3D, wxEVT_COMMAND_MENU_SELECTED, wxCommandEventHandler( GenGisFrame::OnCladogram3D ) );
	this->Connect( ID_RIGHT_CLICK_SLANTED_CLADOGRAM_3D, wxEVT_COMMAND_MENU_SELECTED, wxCommandEventHandler( GenGisFrame::OnSlantedCladogram3D ) );
	this->Connect( ID_RIGHT_CLICK_CLADOGRAM_2D, wxEVT_COMMAND_MENU_SELECTED, wxCommandEventHandler( GenGisFrame::OnCladogram2D ) );
	this->Connect( ID_RIGHT_CLICK_PHYLOGRAM_2D, wxEVT_COMMAND_MENU_SELECTED, wxCommandEventHandler( GenGisFrame::OnPhylogram2D ) );
	this->Connect( ID_RIGHT_CLICK_SLANTED_CLADOGRAM_2D, wxEVT_COMMAND_MENU_SELECTED, wxCommandEventHandler( GenGisFrame::OnSlantedCladogram2D ) );
	this->Connect( ID_RIGHT_CLICK_LAYOUT_ELLIPSE_QUADRANT, wxEVT_COMMAND_MENU_SELECTED, wxCommandEventHandler( GenGisFrame::OnLayoutTreeEllipseQuadrant) );
	this->Connect( ID_RIGHT_CLICK_TREE_LAYOUT_LINE, wxEVT_COMMAND_MENU_SELECTED, wxCommandEventHandler( GenGisFrame::OnTreeLayoutLineVisibility ) );
	this->Connect( ID_RIGHT_CLICK_TREE_LAYOUT_ELLIPSE, wxEVT_COMMAND_MENU_SELECTED, wxCommandEventHandler( GenGisFrame::OnTreeLayoutEllipseVisibility ) );
	this->Connect( ID_RIGHT_CLICK_TREE_LEAF_NODE_LABELS, wxEVT_COMMAND_MENU_SELECTED, wxCommandEventHandler( GenGisFrame::OnLeafNodeLabels ) );
	this->Connect( ID_RIGHT_CLICK_TREE_SIGNIFICANCE_TEST, wxEVT_COMMAND_MENU_SELECTED, wxCommandEventHandler( GenGisFrame::OnTreeSignificanceTest ) );
	this->Connect( ID_RIGHT_CLICK_GEO_AXIS_TO_TREE, wxEVT_COMMAND_MENU_SELECTED, wxCommandEventHandler( GenGisFrame::OnAssignGeoAxisToTree ) );
	this->Connect( ID_RIGHT_CLICK_GEO_AXIS_REVERSE_ORDER, wxEVT_COMMAND_MENU_SELECTED, wxCommandEventHandler( GenGisFrame::OnReverseGeoAxisOrder ) );	
	this->Connect( ID_RIGHT_CLICK_TREE_GEO_AXIS, wxEVT_COMMAND_MENU_SELECTED, wxCommandEventHandler( GenGisFrame::OnTreeGeographicAxisVisibility ) );

	this->Connect( ID_RIGHT_CLICK_LAYOUT_CHARTS_EVENLY_ON_LINE, wxEVT_COMMAND_MENU_SELECTED, wxCommandEventHandler( GenGisFrame::OnLayoutChartsEvenlyOnLine ) );
	this->Connect( ID_RIGHT_CLICK_LAYOUT_CHARTS_LINE_PROJECTED, wxEVT_COMMAND_MENU_SELECTED, wxCommandEventHandler( GenGisFrame::OnLayoutChartsLineProjected ) );
	this->Connect( ID_RIGHT_CLICK_LAYOUT_CHARTS_ELLIPSE_QUADRANT, wxEVT_COMMAND_MENU_SELECTED, wxCommandEventHandler( GenGisFrame::OnLayoutChartsEllipseQuadrant ) );
	this->Connect( ID_RIGHT_CLICK_LAYOUT_CHARTS_EVENLY_ON_ELLIPSE, wxEVT_COMMAND_MENU_SELECTED, wxCommandEventHandler( GenGisFrame::OnLayoutChartsEvenlyOnEllipse ) );
	this->Connect( ID_RIGHT_CLICK_LAYOUT_CHARTS_ELLIPSE_PROJECTED, wxEVT_COMMAND_MENU_SELECTED, wxCommandEventHandler( GenGisFrame::OnLayoutChartsEllipseProjected) );
	this->Connect( ID_RIGHT_CLICK_CHART_SET_LAYOUT_LINE, wxEVT_COMMAND_MENU_SELECTED, wxCommandEventHandler( GenGisFrame::OnChartSetLayoutLine ) );
	this->Connect( ID_RIGHT_CLICK_CHART_SET_LAYOUT_ELLIPSE, wxEVT_COMMAND_MENU_SELECTED, wxCommandEventHandler( GenGisFrame::OnChartSetLayoutEllipse ) );

	this->Connect( ID_MNU_LAYOUT_LINE_DELETE, wxEVT_COMMAND_MENU_SELECTED, wxCommandEventHandler( GenGisFrame::OnLayoutLineDelete ) );
	this->Connect( ID_MNU_LAYOUT_ELLIPSE_DELETE, wxEVT_COMMAND_MENU_SELECTED, wxCommandEventHandler( GenGisFrame::OnLayoutEllipseDelete ) );
	this->Connect( ID_MNU_GEO_AXIS_POLYLINE_DELETE, wxEVT_COMMAND_MENU_SELECTED, wxCommandEventHandler( GenGisFrame::OnGeoAxisPolylineDelete ) );
	this->Connect( ID_POPUP_MNU_HIDE, wxEVT_COMMAND_MENU_SELECTED, wxCommandEventHandler( GenGisFrame::OnSelectedViewHide ) );
	this->Connect( ID_POPUP_MNU_PROPERTIES, wxEVT_COMMAND_MENU_SELECTED, wxCommandEventHandler( GenGisFrame::OnSelectedViewProperties ) );
	this->Connect( ID_POPUP_MNU_REMOVE, wxEVT_COMMAND_MENU_SELECTED, wxCommandEventHandler( GenGisFrame::OnSelectedViewRemove ) );

	this->Connect( ID_POPUP_MNU_SUBTREE_SIGNIFICANCE_TEST, wxEVT_COMMAND_MENU_SELECTED, wxCommandEventHandler( GenGisFrame::OnSubTreeSignificanceTest ) );
	this->Connect( ID_POPUP_MNU_EXTEND_POLYLINE, wxEVT_COMMAND_MENU_SELECTED, wxCommandEventHandler( GenGisFrame::OnExtendGeographicPolyline ) );
}

void GenGisFrame::DisconnectEvents()
{
	this->Disconnect( ID_MNU_LAYER_STUDY_PROPERTIES, wxEVT_COMMAND_MENU_SELECTED, wxCommandEventHandler( GenGisFrame::OnStudyProperties ) );
	this->Disconnect( ID_MNU_LAYER_MAP_PROPERTIES, wxEVT_COMMAND_MENU_SELECTED, wxCommandEventHandler( GenGisFrame::OnMapProperties ) );
	this->Disconnect( ID_MNU_LAYER_LOCATION_SET_PROPERTIES, wxEVT_COMMAND_MENU_SELECTED, wxCommandEventHandler( GenGisFrame::OnLocationSetProperties ) );
	this->Disconnect( ID_MNU_LAYER_LOCATION_PROPERTIES, wxEVT_COMMAND_MENU_SELECTED, wxCommandEventHandler( GenGisFrame::OnLocationProperties ) );
	this->Disconnect( ID_MNU_LAYER_SEQUENCE_PROPERTIES, wxEVT_COMMAND_MENU_SELECTED, wxCommandEventHandler( GenGisFrame::OnSequenceProperties ) );
	this->Disconnect( ID_MNU_LAYER_TREE_PROPERTIES, wxEVT_COMMAND_MENU_SELECTED, wxCommandEventHandler( GenGisFrame::OnTreeProperties ) );

	this->Disconnect( ID_RIGHT_CLICK_CLADOGRAM_3D, wxEVT_COMMAND_MENU_SELECTED, wxCommandEventHandler( GenGisFrame::OnCladogram3D ) );
	this->Disconnect( ID_RIGHT_CLICK_SLANTED_CLADOGRAM_3D, wxEVT_COMMAND_MENU_SELECTED, wxCommandEventHandler( GenGisFrame::OnSlantedCladogram3D ) );
	this->Disconnect( ID_RIGHT_CLICK_CLADOGRAM_2D, wxEVT_COMMAND_MENU_SELECTED, wxCommandEventHandler( GenGisFrame::OnCladogram2D ) );
	this->Disconnect( ID_RIGHT_CLICK_PHYLOGRAM_2D, wxEVT_COMMAND_MENU_SELECTED, wxCommandEventHandler( GenGisFrame::OnPhylogram2D ) );
	this->Disconnect( ID_RIGHT_CLICK_SLANTED_CLADOGRAM_2D, wxEVT_COMMAND_MENU_SELECTED, wxCommandEventHandler( GenGisFrame::OnSlantedCladogram2D ) );
	this->Disconnect( ID_RIGHT_CLICK_LAYOUT_ELLIPSE_QUADRANT, wxEVT_COMMAND_MENU_SELECTED, wxCommandEventHandler( GenGisFrame::OnLayoutTreeEllipseQuadrant) );
	this->Disconnect( ID_RIGHT_CLICK_TREE_LAYOUT_LINE, wxEVT_COMMAND_MENU_SELECTED, wxCommandEventHandler( GenGisFrame::OnTreeLayoutLineVisibility ) );
	this->Disconnect( ID_RIGHT_CLICK_TREE_LAYOUT_ELLIPSE, wxEVT_COMMAND_MENU_SELECTED, wxCommandEventHandler( GenGisFrame::OnTreeLayoutEllipseVisibility ) );
	this->Disconnect( ID_RIGHT_CLICK_TREE_LEAF_NODE_LABELS, wxEVT_COMMAND_MENU_SELECTED, wxCommandEventHandler( GenGisFrame::OnLeafNodeLabels ) );
	this->Disconnect( ID_RIGHT_CLICK_TREE_SIGNIFICANCE_TEST, wxEVT_COMMAND_MENU_SELECTED, wxCommandEventHandler( GenGisFrame::OnTreeSignificanceTest ) );
	this->Disconnect( ID_RIGHT_CLICK_GEO_AXIS_TO_TREE, wxEVT_COMMAND_MENU_SELECTED, wxCommandEventHandler( GenGisFrame::OnAssignGeoAxisToTree ) );
	this->Disconnect( ID_RIGHT_CLICK_GEO_AXIS_REVERSE_ORDER, wxEVT_COMMAND_MENU_SELECTED, wxCommandEventHandler( GenGisFrame::OnReverseGeoAxisOrder ) );	
	this->Disconnect( ID_RIGHT_CLICK_TREE_GEO_AXIS, wxEVT_COMMAND_MENU_SELECTED, wxCommandEventHandler( GenGisFrame::OnTreeGeographicAxisVisibility ) );

	this->Disconnect( ID_RIGHT_CLICK_LAYOUT_CHARTS_EVENLY_ON_LINE, wxEVT_COMMAND_MENU_SELECTED, wxCommandEventHandler( GenGisFrame::OnLayoutChartsEvenlyOnLine ) );
	this->Disconnect( ID_RIGHT_CLICK_LAYOUT_CHARTS_LINE_PROJECTED, wxEVT_COMMAND_MENU_SELECTED, wxCommandEventHandler( GenGisFrame::OnLayoutChartsLineProjected ) );
	this->Disconnect( ID_RIGHT_CLICK_LAYOUT_CHARTS_ELLIPSE_QUADRANT, wxEVT_COMMAND_MENU_SELECTED, wxCommandEventHandler( GenGisFrame::OnLayoutChartsEllipseQuadrant ) );	
	this->Disconnect( ID_RIGHT_CLICK_LAYOUT_CHARTS_EVENLY_ON_ELLIPSE, wxEVT_COMMAND_MENU_SELECTED, wxCommandEventHandler( GenGisFrame::OnLayoutChartsEvenlyOnEllipse ) );
	this->Disconnect( ID_RIGHT_CLICK_LAYOUT_CHARTS_ELLIPSE_PROJECTED, wxEVT_COMMAND_MENU_SELECTED, wxCommandEventHandler( GenGisFrame::OnLayoutChartsEllipseProjected) );
	this->Disconnect( ID_RIGHT_CLICK_CHART_SET_LAYOUT_LINE, wxEVT_COMMAND_MENU_SELECTED, wxCommandEventHandler( GenGisFrame::OnChartSetLayoutLine ) );
	this->Disconnect( ID_RIGHT_CLICK_CHART_SET_LAYOUT_ELLIPSE, wxEVT_COMMAND_MENU_SELECTED, wxCommandEventHandler( GenGisFrame::OnChartSetLayoutEllipse ) );

	this->Disconnect( ID_MNU_LAYOUT_LINE_DELETE, wxEVT_COMMAND_MENU_SELECTED, wxCommandEventHandler( GenGisFrame::OnLayoutLineDelete ) );
	this->Disconnect( ID_MNU_LAYOUT_ELLIPSE_DELETE, wxEVT_COMMAND_MENU_SELECTED, wxCommandEventHandler( GenGisFrame::OnLayoutEllipseDelete ) );
	this->Disconnect( ID_MNU_GEO_AXIS_POLYLINE_DELETE, wxEVT_COMMAND_MENU_SELECTED, wxCommandEventHandler( GenGisFrame::OnGeoAxisPolylineDelete ) );
	this->Disconnect( ID_POPUP_MNU_HIDE, wxEVT_COMMAND_MENU_SELECTED, wxCommandEventHandler( GenGisFrame::OnSelectedViewHide ) );
	this->Disconnect( ID_POPUP_MNU_PROPERTIES, wxEVT_COMMAND_MENU_SELECTED, wxCommandEventHandler( GenGisFrame::OnSelectedViewProperties ) );
	this->Disconnect( ID_POPUP_MNU_REMOVE, wxEVT_COMMAND_MENU_SELECTED, wxCommandEventHandler( GenGisFrame::OnSelectedViewRemove ) );

	this->Disconnect( ID_POPUP_MNU_SUBTREE_SIGNIFICANCE_TEST, wxEVT_COMMAND_MENU_SELECTED, wxCommandEventHandler( GenGisFrame::OnSubTreeSignificanceTest ) );
	this->Disconnect( ID_POPUP_MNU_EXTEND_POLYLINE, wxEVT_COMMAND_MENU_SELECTED, wxCommandEventHandler( GenGisFrame::OnExtendGeographicPolyline ) );
}

void GenGisFrame::AddConsolePanel(wxPanel* panel, wxString name)
{
  m_bottomNotebook->AddPage(panel, name, false);
}

wxNotebook* GenGisFrame::GetBottomNotebook()
{
  return m_bottomNotebook;
}

void GenGisFrame::InitializeTerrainController()
{
  // Initialize the m_terrainController with empty new objects... This is needed after adding the Python console to the bottom notebook, otherwise maps won't load.
  // Chances are this code should be move to some constructor or something... putting it here as a quick fix.
  TerrainModelPtr terrainModel(new TerrainModel());
  TerrainViewPtr terrainView(new TerrainView(terrainModel, App::Inst().GetViewport()->GetCamera()));
  m_terrainController.reset(new TerrainController(terrainModel, terrainView));
}

void GenGisFrame::FillSequenceLegend()
{
  LayerTreeControllerPtr layerTreeController = App::Inst().GetLayerTreeController();
  
  std::map<std::wstring, Colour> seenNames;
  for(uint eachLayer = 0; eachLayer < layerTreeController->GetNumLocationSetLayers(); eachLayer++)    // for all LocationSetLayers...
  {
    LocationSetLayerPtr locationSetLayer = layerTreeController->GetLocationSetLayer(eachLayer);
    for(uint eachLocation = 0; eachLocation < locationSetLayer->GetNumLocations(); eachLocation++)   // for all LocationLayers...
    {
      ChartViewPtr chartView = locationSetLayer->GetLocation(eachLocation)->GetLocationController()->GetChartView();
      ColourMapDiscretePtr colourMap = chartView->GetColourMap();
      std::vector<std::wstring> chartViewNames = chartView->GetNames();
      std::vector<std::wstring>::iterator it;
      for(it = chartViewNames.begin(); it != chartViewNames.end(); ++it)    // for each ChartView's name...
      {
        std::wstring wstr = *it;
        if(seenNames.count(wstr) > 0)
        {
          continue;
        }
				
				Colour colour;
				colourMap->GetColour(wstr, colour);
				float assignToOther = chartView->GetAssignToOther();
				if(chartView->GetPercentage(wstr) > chartView->GetAssignToOther()/100)
					seenNames.insert(make_pair(wstr, colour));    
      }
						
			if(chartViewNames.size() != 0)
			{
				std::wstring otherStr = _T("Other");
				Colour colour;
				colourMap->GetColour(otherStr, colour);			
				seenNames.insert(make_pair(otherStr, colour));	
			}
    }
  }

	m_legendSequenceSizer->GetContainingWindow()->Freeze();
	m_legendSequenceSizer->Clear(true);
  std::map<std::wstring,Colour>::iterator it;
  for(it = seenNames.begin(); it != seenNames.end(); ++it)
  {
    wxColourPickerCtrl* wxtempctrl = new wxColourPickerCtrl(m_legendSequence, wxID_ANY, wxColour((it->second).GetRedInt(), (it->second).GetGreenInt(), (it->second).GetBlueInt()), wxDefaultPosition, wxDefaultSize, wxCLRP_DEFAULT_STYLE);
		wxtempctrl->Enable(false);
		wxtempctrl->SetForegroundColour(wxColour(0,0,250));
    m_legendSequenceSizer->Add(wxtempctrl, 0, wxALL, 5);

    wxStaticText* tempStaticText = new wxStaticText(m_legendSequence, wxID_ANY, wxString((it->first).c_str()), wxDefaultPosition, wxDefaultSize, 0);
  	tempStaticText->Wrap(-1);
  	m_legendSequenceSizer->Add(tempStaticText, 0, wxALIGN_CENTER_VERTICAL|wxALL, 5);
	}

	m_legendSequenceSizer->GetContainingWindow()->Thaw();
	m_legendSequenceSizer->Layout();																										// Redraw the flexGridSizer
  m_legendSequenceSizer->FitInside(m_legendSequenceSizer->GetContainingWindow());			// Force the scrollbars to reset
}

void GenGisFrame::OnSamplesLegendPaint(wxPaintEvent& event)
{
  wxGLCanvas* shape = (wxGLCanvas*)event.GetEventObject();
	int id = event.GetId();

	shape->SetCurrent();
  wxPaintDC dc(shape);

	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	glOrtho(0.0, shape->GetSize().x, 0.0, shape->GetSize().y, 1, 100 );

	glMatrixMode( GL_MODELVIEW );
	glPushMatrix();
	glLoadIdentity();

	glClearColor(1.0, 1.0, 1.0, 1.0);
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	uint x = shape->GetSize().x;
	uint y = shape->GetSize().y;
	glViewport(0, 0, (GLint)shape->GetSize().x, (GLint)shape->GetSize().y);

	bool bDone = false;
	for(uint eachLocationSet = 0; eachLocationSet < App::Inst().GetLayerTreeController()->GetNumLocationSetLayers(); eachLocationSet++)
  {
    LocationSetLayerPtr locationSetLayer = App::Inst().GetLayerTreeController()->GetLocationSetLayer(eachLocationSet);

    for(uint eachLocation = 0; eachLocation < locationSetLayer->GetNumLocations(); eachLocation++)
    {
			LocationViewPtr locationView = locationSetLayer->GetLocation(eachLocation)->GetLocationController()->GetLocationView();
			if(locationView->GetId() == id)
			{
				Colour colour = locationView->GetColour();

				glTranslated(x/2, y/2, -10);
				VisualMarker::RenderMarker(locationView->GetSize(), colour, locationView->GetShape());

				bDone = true;
				break;
			}
		}

		if(bDone)
			break;
	}

	glPopMatrix();
	glMatrixMode( GL_PROJECTION );
	glPopMatrix();
	glMatrixMode( GL_MODELVIEW );

	shape->SwapBuffers();
}

// Fills the Samples legend (on the sidebar) with data.
void GenGisFrame::FillSamplesLegend()
{
	m_legendLocationsSizer->GetContainingWindow()->Freeze();
	m_legendLocationsSizer->Clear(true);

  // Iterate over each LocationSet:
  for(uint eachLocationSet = 0; eachLocationSet < App::Inst().GetLayerTreeController()->GetNumLocationSetLayers(); eachLocationSet++)
  {
    LocationSetLayerPtr locationSetLayer = App::Inst().GetLayerTreeController()->GetLocationSetLayer(eachLocationSet);
    
    // For each Location in this LocationSet:
    for(uint eachLocation = 0; eachLocation < locationSetLayer->GetNumLocations(); eachLocation++)    // For all LocationLayers...
    {
			// add OpenGL canvas for rendering location marker
      LocationViewPtr locationMarkerShape = locationSetLayer->GetLocation(eachLocation)->GetLocationController()->GetLocationView();
      wxGLCanvas* shape = new wxGLCanvas(m_legendLocations, locationMarkerShape->GetId(), wxDefaultPosition, wxDefaultSize, wxALIGN_CENTRE);
			shape->Connect(wxEVT_PAINT, wxPaintEventHandler(GenGisFrame::OnSamplesLegendPaint));
			m_legendLocationsSizer->Add(shape, 0, wxALL|wxALIGN_CENTER_VERTICAL, 5);
     
			// add location marker label
      wxString locationId = wxString(locationSetLayer->GetLocation(eachLocation)->GetLocationController()->GetId().c_str());
	    wxStaticText* id = new wxStaticText( m_legendLocations, wxID_ANY, locationId, wxDefaultPosition, wxDefaultSize, wxALIGN_CENTRE);
	    id->Wrap( -1 );
      m_legendLocationsSizer->Add(id, 0, wxALL|wxALIGN_CENTER_VERTICAL, 5);
    }
  }

	m_legendLocationsSizer->GetContainingWindow()->Thaw();
	m_legendLocationsSizer->Layout();																										// Redraw the flexGridSizer
  m_legendLocationsSizer->FitInside(m_legendLocationsSizer->GetContainingWindow());			// Force the scrollbars to reset
}

void GenGisFrame::OnFileSaveImageAs( wxCommandEvent& event )
{
	SaveImageDlg* saveDlg = new SaveImageDlg(this);
	saveDlg->ShowModal();

	if(saveDlg->GetFilename() != _T(""))
  {
		// save (possibly) high resolution image
		App::Inst().SetResolutionFactor(saveDlg->GetResolutionScalingFactor());
		App::Inst().SetSavingHiResImage(true);

		App::Inst().GetViewport()->SaveImage(std::string(saveDlg->GetFilename().utf8_str()));

		App::Inst().SetResolutionFactor(1);
		App::Inst().SetSavingHiResImage(false);

		// redraw screen
		App::Inst().GetViewport()->Refresh(false); 
  }

	saveDlg->Destroy();
}

// Thin wrapper method for opening maps from File menu
void GenGisFrame::OnLayerOpenMap(wxCommandEvent& event)
{
	if(m_terrainController->GetTerrainView()->GetTerrainModel()->GetTexture())
	{
		wxMessageBox(wxT("GenGIS currently supports only a single map."),
												wxT("Multiple maps not supported"), wxOK | wxICON_INFORMATION);
		return;
	}


	wxFileDialog* openDlg = new wxFileDialog(this, wxT("Open map file..."), wxT(""), wxT(""), wxT("*.*"), wxFD_OPEN | wxFD_FILE_MUST_EXIST | wxFD_CHANGE_DIR);

	wxString wildcards = wxT("Common Raster Data Sources (*.dem; *.asc; *.ascii; *.tif; *.tiff)|*.dem; *.asc; *.ascii; *.tif; *.tiff|")
											 wxT("Arc/Info ASCII Grid (*.asc; *.ascii)|*.asc; *.ascii|")
											 wxT("GeoTIFF (*.tif; *.tiff)|*.tif; *.tiff|")
											 wxT("USGS Optional ASCII DEM (*.dem)|*.dem|")
											 wxT("All files (*.*)|*.*");
 	openDlg->SetWildcard(wildcards);

  if(openDlg->ShowModal() != wxID_OK)
  {
		openDlg->Destroy();
    return;
  }
  
	std::wstring filename = openDlg->GetFilename().wc_str(*wxConvCurrent);
	std::wstring path = openDlg->GetDirectory().wc_str(*wxConvCurrent);

	openDlg->Destroy();
  
  LayerOpenMap(filename, path);
}

// Thin wrapper method to convert std::string to two std::wstrings
void GenGisFrame::LayerOpenMap(std::string file)
{
	if(m_terrainController->GetTerrainView()->GetTerrainModel()->GetTexture())
	{
		wxMessageBox(wxT("GenGIS currently supports only a single map."),
												wxT("Multiple maps not supported"), wxOK | wxICON_INFORMATION);
		return;
	}

	std::string path, filename;
	util::StringTools::ExtractPathAndFilename(file, path, filename);  
	LayerOpenMap(util::StringTools::ToStringW(filename), util::StringTools::ToStringW(path));
}

void GenGisFrame::LayerOpenMap(std::wstring filename, std::wstring path)
{
	LayerPtr selectedLayer = App::Inst().GetLayerTreeController()->GetSelectedLayer();
	if(selectedLayer != LayerPtr() && selectedLayer->GetType() == Layer::STUDY)
	{
		// loading a DEM may take some time so it is nice to have a progress dialog
		std::wstring str = _T("Loading DEM file.");
		str.resize(75, _T(' '));
		ProgressDlgPtr progressDlg(new ProgressDlg(wxT("Loading file"), str, 100, this));

		TerrainModelPtr terrainModel(new TerrainModel());
		StudyControllerPtr studyController = App::Inst().GetLayerTreeController()->GetStudyLayer(0)->GetStudyController();
		DemReader demReader;

		std::wstring fullPath;
		if(path != _T(""))
			fullPath = path + wxT("/") + filename;
		else
			fullPath = filename;

		bool bReader = demReader.Open(fullPath, terrainModel, studyController, progressDlg);
		if(bReader)
		{
			TerrainViewPtr terrainView(new TerrainView(terrainModel, App::Inst().GetViewport()->GetCamera()));
			m_terrainController.reset(new TerrainController(terrainModel, terrainView));
			MapLayerPtr map(new MapLayer(UniqueId::Inst().GenerateId(), App::Inst().GetLayerTreeController()->GetSelectedLayer(), m_terrainController));
			map->SetName(filename.substr(0, filename.rfind(wxT("."), filename.length())));
			map->SetPath(path);
			map->SetFilename(filename);

			progressDlg->Update(0, wxT("Building terrain."));
			bool bBuilt = terrainView->BuildTerrain(progressDlg);
			if(!bBuilt)
			{
				error::Log::Inst().Warning("(Warning) Failed to build terrain.");
				return;
			}

			bool bAddMap = App::Inst().GetLayerTreeController()->AddMap(map);	

			// set map extents
			float width = m_terrainController->GetWidth();
			float height = m_terrainController->GetHeight();
			App::Inst().GetViewport()->GetCamera()->SetModelSpaceExtents(-width*0.5, width*0.5, -height*0.5, height*0.5);

			// position camera and lights in a nice default position
			LightingPtr lighting = App::Inst().GetViewport()->GetLighting();
			if(terrainModel->GetElevationMap())
			{
				ViewReset();

				lighting->setAltitude(0.667f);
				lighting->setGlobalAmbient(0.5f);
			}
			else
			{
				ViewTop();

				lighting->setAltitude(1.0f);
				lighting->setGlobalAmbient(1.0f);
			}

			m_widgetController->SetMiniMapTerrain(m_terrainController);
		}
		else
		{
			wxMessageBox(wxT("Failed to load map file. Check console window for warning messages."), wxT("Failed to read file"), wxOK | wxICON_INFORMATION);
		}
	}
	else
	{
		wxMessageBox(wxT("Please select a study node before adding a new map."),
												wxT("Select study node"), wxOK | wxICON_INFORMATION);
	}
}

void GenGisFrame::OnLayerOpenLocations( wxCommandEvent& event )
{
	wxFileDialog* openDlg = new wxFileDialog(this, wxT("Open location set file..."), wxT(""), wxT(""), wxT("*.*"), 
																							wxFD_OPEN | wxFD_FILE_MUST_EXIST | wxFD_CHANGE_DIR);

	wxString wildcards = _T("Location set file (*.csv)|*.csv| All files (*.*)|*.*");
 	openDlg->SetWildcard(wildcards);

  if(openDlg->ShowModal() != wxID_OK)
  {
		openDlg->Destroy();
    return;
  }

	std::wstring filename = openDlg->GetFilename().wc_str(*wxConvCurrent);
	std::wstring path = openDlg->GetDirectory().wc_str(*wxConvCurrent);

	openDlg->Destroy();
	
  LayerOpenLocations(filename, path);

	FillSamplesLegend();
}

void GenGisFrame::LayerOpenLocations(std::string file)
{
	std::string path, filename;
	util::StringTools::ExtractPathAndFilename(file, path, filename);  
	LayerOpenLocations(util::StringTools::ToStringW(filename), util::StringTools::ToStringW(path));
}

void GenGisFrame::LayerOpenLocations(std::wstring filename, std::wstring path)
{
	LayerPtr selectedLayer = App::Inst().GetLayerTreeController()->GetSelectedLayer();
	if(selectedLayer != LayerPtr() && selectedLayer->GetType() == Layer::MAP)
	{
		std::vector<LocationModelPtr> locationModels;
		StudyControllerPtr studyController = App::Inst().GetLayerTreeController()->GetStudyLayer(0)->GetStudyController();
		if(LocationSetIO::Read(path, filename, studyController, locationModels))
		{	
			// stop refreshing the tree until all sequences are loaded (this is purely for efficency)
			App::Inst().GetLayerTreeController()->GetTreeCtrl()->Freeze();

			ChartSetViewPtr chartSetCtrl(new ChartSetView());
			LocationSetLayerPtr locationSet(new LocationSetLayer(UniqueId::Inst().GenerateId(), 
											App::Inst().GetLayerTreeController()->GetSelectedLayer(),
											chartSetCtrl));
			locationSet->SetName(filename.substr(0, filename.rfind(wxT("."), filename.length())));
			locationSet->SetPath(path);
			locationSet->SetFilename(filename);
			m_locationSetLayer = locationSet;

			// assign default colour map to location
			ColourMapManagerPtr colourMapManager = App::Inst().GetColourMapManager();
			ColourMapPtr defaultColourMap = colourMapManager->GetDefaultDiscreteColourMap();
			ColourMapDiscretePtr newColourMap(new ColourMapDiscrete(defaultColourMap));

			foreach(LocationModelPtr locationModel, locationModels)
			{
				LocationViewPtr locationView(new LocationView(locationModel, m_terrainController, App::Inst().GetViewport()->GetCamera(), UniqueId::Inst().GenerateId()));
				ChartViewPtr pieChartView(new ChartView(locationModel, locationView, newColourMap));
				chartSetCtrl->AddChart(pieChartView);
				LocationControllerPtr locationController(new LocationController(locationModel, locationView, pieChartView));

				LocationLayerPtr locationLayer(new LocationLayer(UniqueId::Inst().GenerateId(), locationSet, locationController));			
				locationLayer->SetName(locationController->GetId());
				locationLayer->SetPath(path);
				locationLayer->SetFilename(filename);

				locationView->SetLayer(locationLayer);
				pieChartView->SetLayer(locationLayer);

				locationSet->AddLocation(locationLayer);
			}
			App::Inst().GetLayerTreeController()->AddLocationSet(locationSet);

			App::Inst().GetLayerTreeController()->GetTreeCtrl()->Thaw();

			// Update Samples legend on sidebar:
			//FillSamplesLegend();
		}
		else
		{
			wxMessageBox(wxT("Failed to read location sites from file. Check console window for warning messages."), 
											wxT("Failed to read file"), wxOK | wxICON_INFORMATION);
		}
	}
	else
	{
		wxMessageBox(wxT("Please select a map node before adding a location sites layer."),
												wxT("Select map node"), wxOK | wxICON_INFORMATION);
	}
}

void GenGisFrame::OnLayerOpenTree( wxCommandEvent& event )
{
	wxFileDialog* openDlg = new wxFileDialog(this, wxT("Open tree file..."), wxT(""), wxT(""), wxT("*.*"), 
																							wxFD_OPEN | wxFD_FILE_MUST_EXIST | wxFD_CHANGE_DIR);

	wxString wildcards = _T("Geographic tree model (*.gtm)|*.gtm|All files (*.*)|*.*");
 	openDlg->SetWildcard(wildcards);

  if(openDlg->ShowModal() != wxID_OK)
  {
		openDlg->Destroy();
    return;
  }

	std::wstring filename = openDlg->GetFilename().wc_str(*wxConvCurrent);
	std::wstring path = openDlg->GetDirectory().wc_str(*wxConvCurrent);

	LayerOpenTree(filename, path);
  
	openDlg->Destroy();
}

void GenGisFrame::LayerOpenTree(std::string file)
{
	std::string path, filename;
	util::StringTools::ExtractPathAndFilename(file, path, filename);  
	LayerOpenTree(util::StringTools::ToStringW(filename), util::StringTools::ToStringW(path));
}

void GenGisFrame::LayerOpenTree(std::wstring filename, std::wstring path)
{
	// determine the selected layer
	LayerPtr selectedLayer = App::Inst().GetLayerTreeController()->GetSelectedLayer();
	if(selectedLayer != LayerPtr() && selectedLayer->GetType() == Layer::MAP)
	{
		GeoTreeIO geoTreeIO;
		Tree<NodeGeoTree>::Ptr geoTree(new Tree<NodeGeoTree>());

		std::string fullPath;
		if(path != _T(""))
			fullPath = util::StringTools::ToString(path) + std::string("\\") + util::StringTools::ToString(filename);
		else
			fullPath = util::StringTools::ToString(filename);

		bool bSuccess = geoTreeIO.Read(geoTree, fullPath);

		if(bSuccess)
		{
			GeoTreeViewPtr geoTreeView(new GeoTreeView(geoTree, m_terrainController, selectedLayer));
			geoTreeView->SetLayout(GeoTreeView::SLANTED_CLADOGRAM_3D);
			
			GeoTreeControllerPtr geoTreeController(new GeoTreeController(geoTreeView));

			TreeLayerPtr treeLayer(new TreeLayer(UniqueId::Inst().GenerateId(), selectedLayer, geoTreeController));
			treeLayer->SetName(filename.substr(0, filename.rfind(wxT("."), filename.length())));
			treeLayer->SetPath(path);
			treeLayer->SetFilename(filename);

			geoTreeView->SetLayer(treeLayer);
			
			App::Inst().GetLayerTreeController()->AddTree(treeLayer);

			geoTreeView->SignalNodeUpdate(boost::bind(&GenGisFrame::SigNodeUpdate, this, _1));
		}
		else
		{
			wxMessageBox(wxT("Failed to read geographic tree model from file."),
												wxT("Failed to read file"), wxOK | wxICON_INFORMATION);
		}
	}
	else
	{
		wxMessageBox(wxT("Please select a map node before adding a tree layer."),
												wxT("Select map node"), wxOK | wxICON_INFORMATION);
	}
}

void GenGisFrame::OnLayerOpenSequenceData( wxCommandEvent& event )
{
	wxFileDialog* openDlg = new wxFileDialog(this, wxT("Open sequence data file..."), wxT(""), wxT(""), wxT("*.*"), 
																							wxFD_OPEN | wxFD_FILE_MUST_EXIST | wxFD_CHANGE_DIR);

	wxString wildcards = _T("Sequence data file (*.csv)|*.csv| All files (*.*)|*.*");
 	openDlg->SetWildcard(wildcards);

  if(openDlg->ShowModal() != wxID_OK)
  {
		openDlg->Destroy();
    return;
  }

	std::wstring filename = openDlg->GetFilename().wc_str(*wxConvCurrent);
	std::wstring path = openDlg->GetDirectory().wc_str(*wxConvCurrent);

	openDlg->Destroy();
  
  LayerOpenSequenceData(filename, path);
}

void GenGisFrame::LayerOpenSequenceData(std::string file)
{
	std::string path, filename;
	util::StringTools::ExtractPathAndFilename(file, path, filename);  
	LayerOpenSequenceData(util::StringTools::ToStringW(filename), util::StringTools::ToStringW(path));
}

void GenGisFrame::LayerOpenSequenceData(std::wstring filename, std::wstring path)
{
	// determine the selected layer
	LayerPtr selectedLayer = App::Inst().GetLayerTreeController()->GetSelectedLayer();
	if(selectedLayer != LayerPtr() && selectedLayer->GetType() == Layer::LOCATION_SET)
	{
		std::vector<SequenceModelPtr> sequenceModels;
		if(SequenceIO::Read(path, filename, sequenceModels))
		{	
			// stop refreshing the tree until all sequences are loaded (this is purely for efficency)
			App::Inst().GetLayerTreeController()->GetTreeCtrl()->Freeze();

			std::set<std::wstring> missingLocations;
			foreach(SequenceModelPtr sequenceModel, sequenceModels)
			{
				SequenceControllerPtr sequenceController(new SequenceController(sequenceModel));

				SequenceLayerPtr sequenceLayer(new SequenceLayer(UniqueId::Inst().GenerateId(), App::Inst().GetLayerTreeController()->GetSelectedLayer(), sequenceController));
				sequenceLayer->SetName(sequenceController->GetSequenceId());
				sequenceLayer->SetPath(path);
				sequenceLayer->SetFilename(filename);

				if(!App::Inst().GetLayerTreeController()->AddSequence(sequenceLayer, sequenceModel, missingLocations))
				{			
					// The above function should never return false!
					wxMessageBox(wxT("Critial error: Failed to load sequence data. Please report this error to the GenGIS team."), wxT("Failed to load sequence data."), wxOK | wxICON_ERROR);
					return;
				}
			}

			// report any locations that are specified in the sequence file, but do not exist
			foreach(std::wstring locationStr, missingLocations)
			{
				error::Log::Inst().Warning(wxT("(Warning) Site id '") + wxString(locationStr.c_str()) + wxT("' was not found in the location set."));
			}

			// all tree to refresh
			App::Inst().GetLayerTreeController()->GetTreeCtrl()->Thaw();
		}
		else
		{
			wxMessageBox(wxT("Failed to read sequence data from file. Check console window for warning messages."), wxT("Failed to read file"), wxOK | wxICON_INFORMATION);
		}
	}
	else
	{
		wxMessageBox(wxT("Please select a location set node before adding sequence data."), wxT("Select location set node"), wxOK | wxICON_INFORMATION);
	}

	App::Inst().GetViewport()->Refresh(false);
}

void GenGisFrame::OnFileExit(wxCommandEvent& event)
{
	Close(true);
}

void GenGisFrame::OnViewReset( wxCommandEvent& event )
{
  ViewReset();
}

void GenGisFrame::ViewReset()
{ 
	if(m_terrainController)
	{
		App::Inst().GetViewport()->GetCamera()->ResetView(); 
		m_widgetController->SetCompassAngle(App::Inst().GetViewport()->GetCamera()->GetYaw());	
		m_terrainController->GetTerrainView()->RefreshQuadtree();
		App::Inst().GetViewport()->Refresh(false);
	}
}

void GenGisFrame::ViewTop()
{ 
	if(m_terrainController)
	{
		App::Inst().GetViewport()->GetCamera()->TopView(); 
		m_widgetController->SetCompassAngle(App::Inst().GetViewport()->GetCamera()->GetYaw());
		m_terrainController->GetTerrainView()->RefreshQuadtree();
		App::Inst().GetViewport()->Refresh(false); 
	}
}

void GenGisFrame::OnViewTop( wxCommandEvent& event )
{
  ViewTop();
}

void GenGisFrame::ViewSidePanel()
{
	if(m_splitterVertical->IsSplit())
	{
		m_splitterVertical->Unsplit(m_pnlSidebar);
		m_verSplitterPos = m_splitterVertical->GetSashPosition();
		m_mnuView->Check(ID_MNU_VIEW_SIDE_PANEL, false);
	}
	else
	{
		// m_splitterHorizontal->GetSashPosition() doesn't work in v2.9 of wxWidgets
		// so just set the side panel to a default size
		m_splitterVertical->SplitVertically(m_pnlSidebar, m_pnlViewport, 200);	//m_verSplitterPos);
		m_mnuView->Check(ID_MNU_VIEW_SIDE_PANEL, true);
	}
}

void GenGisFrame::OnViewSidePanel( wxCommandEvent& event )
{
  ViewSidePanel();
}

void GenGisFrame::ViewConsole()
{
	if(m_splitterHorizontal->IsSplit())
	{
		m_splitterHorizontal->Unsplit(m_pnlBottom);
		m_horSplitterPos = m_splitterHorizontal->GetSashPosition();
		m_mnuView->Check(ID_MNU_VIEW_CONSOLE, false);
	}
	else
	{
		// m_splitterHorizontal->GetSashPosition() doesn't work in v2.9 of wxWidgets
		// so just set the console to a default size
		m_splitterHorizontal->SplitHorizontally(m_pnlTop, m_pnlBottom, App::Inst().GetViewport()->GetCamera()->GetHeight() - 200); //m_horSplitterPos);
		m_mnuView->Check(ID_MNU_VIEW_CONSOLE, true);
	}
}

void GenGisFrame::OnViewConsole(wxCommandEvent& event)
{
  ViewConsole();
}

void GenGisFrame::OnViewCompass( wxCommandEvent& event )
{
	m_widgetController->ToggleCompassVisibility();
}

void GenGisFrame::OnViewMiniMap( wxCommandEvent& event )
{
	m_widgetController->ToggleMiniMapVisibility();
}


void GenGisFrame::OnViewNav( wxCommandEvent& event )
{
	m_widgetController->ToggleNavVisibility();
}

void GenGisFrame::OnViewHideWidgets( wxCommandEvent& event )
{
	m_widgetController->SetCompassVisibility(false);
	m_widgetController->SetMiniMapVisibility(false);
	m_widgetController->SetNavVisibility(false);

	m_menubar->FindItem(ID_MNU_VIEW_COMPASS)->Check(false);
	m_menubar->FindItem(ID_MNU_VIEW_MINIMAP)->Check(false);
	m_menubar->FindItem(ID_MNU_VIEW_NAV)->Check(false);
}

void GenGisFrame::OnViewShowWidgets( wxCommandEvent& event )
{
	m_widgetController->SetCompassVisibility(true);
	m_widgetController->SetMiniMapVisibility(true);
	m_widgetController->SetNavVisibility(true);

	m_menubar->FindItem(ID_MNU_VIEW_COMPASS)->Check(true);
	m_menubar->FindItem(ID_MNU_VIEW_MINIMAP)->Check(true);
	m_menubar->FindItem(ID_MNU_VIEW_NAV)->Check(true);
}


void GenGisFrame::OnViewFine( wxCommandEvent& event )
{
  ViewFine();
}

void GenGisFrame::ViewFine()
{
	m_terrainController->GetTerrainView()->IncEpsilon(-0.0001f);
	App::Inst().GetViewport()->Refresh(false);
}

void GenGisFrame::OnViewCoarse( wxCommandEvent& event )
{
  ViewCoarse();
}

void GenGisFrame::ViewCoarse()
{
	m_terrainController->GetTerrainView()->IncEpsilon( 0.0001f );
	App::Inst().GetViewport()->Refresh(false);
}

void GenGisFrame::OnViewSelectSeqDlg( wxCommandEvent& event )
{
  ViewSelectSegDlg();
}

void GenGisFrame::ViewSelectSegDlg()
{
	SeqSelectionDlg* dlg = new SeqSelectionDlg(App::Inst().GetMainWindow(), m_locationSetLayer);
	dlg->Show();	
}

void GenGisFrame::OnLightingSettingsDlg( wxCommandEvent& event )
{
	LightingPropertiesDlg* dlg = new LightingPropertiesDlg(this);
  dlg->Show();
}

void GenGisFrame::OnFontSettingsDlg( wxCommandEvent& event )
{
	FontPropertiesDlg* dlg = new FontPropertiesDlg(this);
  dlg->Show();
}

void GenGisFrame::OnLayoutObjectSettingsDlg( wxCommandEvent& event )
{
	LayoutObjectPropertiesDlg* dlg = new LayoutObjectPropertiesDlg(this);
  dlg->Show();
}

void GenGisFrame::OnHelpAbout( wxCommandEvent& event )
{
	wxAboutDialogInfo info;
	info.SetName(wxT("GenGIS"));
	info.SetVersion(wxT("v1.08"));
	info.SetCopyright(wxT("Copyright (c) 2009, Robert Beiko (beiko@cs.dal.ca)"));
	info.SetDescription(wxT("An application for merging geographic data with information about biological sequences collected from the environment."));

	info.AddDeveloper(wxT("Donovan Parks"));
	info.AddDeveloper(wxT("Michael Porter"));
	info.AddDeveloper(wxT("Suwen Wang"));
	info.AddDeveloper(wxT("Harman Clair"));
	info.AddDeveloper(wxT("Gregory Smolyn"));
	info.AddDeveloper(wxT("Christian Blouin"));
	info.AddDeveloper(wxT("Jacqueline Whalley"));
	info.AddDeveloper(wxT("Stephen Brooks"));
	info.AddDeveloper(wxT("Robert Beiko (beiko@cs.dal.ca)"));

	info.AddDocWriter(wxT("Robert Beiko (beiko@cs.dal.ca)"));
	info.AddDocWriter(wxT("Donovan Parks"));

	info.SetIcon(wxIcon(App::Inst().GetExeDir() + wxT("images/CrazyEye.ico"), wxBITMAP_TYPE_ICO));
	info.SetLicence(wxT("GenGIS is free software: you can redistribute it and/or modify")
									wxT(" it under the terms of the GNU General Public License as published by")
									wxT(" the Free Software Foundation, either version 3 of the License, or")
									wxT(" (at your option) any later version.")
									wxT("\n\n")
									wxT("GenGIS is distributed in the hope that it will be useful,")
									wxT(" but WITHOUT ANY WARRANTY; without even the implied warranty of")
									wxT(" MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the")
									wxT(" GNU General Public License for more details.")
									wxT("\n\n")
									wxT("You should have received a copy of the GNU General Public License")
									wxT(" along with GenGIS.  If not, see <http://www.gnu.org/licenses/>."));
	info.SetWebSite(wxT("http://kiwi.cs.dal.ca/GenGIS"));
	info.AddArtist(wxT("Crazy Eye icon designed by: Fasticon.com"));

	::wxAboutBox(info);
}



void GenGisFrame::OnModeDrawLayoutLine( wxCommandEvent& event ) 
{ 
	if(!m_layoutLine)
	{
		App::Inst().SetMode(App::LAYOUT_LINE); 
		m_layoutLine.reset(new LayoutLine(m_terrainController));
		App::Inst().SetSelectedView(m_layoutLine);
	}
	else
	{
		wxMessageBox(wxT("Can only have a single unassigned layout line active at a time."),
												wxT("Multiple layout lines"), wxOK | wxICON_INFORMATION);	
		m_toolBar->ToggleTool(ID_TB_MODE_DRAW_LAYOUT_LINE, false);	
	}
}

void GenGisFrame::OnModeDrawLayoutEllipse( wxCommandEvent& event ) 
{ 
	if(!m_layoutEllipse)
	{
		App::Inst().SetMode(App::LAYOUT_ELLIPSE); 
		m_layoutEllipse.reset(new LayoutEllipse(m_terrainController));
		App::Inst().SetSelectedView(m_layoutEllipse);
	}
	else
	{
		wxMessageBox(wxT("Can only have a single unassigned layout ellipse active at a time."),
												wxT("Multiple layout ellipses"), wxOK | wxICON_INFORMATION);	
		m_toolBar->ToggleTool(ID_TB_MODE_DRAW_LAYOUT_ELLIPSE, false);	
	}
}

void GenGisFrame::OnGeoAxisByPolyline( wxCommandEvent& event ) 
{ 
	if(!m_geoAxisPolyline)
	{
		App::Inst().SetMode(App::GEO_AXIS_POLYLINE); 
		m_geoAxisPolyline.reset(new GeoAxisPolyline(m_terrainController));
		App::Inst().SetSelectedView(m_geoAxisPolyline);
	}
	else
	{
		wxMessageBox(wxT("Can only have a single unassigned geographic axis polyline active at a time."),
												wxT("Multiple geographic polylines"), wxOK | wxICON_INFORMATION);	
		m_toolBar->ToggleTool(ID_TB_GEO_AXIS_BY_POLYLINE, false);	
	}
}

void GenGisFrame::SigRender()
{	
  App::Inst().GetLayerTreeController()->OnRender();

	UserElements::Inst().Render();
	LabelController::Inst().Render();

	if(m_layoutEllipse)
		m_layoutEllipse->Render();

	if(m_layoutLine)
		m_layoutLine->Render();

	if(m_geoAxisPolyline)
		m_geoAxisPolyline->Render();
}

void GenGisFrame::SigViewportKeyboard( wxKeyEvent& event )
{
	if(m_layoutLine && App::Inst().GetSelectedView() == m_layoutLine)
	{
		if(event.GetKeyCode() == WXK_ESCAPE)
		{
			if(App::Inst().GetMode() == App::LAYOUT_LINE)
			{
				wxCommandEvent dummy;
				OnLayoutLineDelete(dummy);
			}
		}
		if(event.GetKeyCode() == WXK_DELETE)
		{
			wxCommandEvent dummy;
			OnLayoutLineDelete(dummy);
		}
	}
	else if(m_layoutEllipse && App::Inst().GetSelectedView() == m_layoutEllipse)
	{
		m_layoutEllipse->KeyboardEvent(event);
		if(event.GetKeyCode() == WXK_ESCAPE)
		{
			if(App::Inst().GetMode() == App::LAYOUT_ELLIPSE)
			{
				wxCommandEvent dummy;
				OnLayoutEllipseDelete(dummy);
			}
		}
		if(event.GetKeyCode() == WXK_DELETE)
		{
			wxCommandEvent dummy;
			OnLayoutEllipseDelete(dummy);
		}
	}
	else if(m_geoAxisPolyline && App::Inst().GetSelectedView() == m_geoAxisPolyline)
	{
		if(event.GetKeyCode() == WXK_ESCAPE)
		{
			if(App::Inst().GetMode() == App::GEO_AXIS_POLYLINE)
			{
				wxCommandEvent dummy;
				OnGeoAxisPolylineDelete(dummy);
			}
		}
		if(event.GetKeyCode() == WXK_DELETE)
		{
			wxCommandEvent dummy;
			OnGeoAxisPolylineDelete(dummy);
		}
		if(event.GetKeyCode() == WXK_RETURN)
		{
			App::Inst().SetMode(App::STANDARD);
			m_toolBar->ToggleTool(ID_TB_GEO_AXIS_BY_POLYLINE, false);
			m_geoAxisPolyline->PolylineFinished();
		}
	}
	else if(App::Inst().GetMode() == App::USER_LABEL)
	{
		if(event.GetKeyCode() == WXK_ESCAPE)
		{
			App::Inst().SetMode(App::STANDARD);
			m_toolBar->ToggleTool(ID_TB_LABEL_TOOL, false);
		}
	}
	else if(App::Inst().GetSelectedView())
	{
		App::Inst().GetSelectedView()->KeyboardEvent(event);
	}
	else    // If there is no selected view, then the ViewPtr must be the TerrainView ViewPtr.
	{
    if(event.ShiftDown())    // Holding Shift down when using arrow keys zooms the map
    {
      if(event.GetKeyCode() == WXK_UP)    // Zoom in
      {
      	App::Inst().GetViewport()->GetCamera()->ZoomDelta(-0.025f);
      	App::Inst().GetViewport()->Refresh(false);
      }
      else if(event.GetKeyCode() == WXK_DOWN)
      {
      	App::Inst().GetViewport()->GetCamera()->ZoomDelta(0.025f);
      	App::Inst().GetViewport()->Refresh(false);
      }
    }
    else if(event.ControlDown())    // Control key rotates the map
    {
      if(event.GetKeyCode() == WXK_UP)
      {
        event.Skip();
  			App::Inst().GetViewport()->GetCamera()->PitchDelta(-1.0f);
  			App::Inst().GetViewport()->Refresh(false);
      }
      else if(event.GetKeyCode() == WXK_DOWN)
      {
      	event.Skip();
        App::Inst().GetViewport()->GetCamera()->PitchDelta(1.0f);
  			App::Inst().GetViewport()->Refresh(false);
      }
      else if(event.GetKeyCode() == WXK_LEFT)
      {
      	event.Skip();
  			App::Inst().GetViewport()->GetCamera()->YawDelta(1.0f);
  			m_widgetController->SetCompassAngle(App::Inst().GetViewport()->GetCamera()->GetYaw());
  			App::Inst().GetViewport()->Refresh(false);
      }
      else if(event.GetKeyCode() == WXK_RIGHT)
      {
      	event.Skip();
  			App::Inst().GetViewport()->GetCamera()->YawDelta(-1.0f);
  			m_widgetController->SetCompassAngle(App::Inst().GetViewport()->GetCamera()->GetYaw());
  			App::Inst().GetViewport()->Refresh(false);
      }
    }
    else    // No modifier key is held down, so the arrow keys pan the map
    {
      if(event.GetKeyCode() == WXK_UP)
      {
        event.Skip();
				App::Inst().GetViewport()->GetCamera()->TranslateDelta(0, -0.003f);
				App::Inst().GetViewport()->Refresh(false);
      }
      else if(event.GetKeyCode() == WXK_DOWN)
      {
      	event.Skip();
				App::Inst().GetViewport()->GetCamera()->TranslateDelta(0, 0.003f);
				App::Inst().GetViewport()->Refresh(false);
      }
      else if(event.GetKeyCode() == WXK_LEFT)
      {
      	event.Skip();
				App::Inst().GetViewport()->GetCamera()->TranslateDelta(-0.003f, 0);
				App::Inst().GetViewport()->Refresh(false);
      }
      else if(event.GetKeyCode() == WXK_RIGHT)
      {
      	event.Skip();
				App::Inst().GetViewport()->GetCamera()->TranslateDelta(0.003f, 0);
				App::Inst().GetViewport()->Refresh(false);
      }
    }
	}

	App::Inst().GetViewport()->Refresh(false);

	event.Skip();
}

void GenGisFrame::SigNodeUpdate(NodeInfo nodeInfo)
{
	if(nodeInfo.crossings == NOT_SET)
	{
		m_statusBar->SetStatusText(_T(""), 1);
	}
	else if(util::StringTools::IsEmpty(nodeInfo.name))
	{
		m_statusBar->SetStatusText(_T("Crossings below subtree: ") + wxString(util::StringTools::ToStringW(nodeInfo.crossings).c_str()), 1);
	}
	else
	{
		m_statusBar->SetStatusText(_T("Crossings below ") + wxString(nodeInfo.name.c_str()) + _T(": ") + wxString(util::StringTools::ToStringW(nodeInfo.crossings).c_str()), 1);
	}
}

void GenGisFrame::OnLayoutLineDelete(wxCommandEvent& event)
{
	if(App::Inst().GetSelectedView() == m_layoutLine)
		App::Inst().SetSelectedView(ViewPtr());

	App::Inst().SetMode(App::STANDARD);

	m_layoutLine.reset();
	m_toolBar->ToggleTool(ID_TB_MODE_DRAW_LAYOUT_LINE, false);	

	App::Inst().GetViewport()->Refresh(false);
}


void GenGisFrame::OnLayoutEllipseDelete(wxCommandEvent& event)
{
	if(App::Inst().GetSelectedView() == m_layoutEllipse)
		App::Inst().SetSelectedView(ViewPtr());

	App::Inst().SetMode(App::STANDARD);

	m_layoutEllipse.reset();
	m_toolBar->ToggleTool(ID_TB_MODE_DRAW_LAYOUT_ELLIPSE, false);

	App::Inst().GetViewport()->Refresh(false);
}

void GenGisFrame::OnGeoAxisPolylineDelete(wxCommandEvent& event)
{
	if(App::Inst().GetSelectedView() == m_geoAxisPolyline)
		App::Inst().SetSelectedView(ViewPtr());

	App::Inst().SetMode(App::STANDARD);

	m_geoAxisPolyline.reset();
	m_toolBar->ToggleTool(ID_TB_GEO_AXIS_BY_POLYLINE, false);	

	App::Inst().GetViewport()->Refresh(false);
}

void GenGisFrame::OnSelectedViewHide(wxCommandEvent& event)
{
	App::Inst().GetRightClickView()->SetVisibility(false);
	App::Inst().GetViewport()->Refresh();
}

void GenGisFrame::OnSelectedViewProperties(wxCommandEvent& event)
{
	App::Inst().GetRightClickView()->ShowPropertiesDlg();
	App::Inst().GetViewport()->Refresh();
}

void GenGisFrame::OnSelectedViewRemove(wxCommandEvent& event)
{
	ViewPtr view = App::Inst().GetRightClickView();
	VisualLabelPtr visualLabel = boost::dynamic_pointer_cast<VisualLabel>(view);
	if(visualLabel)
	{
		LabelController::Inst().RemoveLabel(visualLabel->GetId());
		App::Inst().GetViewport()->Refresh();
	}
}

void GenGisFrame::OnLayerRemove( wxCommandEvent& event ) 
{ 
	LayerPtr removedLayer;
	App::Inst().GetLayerTreeController()->OnLayerRemove(event, removedLayer); 

	if(removedLayer->GetType() == Layer::MAP)
	{
		InitializeTerrainController();
		m_widgetController->SetMiniMapTerrain(m_terrainController);
	}
}

void GenGisFrame::OnLayerHideAll( wxCommandEvent& event ) 
{ 
	App::Inst().GetLayerTreeController()->OnHideAll(event); 
}

void GenGisFrame::OnLayerShowAll( wxCommandEvent& event ) 
{ 
	App::Inst().GetLayerTreeController()->OnShowAll(event); 
}

void GenGisFrame::OnLayerTreeSelect( wxTreeEvent& event ) 
{ 
	App::Inst().GetLayerTreeController()->OnSelect(event); 
}

void GenGisFrame::OnLayerTreeClick( wxMouseEvent& event ) 
{ 
	App::Inst().GetLayerTreeController()->OnClick(event); 
}

void GenGisFrame::OnLayerTreeLeftDClick( wxMouseEvent& event )
{  
	App::Inst().GetLayerTreeController()->OnLeftDClick(event); 
}

void GenGisFrame::OnModeStandard( wxCommandEvent& event ) 
{ 
	App::Inst().SetMode(App::STANDARD); 
}

void GenGisFrame::OnStudyProperties(wxCommandEvent& event)
{
	StudyPropertiesDlg* dlg = new StudyPropertiesDlg(App::Inst().GetMainWindow(), m_studyLayer);
  dlg->Show();

	App::Inst().GetLayerTreeController()->SetName(m_studyLayer->GetId(), m_studyLayer->GetName());
}

void GenGisFrame::OnMapProperties(wxCommandEvent& event)
{
	MapPropertiesDlg* dlg = new MapPropertiesDlg(App::Inst().GetMainWindow(), m_mapLayer);
  dlg->Show();

	App::Inst().GetLayerTreeController()->SetName(m_mapLayer->GetId(), m_mapLayer->GetName());
}

void GenGisFrame::OnLocationSetProperties(wxCommandEvent& event)
{
	LocationSetPropertiesDlg* dlg = new LocationSetPropertiesDlg(App::Inst().GetMainWindow(), m_locationSetLayer);
  dlg->Show();

	App::Inst().GetLayerTreeController()->SetName(m_locationSetLayer->GetId(), m_locationSetLayer->GetName());
}

void GenGisFrame::OnLocationProperties(wxCommandEvent& event)
{
	LocationPropertiesDlg* dlg = new LocationPropertiesDlg(App::Inst().GetMainWindow(), m_locationLayer);
  dlg->Show();

	App::Inst().GetLayerTreeController()->SetName(m_locationLayer->GetId(), m_locationLayer->GetName());
}

void GenGisFrame::OnSequenceProperties(wxCommandEvent& event)
{
	SequencePropertiesDlg* dlg = new SequencePropertiesDlg(App::Inst().GetMainWindow(), m_sequenceLayer);
  dlg->Show();

	App::Inst().GetLayerTreeController()->SetName(m_sequenceLayer->GetId(), m_sequenceLayer->GetName());
}

void GenGisFrame::OnTreeProperties(wxCommandEvent& event)
{
	TreePropertiesDlg* dlg = new TreePropertiesDlg(App::Inst().GetMainWindow(), m_treeLayer);
  dlg->Show();

	App::Inst().GetLayerTreeController()->SetName(m_treeLayer->GetId(), m_treeLayer->GetName());
}

void GenGisFrame::OnCladogram3D( wxCommandEvent& event ) 
{ 
	App::Inst().GetLayerTreeController()->OnLayoutTree(GeoTreeView::CLADOGRAM_3D, LayoutLinePtr()); 
}

void GenGisFrame::OnSlantedCladogram3D( wxCommandEvent& event )  
{ 
	App::Inst().GetLayerTreeController()->OnLayoutTree(GeoTreeView::SLANTED_CLADOGRAM_3D, LayoutLinePtr()); 
}

void GenGisFrame::OnCladogram2D( wxCommandEvent& event ) 
{ 
	if(App::Inst().GetLayerTreeController()->OnLayoutTree(GeoTreeView::CLADOGRAM_2D, m_layoutLine))
	{
		m_layoutLine.reset();
	}
}

void GenGisFrame::OnPhylogram2D( wxCommandEvent& event ) 
{ 
	if(App::Inst().GetLayerTreeController()->OnLayoutTree(GeoTreeView::PHYLOGRAM_2D, m_layoutLine))
	{
		m_layoutLine.reset();
	}
}

void GenGisFrame::OnSlantedCladogram2D( wxCommandEvent& event )  
{ 
	if(App::Inst().GetLayerTreeController()->OnLayoutTree(GeoTreeView::SLANTED_CLADOGRAM_2D, m_layoutLine))
	{
		m_layoutLine.reset();
	}
}

void GenGisFrame::OnLayoutTreeEllipseQuadrant( wxCommandEvent& event )
{
	if(App::Inst().GetLayerTreeController()->OnLayoutTree(GeoTreeView::ELLIPSE_QUADRANT, m_layoutEllipse))
	{
		m_layoutEllipse.reset();
	}
}

void GenGisFrame::OnTreeLayoutLineVisibility( wxCommandEvent& event )  
{ 
	App::Inst().GetLayerTreeController()->OnTreeLayoutLineVisibility(); 
}

void GenGisFrame::OnTreeLayoutEllipseVisibility( wxCommandEvent& event )  
{ 
	App::Inst().GetLayerTreeController()->OnTreeLayoutEllipseVisibility(); 
}

void GenGisFrame::OnTreeGeographicAxisVisibility( wxCommandEvent& event )  
{ 
	App::Inst().GetLayerTreeController()->OnTreeGeographicAxisVisibility(); 
}

void GenGisFrame::OnAssignGeoAxisToTree( wxCommandEvent& event )  
{
	if(App::Inst().GetLayerTreeController()->OnAssignGeoAxisToTree(m_geoAxisPolyline))
	{
		m_geoAxisPolyline.reset();
	}
}

void GenGisFrame::OnReverseGeoAxisOrder( wxCommandEvent& event )  
{
	App::Inst().GetLayerTreeController()->OnReverseGeoAxisOrder(event.IsChecked());
	App::Inst().GetViewport()->Refresh(false);
}

void GenGisFrame::OnLeafNodeLabels( wxCommandEvent& event )  
{ 
	App::Inst().GetLayerTreeController()->OnLeafNodeLabels(); 
}

void GenGisFrame::OnLayoutChartsEvenlyOnLine(wxCommandEvent& event)
{
	if(App::Inst().GetLayerTreeController()->OnLayoutCharts(ChartSetView::LAYOUT_LINE_EVEN_SPACING, m_layoutLine))
	{
		m_layoutLine.reset();
	}
}

void GenGisFrame::OnLayoutChartsLineProjected(wxCommandEvent& event)
{
	if(App::Inst().GetLayerTreeController()->OnLayoutCharts(ChartSetView::LAYOUT_LINE_PROJECTED, m_layoutLine))
	{
		m_layoutLine.reset();
	}
}

void GenGisFrame::OnLayoutChartsEllipseQuadrant(wxCommandEvent& event)
{
	if(App::Inst().GetLayerTreeController()->OnLayoutCharts(ChartSetView::ELLIPSE_QUADRANT, m_layoutEllipse))
	{
		m_layoutEllipse.reset();
	}
}

void GenGisFrame::OnLayoutChartsEvenlyOnEllipse(wxCommandEvent& event)
{
	if(App::Inst().GetLayerTreeController()->OnLayoutCharts(ChartSetView::ELLIPSE_EVEN_SPACING, m_layoutEllipse))
	{
		m_layoutEllipse.reset();
	}
}

void GenGisFrame::OnLayoutChartsEllipseProjected(wxCommandEvent& event)
{
	if(App::Inst().GetLayerTreeController()->OnLayoutCharts(ChartSetView::ELLIPSE_CLOSEST_PT, m_layoutEllipse))
	{
		m_layoutEllipse.reset();
	}
}

void GenGisFrame::OnChartSetLayoutLine( wxCommandEvent& event )  
{ 
	App::Inst().GetLayerTreeController()->OnChartSetLayoutLine(); 
}

void GenGisFrame::OnChartSetLayoutEllipse( wxCommandEvent& event )  
{ 
	App::Inst().GetLayerTreeController()->OnChartSetLayoutEllipse(); 
}

void GenGisFrame::OnTreeSignificanceTest( wxCommandEvent& event )
{
	std::map<uint, uint> pdf;
	double pValue = m_treeLayer->GetGeoTreeController()->GetTreeView()->SignificanceTest(1000, pdf);
	error::Log::Inst().Write("Random permutation test (1000 iterations), p-value = " + util::StringTools::ToString(pValue));
}

void GenGisFrame::OnSubTreeSignificanceTest( wxCommandEvent& event )
{
	std::map<uint, uint> pdf;
	double pValue = m_treeLayer->GetGeoTreeController()->GetTreeView()->SignificanceTestSubtree(1000, pdf);
	error::Log::Inst().Write("Random permutation test (1000 iterations), p-value for subtree = " + util::StringTools::ToString(pValue));
}

void GenGisFrame::OnExtendGeographicPolyline( wxCommandEvent& event )
{
	if(App::Inst().GetRightClickView() == m_geoAxisPolyline)
	{
		App::Inst().SetMode(App::GEO_AXIS_POLYLINE); 
		m_geoAxisPolyline->ExtendGeographicAxis();
		App::Inst().SetSelectedView(m_geoAxisPolyline);
	}
	else
	{
		m_treeLayer->GetGeoTreeController()->GetTreeView()->ExtendGeographicAxis();
		App::Inst().SetSelectedView(m_treeLayer->GetGeoTreeController()->GetTreeView());
	}
}

void GenGisFrame::OnLayerTreePopupMenu( wxTreeEvent& event )
{
	event.Skip();

	// cause the tree item under the cursor to become selected
	wxTreeItemId id = event.GetItem();
	if(id.IsOk())
	{
		App::Inst().GetLayerTreeController()->GetTreeCtrl()->SelectItem(id);

		LayerTreeItem* layerItem = (LayerTreeItem*)App::Inst().GetLayerTreeController()->GetTreeCtrl()->GetItemData(id);
		LayerPtr layer = layerItem->GetLayer();

		wxMenu menu;
		if(layer->GetType() == Layer::STUDY)
		{
			m_studyLayer = boost::dynamic_pointer_cast<StudyLayer>(layer);
			menu.Append(ID_MNU_LAYER_STUDY_PROPERTIES, wxT("Properties"));
		}
		else if(layer->GetType() == Layer::MAP)
		{
			m_mapLayer = boost::dynamic_pointer_cast<MapLayer>(layer);

			menu.Append(ID_MNU_LAYER_MAP_PROPERTIES, wxT("Properties"));	
		}
		else if(layer->GetType() == Layer::LOCATION_SET)
		{
			m_locationSetLayer = boost::dynamic_pointer_cast<LocationSetLayer>(layer);
			menu.Append(ID_MNU_LAYER_LOCATION_SET_PROPERTIES, wxT("Properties"));

			bool bHasSampleData = false;
			for(unsigned int i = 0; i < m_locationSetLayer->GetNumLocations(); ++i)
			{
				if(m_locationSetLayer->GetLocation(i)->GetNumSequences() > 0)
				{
					bHasSampleData = true;
					break;
				}
			}

			if(bHasSampleData)
			{
				menu.AppendSeparator();

				menu.Append(ID_RIGHT_CLICK_LAYOUT_CHARTS_EVENLY_ON_LINE, wxT("Layout charts evenly on line"), wxT(""), wxITEM_CHECK);
				menu.Append(ID_RIGHT_CLICK_LAYOUT_CHARTS_LINE_PROJECTED, wxT("Project charts onto layout line"), wxT(""), wxITEM_CHECK);
				menu.AppendSeparator();
				menu.Append(ID_RIGHT_CLICK_LAYOUT_CHARTS_ELLIPSE_QUADRANT, wxT("Per ellipse quadrant layout (recommended)"), wxT(""), wxITEM_CHECK);
				menu.Append(ID_RIGHT_CLICK_LAYOUT_CHARTS_EVENLY_ON_ELLIPSE, wxT("Layout charts evenly on ellipse"), wxT(""), wxITEM_CHECK);
				menu.Append(ID_RIGHT_CLICK_LAYOUT_CHARTS_ELLIPSE_PROJECTED, wxT("Project charts onto layout ellipse"), wxT(""), wxITEM_CHECK);
		
				if(m_locationSetLayer->GetChartSetView()->IsLayoutLine())
				{
					menu.AppendSeparator();
					menu.Append(ID_RIGHT_CLICK_CHART_SET_LAYOUT_LINE, wxT("Show layout line"), wxT(""), wxITEM_CHECK);
					menu.Check(ID_RIGHT_CLICK_CHART_SET_LAYOUT_LINE, m_locationSetLayer->GetChartSetView()->GetLayoutLineVisibility());
				}
				else if(m_locationSetLayer->GetChartSetView()->IsLayoutEllipse())
				{
					menu.AppendSeparator();
					menu.Append(ID_RIGHT_CLICK_CHART_SET_LAYOUT_ELLIPSE, wxT("Show layout ellipse"), wxT(""), wxITEM_CHECK);
					menu.Check(ID_RIGHT_CLICK_CHART_SET_LAYOUT_ELLIPSE, m_locationSetLayer->GetChartSetView()->GetLayoutEllipseVisibility());
				}

				// mark current layout style in menu
				if(m_locationSetLayer->GetChartSetView()->GetLayout() == ChartSetView::LAYOUT_LINE_EVEN_SPACING)
				{
					menu.Check(ID_RIGHT_CLICK_LAYOUT_CHARTS_EVENLY_ON_LINE, true);
				}
				else if(m_locationSetLayer->GetChartSetView()->GetLayout() == ChartSetView::LAYOUT_LINE_PROJECTED)
				{
					menu.Check(ID_RIGHT_CLICK_LAYOUT_CHARTS_LINE_PROJECTED, true);
				}
				else if(m_locationSetLayer->GetChartSetView()->GetLayout() == ChartSetView::ELLIPSE_CLOSEST_PT)
				{
					menu.Check(ID_RIGHT_CLICK_LAYOUT_CHARTS_ELLIPSE_PROJECTED, true);
				}
				else if(m_locationSetLayer->GetChartSetView()->GetLayout() == ChartSetView::ELLIPSE_EVEN_SPACING)
				{
					menu.Check(ID_RIGHT_CLICK_LAYOUT_CHARTS_EVENLY_ON_ELLIPSE, true);
				}
				else if(m_locationSetLayer->GetChartSetView()->GetLayout() == ChartSetView::ELLIPSE_QUADRANT)
				{
					menu.Check(ID_RIGHT_CLICK_LAYOUT_CHARTS_ELLIPSE_QUADRANT, true);
				}
			}
		}
		else if(layer->GetType() == Layer::LOCATION)
		{
			m_locationLayer = boost::dynamic_pointer_cast<LocationLayer>(layer);
			menu.Append(ID_MNU_LAYER_LOCATION_PROPERTIES, wxT("Properties"));
		}
		else if(layer->GetType() == Layer::SEQUENCE)
		{
			m_sequenceLayer = boost::dynamic_pointer_cast<SequenceLayer>(layer);
			menu.Append(ID_MNU_LAYER_SEQUENCE_PROPERTIES, wxT("Properties"));
		}
		else if(layer->GetType() == Layer::TREE)
		{
			m_treeLayer = boost::dynamic_pointer_cast<TreeLayer>(layer);

			menu.Append(ID_MNU_LAYER_TREE_PROPERTIES, wxT("Properties"));
			menu.AppendSeparator();

			menu.Append(ID_RIGHT_CLICK_SLANTED_CLADOGRAM_3D, wxT("3D slanted cladogram"), wxT(""), wxITEM_CHECK);
			menu.Append(ID_RIGHT_CLICK_CLADOGRAM_3D, wxT("3D cladogram"), wxT(""), wxITEM_CHECK);

			menu.AppendSeparator();
			menu.Append(ID_RIGHT_CLICK_SLANTED_CLADOGRAM_2D, wxT("2D slanted cladogram"), wxT(""), wxITEM_CHECK);
			menu.Append(ID_RIGHT_CLICK_CLADOGRAM_2D, wxT("2D cladogram"), wxT(""), wxITEM_CHECK);
			menu.Append(ID_RIGHT_CLICK_PHYLOGRAM_2D, wxT("2D phylogram"), wxT(""), wxITEM_CHECK);

			// DHP: Being removed until functionality is complete
			//menu.AppendSeparator();
			//menu.Append(ID_RIGHT_CLICK_LAYOUT_ELLIPSE_QUADRANT, wxT("Elliptical layout"), wxT(""), wxITEM_CHECK);

			menu.AppendSeparator();
			menu.Append(ID_RIGHT_CLICK_GEO_AXIS_TO_TREE, wxT("Assign geographical axis to tree"), wxT(""), wxITEM_CHECK);
			menu.Append(ID_RIGHT_CLICK_GEO_AXIS_REVERSE_ORDER, wxT("Reverse order of geographic points"), wxT(""), wxITEM_CHECK);
			menu.Check(ID_RIGHT_CLICK_GEO_AXIS_REVERSE_ORDER, m_treeLayer->GetGeoTreeController()->GetTreeView()->GetReverseOrderGeographicPts());

			menu.AppendSeparator();
			menu.Append(ID_RIGHT_CLICK_TREE_LEAF_NODE_LABELS, wxT("Show leaf node labels"), wxT(""), wxITEM_CHECK);
			menu.Check(ID_RIGHT_CLICK_TREE_LEAF_NODE_LABELS, m_treeLayer->GetGeoTreeController()->GetTreeView()->GetFontVisibility());

			// check visibility of layout primatives
			if(m_treeLayer->GetGeoTreeController()->GetTreeView()->IsLayoutLine())
			{
				menu.AppendSeparator();
				menu.Append(ID_RIGHT_CLICK_TREE_LAYOUT_LINE, wxT("Show layout line"), wxT(""), wxITEM_CHECK);
				menu.Check(ID_RIGHT_CLICK_TREE_LAYOUT_LINE, m_treeLayer->GetGeoTreeController()->GetTreeView()->GetLayoutLineVisibility());
			}
			else if(m_treeLayer->GetGeoTreeController()->GetTreeView()->IsLayoutEllipse())
			{
				menu.AppendSeparator();
				menu.Append(ID_RIGHT_CLICK_TREE_LAYOUT_ELLIPSE, wxT("Show layout ellipse"), wxT(""), wxITEM_CHECK);
				menu.Check(ID_RIGHT_CLICK_TREE_LAYOUT_ELLIPSE, m_treeLayer->GetGeoTreeController()->GetTreeView()->GetLayoutEllipseVisibility());
			}
			
			// check visibility of geographic axis
			if(m_treeLayer->GetGeoTreeController()->GetTreeView()->IsGeographicAxis())
			{
				menu.AppendSeparator();
				menu.Append(ID_RIGHT_CLICK_TREE_GEO_AXIS, wxT("Show geographical axis"), wxT(""), wxITEM_CHECK);
				menu.Check(ID_RIGHT_CLICK_TREE_GEO_AXIS, m_treeLayer->GetGeoTreeController()->GetTreeView()->GetGeographicAxisVisibility());
			}

			// mark current layout style in menu
			if(m_treeLayer->GetGeoTreeController()->GetTreeView()->GetLayout() == GeoTreeView::SLANTED_CLADOGRAM_3D)
			{
				menu.Check(ID_RIGHT_CLICK_SLANTED_CLADOGRAM_3D, true);
			}
			else if(m_treeLayer->GetGeoTreeController()->GetTreeView()->GetLayout() == GeoTreeView::CLADOGRAM_3D)
			{
				menu.Check(ID_RIGHT_CLICK_CLADOGRAM_3D, true);
			}
			else if(m_treeLayer->GetGeoTreeController()->GetTreeView()->GetLayout() == GeoTreeView::SLANTED_CLADOGRAM_2D)
			{
				menu.Check(ID_RIGHT_CLICK_SLANTED_CLADOGRAM_2D, true);
				menu.AppendSeparator();
				menu.Append(ID_RIGHT_CLICK_TREE_SIGNIFICANCE_TEST, wxT("Perform significance test"), wxT(""), wxITEM_CHECK);
			}
			else if(m_treeLayer->GetGeoTreeController()->GetTreeView()->GetLayout() == GeoTreeView::CLADOGRAM_2D)
			{
				menu.Check(ID_RIGHT_CLICK_CLADOGRAM_2D, true);
				menu.AppendSeparator();
				menu.Append(ID_RIGHT_CLICK_TREE_SIGNIFICANCE_TEST, wxT("Perform significance test"), wxT(""), wxITEM_CHECK);
			}
			else if(m_treeLayer->GetGeoTreeController()->GetTreeView()->GetLayout() == GeoTreeView::PHYLOGRAM_2D)
			{
				menu.Check(ID_RIGHT_CLICK_PHYLOGRAM_2D, true);
				menu.AppendSeparator();
				menu.Append(ID_RIGHT_CLICK_TREE_SIGNIFICANCE_TEST, wxT("Perform significance test"), wxT(""), wxITEM_CHECK);
			}
			else if(m_treeLayer->GetGeoTreeController()->GetTreeView()->GetLayout() == GeoTreeView::ELLIPSE_QUADRANT)
			{
				menu.Check(ID_RIGHT_CLICK_LAYOUT_ELLIPSE_QUADRANT, true);
			}
		}

		App::Inst().GetMainWindow()->PopupMenu(&menu);
	}
}

void GenGisFrame::OnLabelTool( wxCommandEvent& event )
{
	App::Inst().SetMode(App::USER_LABEL); 
}

void GenGisFrame::OnConfigureDB( wxCommandEvent& event )
{
	ConfigureDatabaseDlg* dlg = new ConfigureDatabaseDlg(App::Inst().GetMainWindow());
	dlg->Show();	
}

void GenGisFrame::OnObtainDataFromDB( wxCommandEvent& event )
{
	DbSelectSampleDlg* dlg = new DbSelectSampleDlg(App::Inst().GetMainWindow());
	dlg->Show();	
}



bool GenGisFrame::MouseLeftDown(const Point2D& mousePt)
{
	if(App::Inst().GetMode() == App::STANDARD)
	{
		if(m_layoutLine && m_layoutLine->MouseLeftDown(mousePt))
		{
			App::Inst().SetSelectedView(m_layoutLine);
			return true;
		}

		if(m_layoutEllipse && m_layoutEllipse->MouseLeftDown(mousePt))
		{
			App::Inst().SetSelectedView(m_layoutEllipse);
			return true;
		}

		if(m_geoAxisPolyline && m_geoAxisPolyline->MouseLeftDown(mousePt))
		{
			App::Inst().SetSelectedView(m_geoAxisPolyline);
			return true;
		}

		if(m_widgetController->MouseLeftDown(mousePt))
			return true;

		if(App::Inst().GetLayerTreeController()->MouseLeftDown(mousePt))
			return true;

		if(LabelController::Inst().MouseLeftDown(mousePt))
			return true;
	}
	else if(App::Inst().GetMode() == App::LAYOUT_LINE)
	{
		App::Inst().SetSelectedView(m_layoutLine);
		m_layoutLine->MouseLeftDown(mousePt);
		if(m_layoutLine->GetStage() == LayoutLine::COMPLETE)
		{
			App::Inst().SetMode(App::STANDARD);
			m_toolBar->ToggleTool(ID_TB_MODE_DRAW_LAYOUT_LINE, false);
			return false;
		}

		return true;
	}
	else if(App::Inst().GetMode() == App::LAYOUT_ELLIPSE)
	{
		App::Inst().SetSelectedView(m_layoutEllipse);
		m_layoutEllipse->MouseLeftDown(mousePt);
		if(m_layoutEllipse->GetStage() == LayoutEllipse::COMPLETE)
		{
			App::Inst().SetMode(App::STANDARD);
			m_toolBar->ToggleTool(ID_TB_MODE_DRAW_LAYOUT_ELLIPSE, false);
			return false;
		}

		return true;
	}
	else if(App::Inst().GetMode() == App::GEO_AXIS_POLYLINE)
	{
		App::Inst().SetSelectedView(m_geoAxisPolyline);
		m_geoAxisPolyline->MouseLeftDown(mousePt);
		if(m_geoAxisPolyline->GetStage() == GeoAxisPolyline::COMPLETE)
		{
			App::Inst().SetMode(App::STANDARD);
			m_toolBar->ToggleTool(ID_TB_GEO_AXIS_BY_POLYLINE, false);
			return false;
		}

		return true;
	}
	else if(App::Inst().GetMode() == App::USER_LABEL)
	{
		LabelToolDlg* dlg = new LabelToolDlg(this);
		Point3D labelPos = m_terrainController->ProjectToTerrain(App::Inst().GetMouseScreenPos(), false);
		if(dlg->ShowModal())
		{
			if(!dlg->GetText().IsEmpty())
			{
				VisualLabelPtr label(new VisualLabel());

				label->SetText(dlg->GetText().c_str());
				label->SetSize(dlg->GetSize());
				label->SetColour(dlg->GetColour());
				label->SetGridPosition(labelPos);
				label->SetVisibility(true);
				label->SetLabelPosition(_T("<Custom>"));

				if(dlg->GetStyle() == _T("Billboard"))
					label->SetRotateLabel(false);
				else if(dlg->GetStyle() == _T("Dynamic"))
					label->SetRotateLabel(true);

				LabelController::Inst().AddLabel(label);

				App::Inst().GetViewport()->Refresh(false);
			}
		}

		App::Inst().SetMode(App::STANDARD);
		m_toolBar->ToggleTool(ID_TB_LABEL_TOOL, false);

		return false;
	}
	
	// nothing has been selected
	App::Inst().SetSelectedView(ViewPtr());

	return false;
}

void GenGisFrame::MouseLeftUp(const Point2D& mousePt)
{
	if(App::Inst().GetMode() == App::STANDARD && App::Inst().GetSelectedView())
		App::Inst().GetSelectedView()->MouseLeftUp(mousePt);
}

void GenGisFrame::MouseDragging(const Point2D& mousePt)
{
	if(App::Inst().GetMode() == App::STANDARD && App::Inst().GetSelectedView())
		App::Inst().GetSelectedView()->MouseDragging(mousePt);
}

void GenGisFrame::MouseLeftDoubleClick(const Point2D& mousePt)
{
	if(App::Inst().GetMode() == App::STANDARD && App::Inst().GetSelectedView())
		App::Inst().GetSelectedView()->MouseLeftDblClick(mousePt);
}

bool GenGisFrame::MouseRightDown(const Point2D& mousePt)
{
	wxMenu menu;

	if(App::Inst().GetMode() == App::STANDARD)
	{
		if(m_layoutLine && m_layoutLine->MouseRightDown(mousePt, menu))
		{
			App::Inst().SetRightClickView(m_layoutLine);

			// This is a bit of a hack. Basically, the LayoutLine has a popup menu
			// that makes sense once it has been associated with another object (like a tree).
			// Here, we want a popup menu for when the LayoutLine is still unassociated with 
			// any other object. As such, we define a different popup menu here than what is returned
			// by MouseRightDown.
			wxMenu menu;

			menu.Append(ID_MNU_LAYOUT_LINE_DELETE, wxT("Delete"));
			menu.AppendSeparator();
			menu.Append(ID_POPUP_MNU_PROPERTIES, wxT("Properties"));
			PopupMenu(&menu);

			return true;
		}
		else if(m_layoutEllipse && m_layoutEllipse->MouseRightDown(mousePt, menu))
		{
			App::Inst().SetRightClickView(m_layoutEllipse);

			// This is a bit of a hack. Basically, the LayoutEllipse has a popup menu
			// that makes sense once it has been associated with another object (like a tree).
			// Here, we want a popup menu for when the LayoutEllipse is still unassociated with 
			// any other object. As such, we define a different popup menu here than what is returned
			// by MouseRightDown.
			wxMenu menu;

			menu.Append(ID_MNU_LAYOUT_ELLIPSE_DELETE, wxT("Delete"));
			menu.AppendSeparator();
			menu.Append(ID_POPUP_MNU_PROPERTIES, wxT("Properties"));
			PopupMenu(&menu);

			return true;
		}
		else if(m_geoAxisPolyline && m_geoAxisPolyline->MouseRightDown(mousePt, menu))
		{
			App::Inst().SetRightClickView(m_geoAxisPolyline);
			PopupMenu(&menu);
			return true;
		}
		else if(LabelController::Inst().MouseRightDown(mousePt, menu))
		{
			PopupMenu(&menu);
			return true;
		}
		else if(App::Inst().GetLayerTreeController()->MouseRightDown(mousePt, menu))
		{
			PopupMenu(&menu);
			return true;
		}
	}

	return false;
}

void GenGisFrame::SigViewportMouse(wxMouseEvent& event)
{
	static bool bDragging = false;
	static bool bActiveElement = false;

	event.Skip();	// note: needs to be called so window will get focus (see wxWidgets documentation)

	// get mouse point in OpenGL screen coordinates
	Point2D mousePos(event.GetPosition().x, App::Inst().GetViewport()->GetHeight()-event.GetPosition().y);
	App::Inst().SetMouseScreenPos(mousePos);	

	// Check if viewport needs to be refreshed on a mouse move
	bool bUpdatedMousePos = false;
	if(event.Moving())
	{
		if(m_terrainController && App::Inst().GetSelectedView() && App::Inst().GetSelectedView()->MouseMove(mousePos))
		{
			bUpdatedMousePos = true;

			// save mouse position in world and screen space
			Point3D worldPos = m_terrainController->ProjectToTerrain(mousePos, false);
			App::Inst().SetMouseWorldPos(worldPos);
			
			App::Inst().GetViewport()->Render();
		}
	}

	// *** Check if the user is clicking a mouse button or dragging the mouse ***
	if(event.LeftDClick())
	{
		// process float click events and quick button presses
		MouseLeftDoubleClick(mousePos);
		App::Inst().GetViewport()->Refresh(false);
	}
	else if(event.LeftDown())
	{
		// a dragging operation is starting
		bDragging = true;

		bActiveElement = MouseLeftDown(mousePos);
		if(bActiveElement)
			App::Inst().GetViewport()->Refresh(false);
	}
	else if(event.RightDown())
	{
		// a dragging operation is starting
		bDragging = true;

		MouseRightDown(mousePos);
	}
	else if(event.LeftUp())
	{
		// done dragging
		bDragging = false;
		if(bActiveElement)
		{			
			MouseLeftUp(mousePos);
			bActiveElement = false;
		}

		App::Inst().SetMouseDragging(false);
		App::Inst().GetViewport()->Refresh(false);
	}
	else if(event.RightUp())
	{
		// done dragging
		bDragging = false;

		App::Inst().SetMouseDragging(false);
		App::Inst().GetViewport()->Refresh(false);
	}
	else if(event.Dragging() && bDragging)	// using event.Dragging() fails when the File->Open dialog is used???
  {
		App::Inst().SetMouseDragging(true);

		// we are in the middle of a dragging operation
		int dx = mousePos.x - m_lastMousePos.x;
		int dy = mousePos.y - m_lastMousePos.y;

    if(event.LeftIsDown())
  	{
			bool bRendered = false;
			if(bActiveElement)
			{				
				if(!bUpdatedMousePos && m_terrainController && App::Inst().GetSelectedView() && App::Inst().GetSelectedView()->MouseMove(mousePos))
				{
					// save mouse position in world and screen space
					Point3D worldPos = m_terrainController->ProjectToTerrain(mousePos, false);
					App::Inst().SetMouseWorldPos(worldPos);

					App::Inst().GetViewport()->Render();
					bRendered = true;
				}

				MouseDragging(mousePos);
			}
			else
			{	
				const float TRANSLATION_SENSITIVITY = 0.001f;
				App::Inst().GetViewport()->GetCamera()->TranslateDelta(dx*TRANSLATION_SENSITIVITY, -dy*TRANSLATION_SENSITIVITY);
			}		

			if(!bRendered)
				App::Inst().GetViewport()->Refresh(false);
  	}

		if(event.RightIsDown())
		{
			const float ANGLE_Y_SENSITIVITY = 0.5f;
			const float PITCH_SENSITIVITY = 0.5f;

			App::Inst().GetViewport()->GetCamera()->YawDelta(-dx*ANGLE_Y_SENSITIVITY);
			App::Inst().GetViewport()->GetCamera()->PitchDelta(-dy*PITCH_SENSITIVITY);

			m_widgetController->SetCompassAngle(App::Inst().GetViewport()->GetCamera()->GetYaw());

			App::Inst().GetViewport()->Refresh(false);
		}
  }

	m_lastMousePos = mousePos;
}

void GenGisFrame::SigViewportMouseWheel(wxMouseEvent& event)
{
	const float WHEEL_ZOOM_SENSITIVITY = 0.025f;
	const float REDRAW_ZOOM_SENSITIVITY = 0.25f;
	static float prevZoom = std::numeric_limits<float>::max();

	int dWheel = event.GetWheelRotation()/event.GetWheelDelta();
	App::Inst().GetViewport()->GetCamera()->ZoomDelta((float)(-dWheel*WHEEL_ZOOM_SENSITIVITY));

	float curZoom = App::Inst().GetViewport()->GetCamera()->GetZoom();
	float deltaZoom = (curZoom - prevZoom) / prevZoom;
	if(prevZoom == std::numeric_limits<float>::max() || (deltaZoom < -REDRAW_ZOOM_SENSITIVITY || deltaZoom > REDRAW_ZOOM_SENSITIVITY))
	{
		prevZoom = curZoom;

		if(m_terrainController)
			m_terrainController->GetTerrainView()->RefreshQuadtree();
	}

	Refresh(false);
}