//=======================================================================
// Author: Donovan Parks
//
// Copyright 2009 Donovan Parks
//
// This file is part of GenGIS.
//
// GenGIS is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// GenGIS is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with GenGIS.  If not, see <http://www.gnu.org/licenses/>.
//=======================================================================

#include "../core/Precompiled.hpp"

#include "../gui/MapPropertiesDlg.hpp"
#include "../gui/ProgressDlg.hpp"
#include "../gui/TerrainMapWidget.hpp"

#include "../core/App.hpp"
#include "../core/TerrainTexture.hpp"
#include "../core/MapLayer.hpp"
#include "../core/TerrainController.hpp"
#include "../core/TerrainView.hpp"
#include "../core/Viewport.hpp"

#include "../utils/StringTools.hpp"

using namespace GenGIS;

const int MAX_ELEVATION = 10000;

MapPropertiesDlg::MapPropertiesDlg(wxWindow* parent, MapLayerPtr mapLayer)
	: MapPropertiesLayout(parent), m_mapLayer(mapLayer), m_terrainMapWidget(new TerrainMapWidget(m_cboColourMap, &m_bColourMapChanged))
{
	Init();
	Fit();
}

MapPropertiesDlg::~MapPropertiesDlg()
{

}

int MapPropertiesDlg::ScaleElevation(float elevation)
{
	float max = m_mapLayer->GetTerrainController()->GetMaxElevationGridSpace();
	float normElevation = (max-elevation) / max;

	// invert the elevation since the slider really represents an error measure
	return int(MAX_ELEVATION*normElevation);
}

float MapPropertiesDlg::ScaleElevation(int elevation)
{
	TerrainControllerPtr terrainController = m_mapLayer->GetTerrainController();

	float normElevation = float(MAX_ELEVATION-elevation) / MAX_ELEVATION;

	return normElevation*terrainController->GetMaxElevationGridSpace();
}

void MapPropertiesDlg::Init()
{
	// set state of controls on General page
	m_txtLayerName->SetValue(wxString(m_mapLayer->GetName().c_str()));
	m_txtLayerDescription->SetValue(wxString(m_mapLayer->GetDescription().c_str()));
	m_txtAuthours->SetValue(wxString(m_mapLayer->GetAuthours().c_str()));

	// set state of controls on Symbology page
	TerrainViewPtr terrainView = m_mapLayer->GetTerrainController()->GetTerrainView();
	m_chkWireframe->SetValue(m_mapLayer->GetTerrainController()->GetTerrainView()->IsWireFrame());
	m_txtVerticalExaggeration->SetValue(wxString(util::StringTools::ToStringW(terrainView->GetVerticalExaggeration()).c_str()));
	m_sliderTransparency->SetValue(int(terrainView->GetTerrainModel()->GetTexture()->GetTransparencyPercentage()));

	m_sliderLevelOfDetail->SetMax(MAX_ELEVATION);
	m_sliderLevelOfDetail->SetMin(0);
	m_sliderLevelOfDetail->SetValue(ScaleElevation(terrainView->GetEpsilon()));

	InitColourMap();

	// set state of controls on Metadata page
	m_txtLayerSource->SetValue(wxString(m_mapLayer->GetPath().c_str()) + _T("\\") + wxString(m_mapLayer->GetFilename().c_str()));
	InitMetaData();
}

void MapPropertiesDlg::InitColourMap()
{
	TerrainTexturePtr terrainTexture = m_mapLayer->GetTerrainController()->GetTerrainView()->GetTerrainModel()->GetTexture();

	// Populate colour map combo box with all available colour maps
	m_terrainMapWidget->SetColourMap(terrainTexture->GetColourMap());
	m_terrainMapWidget->PopulateColourMapComboBox();

	// Set number of entires, interpolation, and intervals
	m_spinNumEntries->SetValue(terrainTexture->GetNumColours());

	if(terrainTexture->GetInterpolation() == TerrainTexture::DISCRETE)
		m_cboInterpolation->SetValue(_T("Discrete"));
	else if(terrainTexture->GetInterpolation() == TerrainTexture::LINEARLY)
		m_cboInterpolation->SetValue(_T("Linearly"));
	
	std::vector<float> intervals;
	terrainTexture->GetIntervals(intervals);
	m_terrainMapWidget->SetIntervalData(intervals);

	// Set colour map
	OnNumEntriesChange();

	if(terrainTexture->GetInterpolation() == TerrainTexture::COLOUR_GRID)
	{
		m_cboColourMap->Enable(false);
		m_cboInterpolation->Enable(false);
		m_spinNumEntries->Enable(false);
	}

	m_bColourMapChanged = false;
}

void MapPropertiesDlg::OnNumEntriesChange() 
{ 
	TerrainTexturePtr terrainTexture = m_mapLayer->GetTerrainController()->GetTerrainView()->GetTerrainModel()->GetTexture();

	m_terrainMapWidget->SetEntries(m_scrolledWindowColourMap, m_spinNumEntries->GetValue(),
																		m_mapLayer->GetTerrainController()->GetMaxElevation(), 
																		m_mapLayer->GetTerrainController()->GetMinElevation());

	m_bColourMapChanged = true;
}


void MapPropertiesDlg::OnColourMapChange( wxCommandEvent& event ) 
{ 
	m_terrainMapWidget->SetColourMap(); 
	m_bColourMapChanged = true;
}

void MapPropertiesDlg::OnInterpolationChange( wxCommandEvent& event )
{
	m_bColourMapChanged = true;
}

void MapPropertiesDlg::OnEvenlySpace( wxCommandEvent& event )
{
	m_terrainMapWidget->EvenlySpaceEntries(m_scrolledWindowColourMap,
																					m_mapLayer->GetTerrainController()->GetMaxElevation(), 
																					m_mapLayer->GetTerrainController()->GetMinElevation());

	m_bColourMapChanged = true;
}

void MapPropertiesDlg::InitMetaData()
{
	FileMetaData* metaData = m_mapLayer->GetTerrainController()->GetMetaData();
	m_txtMetaData->BeginBold();
	m_txtMetaData->AppendText(_T("Driver:\n"));
	m_txtMetaData->EndBold();
	m_txtMetaData->AppendText(wxString(util::StringTools::ToStringW(metaData->driverDesc).c_str()) + _T("\n"));
	m_txtMetaData->AppendText(wxString(util::StringTools::ToStringW(metaData->driverMetaData).c_str()) + _T("\n"));

	m_txtMetaData->BeginBold();
	m_txtMetaData->AppendText(_T("\nProjection:\n"));
	m_txtMetaData->EndBold();
	m_txtMetaData->AppendText(wxString(util::StringTools::ToStringW(metaData->projection).c_str()) + _T("\n"));

	m_txtMetaData->BeginBold();
	m_txtMetaData->AppendText(_T("\nDimensions:\n"));
	m_txtMetaData->EndBold();
	m_txtMetaData->AppendText(_T("X: ") + wxString(util::StringTools::ToStringW(metaData->xSize).c_str())
															+ _T(", Y: ") + wxString(util::StringTools::ToStringW(metaData->ySize).c_str())
															+ _T(", Bands: ") + wxString(util::StringTools::ToStringW(metaData->bands).c_str()) + _T("\n"));

	m_txtMetaData->BeginBold();
	m_txtMetaData->AppendText(_T("\nPixel Size:\n"));
	m_txtMetaData->EndBold();
	m_txtMetaData->AppendText(wxString(util::StringTools::ToStringW(metaData->pixelSizeX, 4).c_str()) + _T(", ") 
																+ wxString(util::StringTools::ToStringW(metaData->pixelSizeY, 4).c_str()) + _T("\n"));

	m_txtMetaData->BeginBold();
	m_txtMetaData->AppendText(_T("\nOrigin:\n"));
	m_txtMetaData->EndBold();
	m_txtMetaData->AppendText(wxString(util::StringTools::ToStringW(metaData->originX, 2).c_str()) + _T(", ") 
																+ wxString(util::StringTools::ToStringW(metaData->originY, 2).c_str()) + _T("\n"));

	m_txtMetaData->BeginBold();
	m_txtMetaData->AppendText(_T("\nExtents:\n"));
	m_txtMetaData->EndBold();
	m_txtMetaData->AppendText(_T("Lower, left corner: ") + wxString(util::StringTools::ToStringW(metaData->extents.x, 2).c_str()) + _T(", ") 
																+ wxString(util::StringTools::ToStringW(metaData->extents.y, 2).c_str()) + _T("\n"));
	m_txtMetaData->AppendText(_T("Upper, right corner: ") + wxString(util::StringTools::ToStringW(metaData->extents.dx, 2).c_str()) + _T(", ") 
																+ wxString(util::StringTools::ToStringW(metaData->extents.dy, 2).c_str()) + _T("\n"));
	
	m_txtMetaData->BeginBold();
	m_txtMetaData->AppendText(_T("\nData Type:\n"));
	m_txtMetaData->EndBold();
	m_txtMetaData->AppendText(wxString(util::StringTools::ToStringW(metaData->dataType).c_str()));
}

void MapPropertiesDlg::Apply()
{
	// set properties based on state of controls in General page
	m_mapLayer->SetName(m_txtLayerName->GetValue().c_str());
	m_mapLayer->SetDescription(m_txtLayerDescription->GetValue().c_str());
	m_mapLayer->SetAuthours(m_txtAuthours->GetValue().c_str());

	// set properties based on state of controls in Symbology page
	m_mapLayer->GetTerrainController()->GetTerrainView()->SetWireFrame(m_chkWireframe->IsChecked());
	m_mapLayer->GetTerrainController()->GetTerrainView()->SetVerticalExaggeration(util::StringTools::ToDouble(m_txtVerticalExaggeration->GetValue().c_str()));

	m_mapLayer->GetTerrainController()->GetTerrainView()->GetTerrainModel()->GetTexture()->SetTransparencyPercentage(m_sliderTransparency->GetValue(), this);
	m_mapLayer->GetTerrainController()->GetTerrainView()->SetEpsilon(ScaleElevation(m_sliderLevelOfDetail->GetValue()));

	ApplyColourMap();

	App::Inst().GetViewport()->Refresh(false);
	App::Inst().GetLayerTreeController()->SetName(m_mapLayer->GetId(), m_mapLayer->GetName());
}

void MapPropertiesDlg::ApplyColourMap()
{
	if(m_bColourMapChanged)
	{
		TerrainTexturePtr terrainTexture = m_mapLayer->GetTerrainController()->GetTerrainView()->GetTerrainModel()->GetTexture();

		TerrainTexture::INTERPOLATE interpolate;
		if(m_cboInterpolation->GetValue() == _T("Discrete"))
			interpolate = TerrainTexture::DISCRETE;
		else if(m_cboInterpolation->GetValue() == _T("Linearly"))
			interpolate = TerrainTexture::LINEARLY;

		std::vector<float> entries;
		m_terrainMapWidget->GetEntries(entries);

		terrainTexture->SetColourMap(m_terrainMapWidget->GetColourMap(), entries, interpolate, this);
	}

	m_bColourMapChanged = false;
}
