//=======================================================================
// Author: Donovan Parks
//
// Copyright 2009 Donovan Parks
//
// This file is part of GenGIS.
//
// GenGIS is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// GenGIS is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with GenGIS.  If not, see <http://www.gnu.org/licenses/>.
//=======================================================================

#include "../core/Precompiled.hpp"

#include "../core/App.hpp"
#include "../gui/TreePropertiesDlg.hpp"
#include "../utils/StringTools.hpp"
#include "../gui/ProgressDlg.hpp"
#include "../core/GeoTreeController.hpp"
#include "../core/GeoTreeView.hpp"
#include "../core/TreeLayer.hpp"
#include "../core/Viewport.hpp"

using namespace GenGIS;

TreePropertiesDlg::TreePropertiesDlg(wxWindow* parent, TreeLayerPtr treeLayer)
	: TreePropertiesLayout(parent), m_treeLayer(treeLayer)
{
	Init();
	Fit();
}

TreePropertiesDlg::~TreePropertiesDlg()
{

}

void TreePropertiesDlg::Init()
{
	// *** Set state of controls on General page
	m_txtLayerName->SetValue(wxString(m_treeLayer->GetName().c_str()));
	m_txtLayerDescription->SetValue(wxString(m_treeLayer->GetDescription().c_str()));
	m_txtAuthours->SetValue(wxString(m_treeLayer->GetAuthours().c_str()));

	// *** Set state of controls on Symbology page
	GeoTreeViewPtr geoTreeView = m_treeLayer->GetGeoTreeController()->GetTreeView();

	// tree layout properties
	if(geoTreeView->GetLayout() == GeoTreeView::SLANTED_CLADOGRAM_3D)
		m_cboTreeLayout->SetValue(_T("3D slanted cladogram"));
	else if(geoTreeView->GetLayout() == GeoTreeView::CLADOGRAM_3D)
		m_cboTreeLayout->SetValue(_T("3D cladogram"));
	else if(geoTreeView->GetLayout() == GeoTreeView::CLADOGRAM_2D)
		m_cboTreeLayout->SetValue(_T("2D cladogram"));
	else if(geoTreeView->GetLayout() == GeoTreeView::PHYLOGRAM_2D)
		m_cboTreeLayout->SetValue(_T("2D phylogram"));
	else if(geoTreeView->GetLayout() == GeoTreeView::ELLIPSE_QUADRANT)
		m_cboTreeLayout->SetValue(_T("Elliptical Layout"));

	m_spinTreeThickness->SetValue(geoTreeView->GetLineThickness());
	m_txtTreeHeight->SetValue(wxString::Format(wxT("%f"), geoTreeView->GetHeight()));
	
	if(geoTreeView->GetOrientation() == GeoTreeView::VERTICAL)
		m_cbo2dOrientation->SetValue(_T("Vertical"));
	else if(geoTreeView->GetOrientation() == GeoTreeView::HORIZONTAL)
		m_cbo2dOrientation->SetValue(_T("Horizontal"));

	m_chkDrawLayoutPrimative->SetValue(geoTreeView->GetLayoutPrimativeVisibility());

	// tree style properties
	Colour treeColour = geoTreeView->GetColour();
	m_colourTree->SetColour(wxColour(treeColour.GetRedInt(), treeColour.GetGreenInt(), treeColour.GetBlueInt(), treeColour.GetAlphaInt()));
	
	if(geoTreeView->GetTreeColourStyle() == GeoTreeView::COLOUR_SINGLE)
		m_cboTreeColourStyle->SetValue(_T("Single colour"));
	else if(geoTreeView->GetTreeColourStyle() == GeoTreeView::COLOUR_DISCRETE)
		m_cboTreeColourStyle->SetValue(_T("Propogate discrete colours"));
	else if(geoTreeView->GetTreeColourStyle() == GeoTreeView::COLOUR_CONTINUOUS)
		m_cboTreeColourStyle->SetValue(_T("Propogate continuous colours"));

	m_spinLeafNodeSize->SetValue(geoTreeView->GetLeafNodeSize());
	m_spinInternalNodeSize->SetValue(geoTreeView->GetInternalNodeSize());

	m_spinBranchBorderSize->SetValue(geoTreeView->GetBranchBorderSize());
	m_spinLeafNodeBorderSize->SetValue(geoTreeView->GetLeafNodeBorderSize());
	m_spinInternalNodeBorderSize->SetValue(geoTreeView->GetInternalNodeBorderSize());
	m_spinGeoPointBorderSize->SetValue(geoTreeView->GetGeoPtsBorderSize());
	m_spinLocationLineBorderSize->SetValue(geoTreeView->GetLocationLineBorderSize());
	m_spinCorrelationLineBorderSize->SetValue(geoTreeView->GetCorrelationLineBorderSize());

	Colour leafNodeBorderColour = geoTreeView->GetLeafNodeBorderColour();
	m_colourLeafNodeBorder->SetColour(wxColour(leafNodeBorderColour.GetRedInt(), leafNodeBorderColour.GetGreenInt(), 
																							leafNodeBorderColour.GetBlueInt(), leafNodeBorderColour.GetAlphaInt()));

	Colour internalNodeBorderColour = geoTreeView->GetInternalNodeBorderColour();
	m_colourInternalNodeBorder->SetColour(wxColour(internalNodeBorderColour.GetRedInt(), internalNodeBorderColour.GetGreenInt(), 
																							internalNodeBorderColour.GetBlueInt(), internalNodeBorderColour.GetAlphaInt()));

	Colour geoPtsBorderColour = geoTreeView->GetGeoPtsBorderColour();
	m_colourGeoPointBorder->SetColour(wxColour(geoPtsBorderColour.GetRedInt(), geoPtsBorderColour.GetGreenInt(), 
																							geoPtsBorderColour.GetBlueInt(), geoPtsBorderColour.GetAlphaInt()));

	Colour branchBorderColour = geoTreeView->GetBranchBorderColour();
	m_colourBranchBorder->SetColour(wxColour(branchBorderColour.GetRedInt(), branchBorderColour.GetGreenInt(), 
																							branchBorderColour.GetBlueInt(), branchBorderColour.GetAlphaInt()));

	Colour locationLineBorderColour = geoTreeView->GetLocationLineBorderColour();
	m_colourLocationLineBorder->SetColour(wxColour(locationLineBorderColour.GetRedInt(), locationLineBorderColour.GetGreenInt(), 
																							locationLineBorderColour.GetBlueInt(), locationLineBorderColour.GetAlphaInt()));

	Colour correlationLineBorderColour = geoTreeView->GetCorrelationLineBorderColour();
	m_colourCorrelationLineBorder->SetColour(wxColour(correlationLineBorderColour.GetRedInt(), correlationLineBorderColour.GetGreenInt(), 
																							correlationLineBorderColour.GetBlueInt(), correlationLineBorderColour.GetAlphaInt()));
	// location line properties
	m_spinLocationLineThickness->SetValue(geoTreeView->LocationLines().GetThickness());

	if(geoTreeView->LocationLines().GetLineStyle() == VisualLine::SOLID)
		m_cboLocationLineStyle->SetValue(_T("Solid"));
	else if(geoTreeView->LocationLines().GetLineStyle() == VisualLine::SHORT_DASH)
		m_cboLocationLineStyle->SetValue(_T("Short dash"));
	else if(geoTreeView->LocationLines().GetLineStyle() == VisualLine::LONG_DASH)
		m_cboLocationLineStyle->SetValue(_T("Long dash"));
	else if(geoTreeView->LocationLines().GetLineStyle() == VisualLine::HIDDEN)
		m_cboLocationLineStyle->SetValue(_T("Hidden"));


	// correlation line properties
	m_spinCorrelationLineThickness->SetValue(geoTreeView->CorrelationLines().GetThickness());

	if(geoTreeView->CorrelationLines().GetLineStyle() == VisualLine::SOLID)
		m_cboCorrelationLineStyle->SetValue(_T("Solid"));
	else if(geoTreeView->CorrelationLines().GetLineStyle() == VisualLine::SHORT_DASH)
		m_cboCorrelationLineStyle->SetValue(_T("Short dash"));
	else if(geoTreeView->CorrelationLines().GetLineStyle() == VisualLine::LONG_DASH)
		m_cboCorrelationLineStyle->SetValue(_T("Long dash"));
	else if(geoTreeView->CorrelationLines().GetLineStyle() == VisualLine::HIDDEN)
		m_cboCorrelationLineStyle->SetValue(_T("Hidden"));

	// drop line properties
	Colour colour = geoTreeView->Get3dDropLineStyle().GetColour();
	m_colourDropLine->SetColour(wxColour(colour.GetRedInt(), colour.GetGreenInt(), colour.GetBlueInt(), colour.GetAlphaInt()));

	if(geoTreeView->Get3dDropLineStyle().GetLineStyle() == VisualLine::SOLID)
		m_cboDropLineStyle->SetValue(_T("Solid"));
	else if(geoTreeView->Get3dDropLineStyle().GetLineStyle() == VisualLine::SHORT_DASH)
		m_cboDropLineStyle->SetValue(_T("Short dash"));
	else if(geoTreeView->Get3dDropLineStyle().GetLineStyle() == VisualLine::LONG_DASH)
		m_cboDropLineStyle->SetValue(_T("Long dash"));
	else if(geoTreeView->Get3dDropLineStyle().GetLineStyle() == VisualLine::HIDDEN)
		m_cboDropLineStyle->SetValue(_T("Hidden"));

	// geography line properties
	m_spinGeoLineOffset->SetValue(uint(geoTreeView->GetGeographyLineOffsetPercentage()*100 + 0.5));

	m_spinGeoLineThickness->SetValue(geoTreeView->GeographyLine().GetSize());

	if(geoTreeView->GeographyLine().GetLineStyle() == VisualLine::SOLID)
		m_cboGeoLineStyle->SetValue(_T("Solid"));
	else if(geoTreeView->GeographyLine().GetLineStyle() == VisualLine::SHORT_DASH)
		m_cboGeoLineStyle->SetValue(_T("Short dash"));
	else if(geoTreeView->GeographyLine().GetLineStyle() == VisualLine::LONG_DASH)
		m_cboGeoLineStyle->SetValue(_T("Long dash"));
	else if(geoTreeView->GeographyLine().GetLineStyle() == VisualLine::HIDDEN)
		m_cboGeoLineStyle->SetValue(_T("Hidden"));

	// geographic point properties
	m_spinGeoPointSize->SetValue(geoTreeView->GeographicPoints().GetSize());

	m_chkDrawGeoPoint->SetValue(geoTreeView->GeographicPoints().GetVisibility());
	m_chkEvenlySpeadGeoPts->SetValue(geoTreeView->GetSpreadGeographyPts());

	m_chkDrawGeoAxis->SetValue(geoTreeView->GetGeographicAxisVisibility());

	m_chkOptimizeLeafNodes->SetValue(geoTreeView->GetOptimizeTopology());


	// *** Set state of controls on Label page
	m_spinLabelSize->SetValue(geoTreeView->GetFontSize());

	Colour fontColour = geoTreeView->GetFontColour();
	m_colourLabel->SetColour(wxColour(fontColour.GetRedInt(), fontColour.GetGreenInt(), fontColour.GetBlueInt(), fontColour.GetAlphaInt()));

	m_chkShowLabels->SetValue(geoTreeView->GetFontVisibility());

	// set state of controls on Metadata page
	m_txtLayerSource->SetValue(wxString(m_treeLayer->GetPath().c_str()) + _T("\\") + wxString(m_treeLayer->GetFilename().c_str()));
	SetMetaData();
}

void TreePropertiesDlg::SetMetaData()
{
	GeoTreeControllerPtr controller = m_treeLayer->GetGeoTreeController();

	m_txtMetaData->BeginBold();
	m_txtMetaData->AppendText(_T("Number of nodes:\n"));
	m_txtMetaData->EndBold();
	m_txtMetaData->AppendText(wxString(util::StringTools::ToStringW(controller->GetNumberOfNodes()).c_str()) + _T("\n"));

	m_txtMetaData->BeginBold();
	m_txtMetaData->AppendText(_T("\nNumber of leaf nodes:\n"));
	m_txtMetaData->EndBold();
	m_txtMetaData->AppendText(wxString(util::StringTools::ToStringW(controller->GetNumberOfLeaves()).c_str()) + _T("\n"));

	m_txtMetaData->BeginBold();
	m_txtMetaData->AppendText(_T("\nDepth of tree:\n"));
	m_txtMetaData->EndBold();
	m_txtMetaData->AppendText(wxString(util::StringTools::ToStringW(controller->GetDepthOfTree()).c_str()) + _T("\n"));

	m_txtMetaData->BeginBold();
	m_txtMetaData->AppendText(_T("\nMultifurcating tree:\n"));
	m_txtMetaData->EndBold();
	if(controller->IsMultifurcating())
		m_txtMetaData->AppendText(_T("True\n"));
	else
		m_txtMetaData->AppendText(_T("False\n"));

	m_txtMetaData->BeginBold();
	m_txtMetaData->AppendText(_T("\nTopology optimized to leaf node geography:\n"));
	m_txtMetaData->EndBold();
	if(m_treeLayer->GetGeoTreeController()->GetTreeView()->GetOptimizeTopology())
		m_txtMetaData->AppendText(_T("True"));
	else
		m_txtMetaData->AppendText(_T("False"));
}

void TreePropertiesDlg::Apply()
{
	// *** Set properties based on state of controls in General page
	m_treeLayer->SetName(m_txtLayerName->GetValue().c_str());
	m_treeLayer->SetDescription(m_txtLayerDescription->GetValue().c_str());
	m_treeLayer->SetAuthours(m_txtAuthours->GetValue().c_str());

	// *** Set properties based on state of control in Symbology page
	GeoTreeViewPtr geoTreeView = m_treeLayer->GetGeoTreeController()->GetTreeView();

	if(m_cboTreeLayout->GetValue() == _T("3D slanted cladogram"))
		geoTreeView->SetLayout(GeoTreeView::SLANTED_CLADOGRAM_3D);
	else if(m_cboTreeLayout->GetValue() == _T("3D cladogram"))
		geoTreeView->SetLayout(GeoTreeView::CLADOGRAM_3D);
	else if(m_cboTreeLayout->GetValue() == _T("2D slanted cladogram"))
		geoTreeView->SetLayout(GeoTreeView::SLANTED_CLADOGRAM_2D);
	else if(m_cboTreeLayout->GetValue() == _T("2D cladogram"))
		geoTreeView->SetLayout(GeoTreeView::CLADOGRAM_2D);
	else if(m_cboTreeLayout->GetValue() == _T("2D phylogram"))
		geoTreeView->SetLayout(GeoTreeView::PHYLOGRAM_2D);
	else if(m_cboTreeLayout->GetValue() == _T("Elliptical layout"))
		geoTreeView->SetLayout(GeoTreeView::ELLIPSE_QUADRANT);

	geoTreeView->SetLineThickness(m_spinTreeThickness->GetValue());

	geoTreeView->SetHeight(util::StringTools::ToDouble(std::wstring(m_txtTreeHeight->GetValue())));

	if(m_cbo2dOrientation->GetValue() == _T("Vertical"))
		geoTreeView->SetOrientation(GeoTreeView::VERTICAL);
	else if(m_cbo2dOrientation->GetValue() == _T("Horizontal"))
		geoTreeView->SetOrientation(GeoTreeView::HORIZONTAL);

	geoTreeView->SetLayoutPrimativeVisibility(m_chkDrawLayoutPrimative->GetValue());

	geoTreeView->SetColour(Colour(m_colourTree->GetColour()));

	if(m_cboTreeColourStyle->GetValue() == _T("Single colour"))
		geoTreeView->SetTreeColourStyle(GeoTreeView::COLOUR_SINGLE);
	else if(m_cboTreeColourStyle->GetValue() == _T("Propogate discrete colours"))
		geoTreeView->SetTreeColourStyle(GeoTreeView::COLOUR_DISCRETE);
	else if(m_cboTreeColourStyle->GetValue() == _T("Propogate continuous colours"))
		geoTreeView->SetTreeColourStyle(GeoTreeView::COLOUR_CONTINUOUS);

	geoTreeView->SetLeafNodeSize(m_spinLeafNodeSize->GetValue());
	geoTreeView->SetInternalNodeSize(m_spinInternalNodeSize->GetValue());

	geoTreeView->SetLeafNodeBorderSize(m_spinLeafNodeBorderSize->GetValue());
	geoTreeView->SetInternalNodeBorderSize(m_spinInternalNodeBorderSize->GetValue());
	geoTreeView->SetGeoPtsBorderSize(m_spinGeoPointBorderSize->GetValue());
	geoTreeView->SetBranchBorderSize(m_spinBranchBorderSize->GetValue());
	geoTreeView->SetLocationLineBorderSize(m_spinLocationLineBorderSize->GetValue());
	geoTreeView->SetCorrelationLineBorderSize(m_spinCorrelationLineBorderSize->GetValue());

	geoTreeView->SetLeafNodeBorderColour(m_colourLeafNodeBorder->GetColour());
	geoTreeView->SetInternalNodeBorderColour(m_colourInternalNodeBorder->GetColour());
	geoTreeView->SetGeoPtsBorderColour(m_colourGeoPointBorder->GetColour());
	geoTreeView->SetBranchBorderColour(m_colourBranchBorder->GetColour());
	geoTreeView->SetLocationLineBorderColour(m_colourLocationLineBorder->GetColour());
	geoTreeView->SetCorrelationLineBorderColour(m_colourCorrelationLineBorder->GetColour());

	// set location line properties
	geoTreeView->LocationLines().SetThickness(m_spinLocationLineThickness->GetValue());

	geoTreeView->LocationLines().SetVisibility(true);
	if(m_cboLocationLineStyle->GetValue() == _T("Solid"))
		geoTreeView->LocationLines().SetLineStyle(VisualLine::SOLID);
	else if(m_cboLocationLineStyle->GetValue() == _T("Short dash"))
		geoTreeView->LocationLines().SetLineStyle(VisualLine::SHORT_DASH);
	else if(m_cboLocationLineStyle->GetValue() == _T("Long dash"))
		geoTreeView->LocationLines().SetLineStyle(VisualLine::LONG_DASH);
	else if(m_cboLocationLineStyle->GetValue() == _T("Hidden"))
	{
		geoTreeView->LocationLines().SetLineStyle(VisualLine::HIDDEN);
		geoTreeView->LocationLines().SetVisibility(false);
	}

	// set correlation line properties
	geoTreeView->CorrelationLines().SetThickness(m_spinCorrelationLineThickness->GetValue());

	geoTreeView->CorrelationLines().SetVisibility(true);
	if(m_cboCorrelationLineStyle->GetValue() == _T("Solid"))
		geoTreeView->CorrelationLines().SetLineStyle(VisualLine::SOLID);
	else if(m_cboCorrelationLineStyle->GetValue() == _T("Short dash"))
		geoTreeView->CorrelationLines().SetLineStyle(VisualLine::SHORT_DASH);
	else if(m_cboCorrelationLineStyle->GetValue() == _T("Long dash"))
		geoTreeView->CorrelationLines().SetLineStyle(VisualLine::LONG_DASH);
	else if(m_cboCorrelationLineStyle->GetValue() == _T("Hidden"))
	{
		geoTreeView->CorrelationLines().SetLineStyle(VisualLine::HIDDEN);
		geoTreeView->CorrelationLines().SetVisibility(false);
	}

	// set 3d drop line properties
	VisualLine dropLineStyle = geoTreeView->Get3dDropLineStyle();
	dropLineStyle.SetColour(Colour(m_colourDropLine->GetColour()));

	dropLineStyle.SetVisibility(true);
	if(m_cboDropLineStyle->GetValue() == _T("Solid"))
		dropLineStyle.SetLineStyle(VisualLine::SOLID);
	else if(m_cboDropLineStyle->GetValue() == _T("Short dash"))
		dropLineStyle.SetLineStyle(VisualLine::SHORT_DASH);
	else if(m_cboDropLineStyle->GetValue() == _T("Long dash"))
		dropLineStyle.SetLineStyle(VisualLine::LONG_DASH);
	else if(m_cboDropLineStyle->GetValue() == _T("Hidden"))
	{
		dropLineStyle.SetLineStyle(VisualLine::HIDDEN);
		dropLineStyle.SetVisibility(true);
	}

	geoTreeView->Set3dDropLineStyle(dropLineStyle);

	// set geography line properties
	geoTreeView->SetGeographyLineOffsetPercentage(m_spinGeoLineOffset->GetValue() / 100.0);
	geoTreeView->GeographyLine().SetSize(m_spinGeoLineThickness->GetValue());

	geoTreeView->GeographyLine().SetVisibility(true);
	if(m_cboGeoLineStyle->GetValue() == _T("Solid"))
		geoTreeView->GeographyLine().SetLineStyle(VisualLine::SOLID);
	else if(m_cboGeoLineStyle->GetValue() == _T("Short dash"))
		geoTreeView->GeographyLine().SetLineStyle(VisualLine::SHORT_DASH);
	else if(m_cboGeoLineStyle->GetValue() == _T("Long dash"))
		geoTreeView->GeographyLine().SetLineStyle(VisualLine::LONG_DASH);
	else if(m_cboGeoLineStyle->GetValue() == _T("Hidden"))
	{
		geoTreeView->GeographyLine().SetLineStyle(VisualLine::HIDDEN);
		geoTreeView->GeographyLine().SetVisibility(true);
	}

	geoTreeView->SetOptimizeTopology(m_chkOptimizeLeafNodes->GetValue());

	// geographic point properties
	geoTreeView->GeographicPoints().SetSize(m_spinGeoPointSize->GetValue());

	geoTreeView->GeographicPoints().SetVisibility(m_chkDrawGeoPoint->GetValue());
	geoTreeView->SetSpreadGeographyPts(m_chkEvenlySpeadGeoPts->GetValue());

	geoTreeView->SetGeographicAxisVisibility(m_chkDrawGeoAxis->GetValue());

	// *** Set properties based on state of control in Label page
	geoTreeView->SetFontSize(m_spinLabelSize->GetValue());

	geoTreeView->SetFontColour(Colour(m_colourLabel->GetColour()));

	geoTreeView->SetFontVisibility(m_chkShowLabels->GetValue());

	geoTreeView->ForceTreeLayout();

	App::Inst().GetViewport()->Refresh(false);
	App::Inst().GetLayerTreeController()->SetName(m_treeLayer->GetId(), m_treeLayer->GetName());
}

void TreePropertiesDlg::OnLayoutStyleChange( wxCommandEvent& event )
{
	GeoTreeViewPtr geoTreeView = m_treeLayer->GetGeoTreeController()->GetTreeView();
	
	// only allow 2D views if a layout line exists
	if(!geoTreeView->IsLayoutLine())
	{
		if(m_cboTreeLayout->GetValue() == _T("2D slanted cladogram")
				|| m_cboTreeLayout->GetValue() == _T("2D cladogram")
				|| m_cboTreeLayout->GetValue() == _T("2D phylogram"))
		{
			wxMessageBox(wxT("Please draw a layout line before switching to a 2D tree layout."),
										wxT("Missing layout line"), wxOK | wxICON_INFORMATION);

			// change it back to what it previously was
			if(geoTreeView->GetLayout() == GeoTreeView::SLANTED_CLADOGRAM_3D)
				m_cboTreeLayout->SetValue(_T("3D slanted cladogram"));
			else if(geoTreeView->GetLayout() == GeoTreeView::CLADOGRAM_3D)
				m_cboTreeLayout->SetValue(_T("3D cladogram"));
			else if(geoTreeView->GetLayout() == GeoTreeView::ELLIPSE_QUADRANT)
				m_cboTreeLayout->SetValue(_T("Elliptical layout"));
		}
	}
}
