//=======================================================================
// Author: Mike Porter
//
// Copyright 2009 Mike Porter
//
// This file is part of GenGIS.
//
// GenGIS is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// GenGIS is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with GenGIS.  If not, see <http://www.gnu.org/licenses/>.
//=======================================================================

#include "../core/Precompiled.hpp"

// Don't use this code in Windows debug builds!
#ifndef _DEBUG

#include "../utils/Tree.hpp"
#include "../core/NodeGeoTree.hpp"

#include "PythonInterface.hpp"
#include "../utils/StringTools.hpp"

//==============================================================================================
//   Custom-rolled C++ std::map to Python dictionary wrapper.
//==============================================================================================

template<class Key, class Val> class map_item
{
  public:
  
    static Val& get(std::map<Key,Val> &self, const Key idx)
    {
      if(self.find(idx) != self.end())
      {
        return self[idx];
      }
      PyErr_SetString(PyExc_KeyError,"Map key not found");
      boost::python::throw_error_already_set();
      return self[idx];
    }
    
    static void set(std::map<Key,Val> &self, const Key idx, const Val val)
    {
      self[idx]=val;
    }
    
    static void del(std::map<Key,Val> &self, const Key n)
    {
      self.erase(n);
    }
    
    static bool in(std::map<Key,Val> const &self, const Key n)
    {
      return self.find(n) != self.end();
    }  
    
    static boost::python::list keys(std::map<Key,Val> const &self)
    {
      boost::python::list t;
      typename std::map<Key,Val>::const_iterator it;
      for(it=self.begin(); it!=self.end(); ++it)
      {
        t.append(it->first);
      }
      return t;
    }
    
    static boost::python::list values(std::map<Key,Val> const &self)
    {
      boost::python::list t;
      typename std::map<Key,Val>::const_iterator it;
      for(it=self.begin(); it!=self.end(); ++it)
      {
        t.append(it->second);
      }
      return t;
    }
    
    static boost::python::list items(std::map<Key,Val> const &self)
    {
      boost::python::list t;
      typename std::map<Key,Val>::const_iterator it;
      for(it=self.begin(); it!=self.end(); ++it)
      {
        t.append(boost::python::make_tuple(it->first, it->second));
      }
      return t;
    }
    
    static std::wstring str(std::map<Key,Val> const &self)
    {
      std::wstring trailingComma = L"";
      std::wstring dictionaryString = L"{";
      typename std::map<Key,Val>::const_iterator it;
      for(it=self.begin(); it!=self.end(); ++it)
      {
        dictionaryString += trailingComma;
        dictionaryString += it->first;
        dictionaryString += L": ";
        dictionaryString += it->second;
        trailingComma = L", ";
      }
      dictionaryString += L"}";
      return dictionaryString;
    }
};

using namespace GenGIS;

//==============================================================================================
//   Automatically generated method pointers (for overloaded methods with default parameters)
//==============================================================================================

BOOST_PYTHON_FUNCTION_OVERLOADS(commands_overloads, APIv1::commands, 0, 1)
BOOST_PYTHON_FUNCTION_OVERLOADS(detail_overloads, APIv1::detail, 0, 1)
BOOST_PYTHON_FUNCTION_OVERLOADS(up_overloads, APIv1::up, 0, 1)
BOOST_PYTHON_FUNCTION_OVERLOADS(down_overloads, APIv1::down, 0, 1)
BOOST_PYTHON_FUNCTION_OVERLOADS(left_overloads, APIv1::left, 0, 1)
BOOST_PYTHON_FUNCTION_OVERLOADS(right_overloads, APIv1::right, 0, 1)
BOOST_PYTHON_FUNCTION_OVERLOADS(rotate_overloads, APIv1::rotate, 0, 1)
BOOST_PYTHON_FUNCTION_OVERLOADS(pitch_overloads, APIv1::pitch, 0, 1)
BOOST_PYTHON_FUNCTION_OVERLOADS(zoom_overloads, APIv1::zoom, 0, 1)

//==============================================
//   Method pointers (for overloaded methods)
//==============================================

MapLayerPtr (StudyLayer::*GetMapByIndex)(unsigned int index) const = &StudyLayer::GetMap;
LocationSetLayerPtr (MapLayer::*GetLocationSetByIndex)(unsigned int index) const = &MapLayer::GetLocationSet;
TreeLayerPtr (MapLayer::*GetTreeByIndex)(unsigned int index) const = &MapLayer::GetTree;
LocationLayerPtr (LocationSetLayer::*GetLocationByIndex)(unsigned int index) const = &LocationSetLayer::GetLocation;

//====================================================================================================
//   Thin wrappers (mainly for methods I want to add to one class but are methods of another class)
//====================================================================================================

// Map
// FileHeader* GetHeaderFromMapLayer(MapLayer m) { return m.GetTerrainController()->GetHeader(); }
// FileMetaData* GetMetaDataFromMapLayer(MapLayer m) { return m.GetTerrainController()->GetMetaData(); }
float GetWidthFromMapLayer(MapLayer m) { return m.GetTerrainController()->GetWidth(); }
float GetHeightFromMapLayer(MapLayer m) { return m.GetTerrainController()->GetHeight(); }
float GetMinElevationFromMapLayer(MapLayer m) { return m.GetTerrainController()->GetMinElevation(); }
float GetMaxElevationFromMapLayer(MapLayer m) { return m.GetTerrainController()->GetMaxElevation(); }

// Location
const std::wstring GetNameFromLocationLayer(LocationLayer l) { return l.GetLocationController()->GetName(); }
const std::wstring GetIdFromLocationLayer(LocationLayer l) { return l.GetLocationController()->GetId(); }
const float GetNorthingFromLocationLayer(LocationLayer l) { return l.GetLocationController()->GetNorthing(); }
const float GetEastingFromLocationLayer(LocationLayer l) { return l.GetLocationController()->GetEasting(); }
bool GetVisibilityLocationLayer(LocationLayer l) { return l.IsVisible(); }
ChartViewPtr GetChartLayer(LocationLayer l) { return l.GetLocationController()->GetChartView(); }
std::map<std::wstring,std::wstring> GetDataMapFromLocationLayer(LocationLayer l) { return l.GetLocationController()->GetData(); }
void AddSequenceToLocationLayer(LocationLayer l, SequenceModelPtr sequence) { l.GetLocationController()->AddSequence(sequence); }
void SetColourLocationLayer(LocationLayer l, const Colour& colour) { l.GetLocationController()->SetColour(colour); }
void SetSizeLocationLayer(LocationLayer l, float size) { l.GetLocationController()->SetSize(size); }
void SetVisibilityLocationLayer(LocationLayer l, bool state) { l.GetLocationController()->SetVisibility(state); }
void SetRenderingOrderLocationLayer(LocationLayer l, float renderingDepth) { l.GetLocationController()->SetRenderingOrder(renderingDepth); }

// Sequence
const std::wstring GetSequenceIdFromSequenceLayer(SequenceLayer s) { return s.GetSequenceController()->GetSequenceId(); }
const std::wstring GetSiteIdFromSequenceLayer(SequenceLayer s) { return s.GetSequenceController()->GetSiteId(); }
std::map<std::wstring,std::wstring> GetDataMapFromSequenceLayer(SequenceLayer s) { return s.GetSequenceController()->GetData(); }
bool GetActiveSequenceLayer(SequenceLayer s) { return s.GetSequenceController()->IsActive(); }
void SetActiveSequenceLayer(SequenceLayer s, bool state) { return s.GetSequenceController()->SetActive(state); }

// Colour map discrete
void SetColourDiscrete(ColourMapDiscrete& cmd, const std::wstring& name, const Colour& colour) { cmd.SetColour(name, colour); }
bool GetColourDiscrete(ColourMapDiscrete& cmd, const std::wstring& name, Colour& colour) { return cmd.GetColour(name, colour); }

//===============================
//   GenGisAPIv1 Python module
//===============================

BOOST_PYTHON_MODULE(GenGisAPIv1)
{
	//====================
  //   Enumerated types
  //====================
	boost::python::enum_<VisualMarker::MARKER_SHAPE>("MARKER_SHAPE")
		.value("CIRCLE", VisualMarker::CIRCLE)
    .value("SQUARE", VisualMarker::SQUARE)
		.value("TRIANGLE", VisualMarker::TRIANGLE)
		.value("STAR", VisualMarker::STAR)
		.value("PLUS_SIGN", VisualMarker::PLUS_SIGN)
		.value("HEXAGON", VisualMarker::HEXAGON)
		.value("DIAMOND", VisualMarker::DIAMOND)
		.value("SQUARE", VisualMarker::SQUARE)
    .value("INVERTED_TRIANGLE", VisualMarker::INVERTED_TRIANGLE)
		.export_values()
    ;

	boost::python::enum_<VisualMarker::MARKER_RENDERING_STYLE>("MARKER_RENDERING_STYLE")
		.value("DECAL", VisualMarker::DECAL)
    .value("ALWAYS_VISIBLE", VisualMarker::ALWAYS_VISIBLE)
		.value("MARKER_3D", VisualMarker::MARKER_3D)
		.export_values()
    ;

	boost::python::enum_<VisualLabel::LABEL_RENDERING_STYLE>("LABEL_RENDERING_STYLE")
		.value("PERSPECTIVE", VisualLabel::PERSPECTIVE)
    .value("ORTHO", VisualLabel::ORTHO)
		.export_values()
    ;

	boost::python::enum_<VisualLine::LINE_STYLE>("LINE_STYLE")
		.value("SOLID", VisualLine::SOLID)
    .value("SHORT_DASH", VisualLine::SHORT_DASH)
		.value("LONG_DASH", VisualLine::LONG_DASH)
		.export_values()
    ;

	boost::python::enum_<ChartView::CHART_TYPE>("CHART_TYPE")
		.value("PIE_CHART_2D", ChartView::PIE_CHART_2D)
    .value("PIE_CHART_3D", ChartView::PIE_CHART_3D)
		.export_values()
    ;

  //=============
  //   Methods
  //=============
  
  // console & log file I/O
  boost::python::def("write", &APIv1::write, "Writes `text` to the GenGis console and the log file.\nUsed by the API to print strings to the console.");
  boost::python::def("warning", &APIv1::warning, "Writes `text` to the GenGis console and the log file.\nUsed by the API to print warnings to the console.");
  boost::python::def("error", &APIv1::error, "Writes `text` to the GenGis console and the log file.\nUsed by the API to print errors to the console.");
  
  // Misc
  boost::python::def("commands", &APIv1::commands, commands_overloads(boost::python::args("detail"), "Prints all API commands.\n\nIf `detail` is True, a description will be included with each command."));
  
  // console display
  boost::python::def("console", &APIv1::displayConsole, "If `display` is True, the console is displayed.\nIf `display` is False, the console is hidden.\nOtherwise, the console is toggled on or off.");
  boost::python::def("console", &APIv1::toggleConsole, "If `display` is True, the console is displayed.\nIf `display` is False, the console is hidden.\nOtherwise, the console is toggled on or off.");
  boost::python::def("consoleVisible", &APIv1::consoleVisible, "Returns True if the console is visible, False otherwise.");
  
  // general display functions
  boost::python::def("wxSafeYield", &APIv1::safeYield, "Yields control to pending messages in the windowing system.");
	boost::python::def("viewportWidth", &APIv1::viewportWidth, "Return width of Viewport.");
	boost::python::def("viewportHeight", &APIv1::viewportHeight, "Return height of Viewport.");

  // side panel display
  boost::python::def("sidePanel", &APIv1::displaySidePanel, "If `display` is True, the side panel is displayed.\nIf `display` is False, the side panel is hidden.\nOtherwise, the side panel is toggled on or off.");
  boost::python::def("sidePanel", &APIv1::toggleSidePanel, "If `display` is True, the side panel is displayed.\nIf `display` is False, the side panel is hidden.\nOtherwise, the side panel is toggled on or off.");
  boost::python::def("sidePanelVisible", &APIv1::sidePanelVisible, "Returns True if the side panel is visible, False otherwise.");
  
  // GenGIS 3D Terrain functions
  boost::python::def("viewTop", &APIv1::viewTop, "Sets the 3D terrain to a top-down view.");
  boost::python::def("viewReset", &APIv1::viewReset, "Resets the 3D terrain view.");
  boost::python::def("detail", &APIv1::detail, detail_overloads(boost::python::args("level"), "Adjusts the terrain detail by `number` units; `number` defaults to 1.\nIf `number` is positive the terrain will sharpen, a negative `number` will make the terrain more coarse."));
  boost::python::def("wireframe", &APIv1::displayWireframe, "If `display` is omitted, this function toggles the map between wireframe and smooth.\nIf `display` is True, this function draws the map as a wireframe.\nIf `display` is False, this function draws the map as a smooth texture.");
  boost::python::def("wireframe", &APIv1::toggleWireframe, "If `display` is omitted, this function toggles the map between wireframe and smooth.\nIf `display` is True, this function draws the map as a wireframe.\nIf `display` is False, this function draws the map as a smooth texture.");
  boost::python::def("isWireframe", &APIv1::isWireframe, "Returns True is the terrain is displayed as a wireframe, False otherwise.");
  boost::python::def("refresh", &APIv1::refresh, "Re-draws the terrain.");
  
  // GenGIS control
  boost::python::def("quit", &APIv1::quit, "Quits GenGIS.");
  boost::python::def("openMap", &APIv1::openMap, "Opens the map file at `file`.");
  boost::python::def("openLocations", &APIv1::openLocations, "Opens the samples file at `file`.");
	boost::python::def("openTree", &APIv1::openTree, "Opens the tree file at `file`.");
  boost::python::def("openSequenceData", &APIv1::openSequenceData, "Opens the sequence data file at `file`.");
  boost::python::def("saveImage", &APIv1::saveImage, "Saves an image of the terrain view, with all samples and charts. The image is saved as a png file at `path`.");
  
  // Navigation methods
  boost::python::def("up", &APIv1::up, up_overloads(boost::python::args("number"), "Moves the map `number` units forward; `number` defaults to 1.\n\nThis function is the opposite of down()."));
  boost::python::def("down", &APIv1::down, down_overloads(boost::python::args("number"), "Moves the map `number` units backward; `number` defaults to 1.\n\nThis function is the opposite of up()."));
  boost::python::def("left", &APIv1::left, left_overloads(boost::python::args("number"), "Moves the map `number` units to the left; `number` defaults to 1.\n\nThis function is the opposite of right()."));
  boost::python::def("right", &APIv1::right, right_overloads(boost::python::args("number"), "Moves the map `number` units to the right; `number` defaults to 1.\n\nThis function is the opposite of left()."));
  boost::python::def("rotate", &APIv1::rotate, rotate_overloads(boost::python::args("number"), "Rotates the map `number` degrees counter-clockwise; `number` defaults to 1.\nTo rotate clockwise, `number` should be negative."));
  boost::python::def("pitch", &APIv1::pitch, pitch_overloads(boost::python::args("number"), "Pitches the camera `number` units; `number` defaults to 1.\nIf `number` is positive the camera will pitch up, otherwise the camera will pitch down."));
  boost::python::def("zoom", &APIv1::zoom, zoom_overloads(boost::python::args("number"), "Zooms the camera `number` units; `number` defaults to 1.\nIf `number` is positive the camera will zoom in, otherwise the camera will zoom out."));

  boost::python::def("getPosition", &APIv1::getPosition, boost::python::return_value_policy<boost::python::copy_const_reference>(), "Get position of camera.");
  boost::python::def("setPosition", &APIv1::setPosition, "Set position of camera.");
  boost::python::def("getYaw", &APIv1::getYaw, "Get yaw angle of camera (in degrees).");
  boost::python::def("setYaw", &APIv1::setYaw, "Set yaw angle of camera (in degrees).");
  boost::python::def("getPitch", &APIv1::getPitch, "Get pitch of camera (in degrees).");
  boost::python::def("setPitch", &APIv1::setPitch, "Set pitch of camera (in degrees).");

	// User Graphical element methods
	boost::python::def("addLine", &APIv1::addLine, "Add line to Viewport.");
	boost::python::def("removeLine", &APIv1::removeLine, "Remove line with a given id.");
	boost::python::def("clearLines", &APIv1::clearLines, "Clear all custom lines in Viewport.");
	boost::python::def("addMarker", &APIv1::addMarker, "Add marker to Viewport.");
	boost::python::def("removeMarker", &APIv1::removeMarker, "Remove marker with a given id.");
	boost::python::def("clearMarkers", &APIv1::clearMarkers, "Clear all custom markers in Viewport.");
	boost::python::def("addLabel", &APIv1::addLabel, "Add label to Viewport.");
	boost::python::def("removeLabel", &APIv1::removeLabel, "Remove label with a given id.");
	boost::python::def("clearLabels", &APIv1::clearLabels, "Clear all custom labels in Viewport.");
	boost::python::def("convertGeoCoord", &APIv1::convertGeoCoord, "Convert cartesian geographic coordinates to 3D object coordinates.");
	boost::python::def("convertLatLong", &APIv1::convertLatLong, "Convert lat/long coordinates to 3D object coordinates.");
	 
    
  //=============
  //   Objects
  //=============
 
  // Containers
  boost::python::class_<std::vector<MapLayer::Ptr> >("mapVector")
    .def(boost::python::vector_indexing_suite<std::vector<MapLayer::Ptr>, true>())
  ;
  
  boost::python::class_<std::vector<LocationSetLayerPtr> >("locationSetVector")
    .def(boost::python::vector_indexing_suite<std::vector<LocationSetLayerPtr>, true>())
  ;
  
  boost::python::class_<std::vector<LocationLayerPtr> >("locationVector")
    .def(boost::python::vector_indexing_suite<std::vector<LocationLayerPtr>, true>())
  ;
  
  boost::python::class_<std::vector<SequenceLayer::Ptr> >("sequenceVector")
    .def(boost::python::vector_indexing_suite<std::vector<SequenceLayer::Ptr>, true>())
  ;

	boost::python::class_<std::vector<std::wstring> >("wstringVector")
    .def(boost::python::vector_indexing_suite<std::vector<std::wstring>, true>())
  ;
  
  // map of two wstrings
  boost::python::class_<std::map<std::wstring,std::wstring> >("StringMap")
    .def("__len__", &std::map<std::wstring,std::wstring>::size)
    .def("__getitem__", &map_item<std::wstring,std::wstring>().get, boost::python::return_value_policy<boost::python::copy_non_const_reference>() )
    .def("__setitem__", &map_item<std::wstring,std::wstring>().set)
    .def("__delitem__", &map_item<std::wstring,std::wstring>().del)
    .def("__str__", &map_item<std::wstring,std::wstring>().str)
    .def("__repr__", &map_item<std::wstring,std::wstring>().str)
    .def("clear", &std::map<std::wstring,std::wstring>::clear)
    .def("__contains__", &map_item<std::wstring,std::wstring>().in)
    .def("has_key", &map_item<std::wstring,std::wstring>().in)
    .def("keys", &map_item<std::wstring,std::wstring>().keys)
    .def("values", &map_item<std::wstring,std::wstring>().values)
    .def("items", &map_item<std::wstring,std::wstring>().items)
  ;
  
  // Study
  boost::python::class_<StudyLayer, boost::noncopyable, boost::shared_ptr<StudyLayer> >("Study", boost::python::no_init)
    .def("GetMap", GetMapByIndex)
    .def("GetNumMaps", &StudyLayer::GetNumMaps)
    .def_readonly("maps", &StudyLayer::GetAllMaps)
  ;
  
  // Add the only study to the global scope. By "only" I mean at the moment only 1 study ever exists. This code will have to be changed when support for more than 1 study is added to GenGIS.
  boost::python::scope().attr("study") = App::Inst().GetLayerTreeController()->GetStudyLayer(0);

  // Map
  boost::python::class_<MapLayer, boost::noncopyable, boost::shared_ptr<MapLayer> >("Map", boost::python::no_init)
    .add_property("width", GetWidthFromMapLayer)
    .add_property("height", GetHeightFromMapLayer)
    .add_property("minElevation", GetMinElevationFromMapLayer)
    .add_property("maxElevation", GetMaxElevationFromMapLayer)
    .def("GetLocationSet", GetLocationSetByIndex)
    .def("GetNumLocationSets", &MapLayer::GetNumLocationSets)
    .def_readonly("locationSets", &MapLayer::GetAllLocationSets)
		.def("GetTree", GetTreeByIndex)
		.def("GetNumTrees", &MapLayer::GetNumTrees)
		.def_readonly("trees", &MapLayer::GetAllLocationSets)
  ;

	// Tree
  boost::python::class_<TreeLayer, boost::noncopyable, boost::shared_ptr<TreeLayer> >("Tree", boost::python::no_init)
		.def("GetGeoTreeController", &TreeLayer::GetGeoTreeController)
  ;
  
  // LocationSet
  boost::python::class_<LocationSetLayer, boost::noncopyable, boost::shared_ptr<LocationSetLayer> >("LocationSet", boost::python::no_init)
    .def("GetLocation", GetLocationByIndex)
    .def("GetNumLocations", &LocationSetLayer::GetNumLocations)
		.def("GetChartSet", &LocationSetLayer::GetChartSetView)
    .def_readonly("locations", &LocationSetLayer::GetAllLocations)
  ;
  
  // Location
  boost::python::class_<LocationLayer, boost::noncopyable, boost::shared_ptr<LocationLayer> >("Location", boost::python::no_init)
    .add_property("name", GetNameFromLocationLayer)
    .add_property("id", GetIdFromLocationLayer)
    .add_property("northing", GetNorthingFromLocationLayer)
    .add_property("easting", GetEastingFromLocationLayer)
    .add_property("data", GetDataMapFromLocationLayer)
    .def("GetSequence", &LocationLayer::GetSequence)
    .def("GetNumSequences", &LocationLayer::GetNumSequences)
		.def("GetChart", GetChartLayer)
		.def("GetVisibility", GetVisibilityLocationLayer)
	// ~mikep. Not sure if these "remove" methods are doing anything to the C++ side of the data... is this const or something???
		.def("RemoveSequence", &LocationLayer::RemoveSequence)
		.def("RemoveAllSequences", &LocationLayer::RemoveAllSequences)
		.def("AddSequence", AddSequenceToLocationLayer)
    .def("SetColour", SetColourLocationLayer)
		.def("SetSize", SetSizeLocationLayer)
		.def("SetVisibility", SetVisibilityLocationLayer)
		.def("SetRenderingOrder", SetRenderingOrderLocationLayer)		
    .def_readonly("sequences", &LocationLayer::GetAllSequences)
  ;
  
  // Sequence
  boost::python::class_<SequenceLayer, boost::noncopyable, boost::shared_ptr<SequenceLayer> >("Sequence", boost::python::no_init)
    .add_property("sequenceId", GetSequenceIdFromSequenceLayer)
    .add_property("siteId", GetSiteIdFromSequenceLayer)
    .add_property("data", GetDataMapFromSequenceLayer)
		.def("GetActive", GetActiveSequenceLayer)
		.def("SetActive", SetActiveSequenceLayer)
  ;

	// GeoTreeController
	boost::python::class_<GeoTreeController, boost::noncopyable, boost::shared_ptr<GeoTreeController> >("GeoTreeController", boost::python::no_init)
		.def("GetNumberOfNodes", &GeoTreeController::GetNumberOfNodes)
		.def("GetNumberOfLeaves", &GeoTreeController::GetNumberOfLeaves)
		.def("GetDepthOfTree", &GeoTreeController::GetDepthOfTree)
		.def("IsMultifurcating", &GeoTreeController::IsMultifurcating)
		.def("ProjectTree", &GeoTreeController::ProjectTree)
		.def("RestoreTree", &GeoTreeController::RestoreTree)
		.def("GetLeafNames", &GeoTreeController::GetLeafNames)
	;

	// Chart set view
	boost::python::class_<ChartSetView, boost::noncopyable, boost::shared_ptr<ChartSetView> >("ChartSetView", boost::python::no_init)
		.def("UpdateCharts", &ChartSetView::UpdateCharts)
	;

	// Chart view
	boost::python::class_<ChartView, boost::noncopyable, boost::shared_ptr<ChartView> >("ChartView", boost::python::no_init)
		.def("GetVisibility", &ChartView::IsVisible)
		.def("SetVisibility", &ChartView::SetVisibility)
		.def("GetDepth", &ChartView::GetDepth)
		.def("SetDepth", &ChartView::SetDepth)
		.def("GetWidth", &ChartView::GetWidth)
		.def("SetWidth", &ChartView::SetWidth)
		.def("GetHeight", &ChartView::GetHeight)
		.def("SetHeight", &ChartView::SetHeight)
		.def("GetBorderSize", &ChartView::GetBorderSize)
		.def("SetBorderSize", &ChartView::SetBorderSize)
		.def("GetWedgeBorders", &ChartView::GetWedgeBorders)
		.def("SetWedgeBorders", &ChartView::SetWedgeBorders)
		.def("GetSizeProportionalToSeq", &ChartView::GetSizeProportionalToSeq)
		.def("SetSizeProportionalToSeq", &ChartView::SetSizeProportionalToSeq)
		.def("GetMinSize", &ChartView::GetMinSize)
		.def("SetMinSize", &ChartView::SetMinSize)
		.def("GetMaxSize", &ChartView::GetMaxSize)
		.def("SetMaxSize", &ChartView::SetMaxSize)
		.def("GetAdjustedSize", &ChartView::GetAdjustedSize)
		.def("SetNormalizedNumSeq", &ChartView::SetNormalizedNumSeq)
		.def("SetField", &ChartView::SetField)
		.def("GetField", &ChartView::GetField, boost::python::return_value_policy<boost::python::copy_const_reference>())
		.def("GetNames", &ChartView::GetNames)
		.def("GetPercentage", &ChartView::GetPercentage)
		.def("SetChartType", &ChartView::SetChartType)
		.def("GetChartType", &ChartView::GetChartType)
		.def("SetAssignToOther", &ChartView::SetAssignToOther)
		.def("GetAssignToOther", &ChartView::GetAssignToOther)
		.def("SetCentre", &ChartView::SetCentre)
		.def("GetCentre", &ChartView::GetCentre, boost::python::return_value_policy<boost::python::copy_const_reference>())
		.def("GetLocationPos", &ChartView::GetLocationPos)
		.def("SetScreenPos", &ChartView::SetScreenPos)
		.def("GetScreenPos", &ChartView::GetScreenPos)
		.def("SetLabelBindToChart", &ChartView::SetLabelBindToChart)
		.def("GetLabelBindToChart", &ChartView::GetLabelBindToChart)
		.def("SetLabel", &ChartView::SetLabel)
		.def("GetLabel", &ChartView::GetLabel)
		.def("SetColourMap", &ChartView::SetColourMap)
		.def("GetColourMap", &ChartView::GetColourMap)
		.def("UpdateChart", &ChartView::UpdateChart)
	;

	// Discrete colour map
	boost::python::class_<ColourMapDiscrete, boost::noncopyable, boost::shared_ptr<ColourMapDiscrete> >("ColourMapDiscrete")
		.def("SetColour", SetColourDiscrete)
		.def("GetColour", GetColourDiscrete)
	;

	// Visual label
	boost::python::class_<VisualLabel>("VisualLabel")
		.def(boost::python::init<const std::wstring&, const Colour&, float>())
		.def("GetText", &VisualLabel::GetText, boost::python::return_value_policy<boost::python::copy_const_reference>())
		.def("GetColour", &VisualLabel::GetColour, boost::python::return_value_policy<boost::python::copy_const_reference>())
		.def("GetSize", &VisualLabel::GetSize)
		.def("GetVisibility", &VisualLabel::IsVisible)
		.def("GetGridPosition", &VisualLabel::GetGridPosition, boost::python::return_value_policy<boost::python::copy_const_reference>())
		.def("GetScreenPosition", &VisualLabel::GetScreenPosition, boost::python::return_value_policy<boost::python::copy_const_reference>())
		.def("GetRotateLabel", &VisualLabel::GetRotateLabel)
		.def("GetRotateLabel", &VisualLabel::GetRotateLabel)
		.def("GetRenderingStyle", &VisualLabel::GetRenderingStyle)
		.def("SetText", &VisualLabel::SetText)
		.def("SetColour", &VisualLabel::SetColour)
		.def("SetSize", &VisualLabel::SetSize)
		.def("SetVisibility", &VisualLabel::SetVisibility)
		.def("SetGridPosition", &VisualLabel::SetGridPosition)
		.def("SetScreenPosition", &VisualLabel::SetScreenPosition)
		.def("SetRenderingStyle", &VisualLabel::SetRenderingStyle)
	;

	// Visual marker
	boost::python::class_<VisualMarker>("VisualMarker")
		.def(boost::python::init<const Colour&, float, VisualMarker::MARKER_SHAPE, const Point3D&>())
		.def("GetPosition", &VisualMarker::GetPosition, boost::python::return_value_policy<boost::python::copy_const_reference>())
		.def("GetShape", &VisualMarker::GetShape)
		.def("GetRotation", &VisualMarker::GetRotation)
		.def("GetDepth", &VisualMarker::GetDepth)
		.def("GetRenderingStyle", &VisualMarker::GetRenderingStyle)
		.def("GetColour", &VisualMarker::GetColour, boost::python::return_value_policy<boost::python::copy_const_reference>())
		.def("GetSize", &VisualMarker::GetSize)
		.def("GetVisibility", &VisualMarker::IsVisible)
		.def("GetRenderingOrder", &VisualMarker::GetRenderingOrder)
		.def("SetPosition", &VisualMarker::SetPosition)
		.def("SetShape", &VisualMarker::SetShape)
		.def("SetRotation", &VisualMarker::SetRotation)
		.def("SetDepth", &VisualMarker::SetDepth)
		.def("SetRenderingStyle", &VisualMarker::SetRenderingStyle)
		.def("SetColour", &VisualMarker::SetColour)
		.def("SetSize", &VisualMarker::SetSize)
		.def("SetVisibility", &VisualMarker::SetVisibility)
		.def("SetRenderingOrder", &VisualMarker::SetRenderingOrder)
	;

	// Visual line
	boost::python::class_<VisualLine>("VisualLine")
		.def(boost::python::init<const Colour&, float, VisualLine::LINE_STYLE, const Line3D&>())
		.def("GetColour", &VisualLine::GetColour, boost::python::return_value_policy<boost::python::copy_const_reference>())
		.def("GetSize", &VisualLine::GetSize)
		.def("GetVisibility", &VisualLine::IsVisible)
		.def("GetLineStyle", &VisualLine::GetLineStyle)
		.def("GetStartPt", &VisualLine::GetStartPt, boost::python::return_value_policy<boost::python::copy_const_reference>())
		.def("GetEndPt", &VisualLine::GetEndPt, boost::python::return_value_policy<boost::python::copy_const_reference>())
		.def("SetColour", &VisualLine::SetColour)
		.def("SetSize", &VisualLine::SetSize)
		.def("SetVisibility", &VisualLine::SetVisibility)
		.def("SetLineStyle", &VisualLine::SetLineStyle)
		.def("SetStartPt", &VisualLine::SetStartPt)
		.def("SetEndPt", &VisualLine::SetEndPt)
	;

  // Colour
  boost::python::class_<Colour>("Colour")
    .def(boost::python::init<float, float, float, boost::python::optional<float> >())
		.def("GetRed", &Colour::GetRed)
		.def("GetGreen", &Colour::GetGreen)
		.def("GetBlue", &Colour::GetBlue)
		.def("GetAlpha", &Colour::GetAlpha)
		.def("SetRGBA", &Colour::SetRGBA)
	;

	// Point3D
  boost::python::class_<Point3D>("Point3D")
		.def(boost::python::init<float, float, float>())
		.def_readwrite("x", &Point3D::x)
		.def_readwrite("y", &Point3D::y)
		.def_readwrite("z", &Point3D::z)
  ;

	// Line3D
	boost::python::class_<Line3D>("Line3D")
		.def(boost::python::init<const Point3D&, const Point3D&>())
		.def_readwrite("start", &Line3D::start)
		.def_readwrite("end", &Line3D::end)
  ;
}

PythonInterface::PythonInterface()
{
  // Initialize Python
  Py_Initialize();
  PyEval_InitThreads();

  // WINDOWS: Add GenGIS\bin\lib\site-packages to Python's path, so it can find the Python modules (ie wxPython) contained within.
  PyRun_SimpleString("import sys\nsys.path.insert(0,\".\\lib\\site-packages\")");
  // MAC: Add site-packages directory to sys.path
  PyRun_SimpleString("import sys\nsys.path.insert(0,\"./GenGIS.app/Contents/Resources/site-packages\")");
  
  // Load the wxPython core API.
  if(!wxPyCoreAPI_IMPORT())
  {
    wxLogError(wxT("Error importing the wxPython API!"));
    PyErr_Print();
    Py_Finalize();
    return;
  }
  
  // Save the current Python thread state and release the Global Interpreter Lock.
  // m_mainTState = wxPyBeginAllowThreads();         // ~mikep. This make everything crash. When I get the time, find out why and fix it......
  m_mainTState = NULL;

  // Add Python to GenGIS:
  ((GenGisFrame*)GenGIS::App::Inst().GetMainWindow())->AddConsolePanel(GetPythonPanel(), _T("Python"));
  
  // Initialize the Boost module:
  try
  {
    initGenGisAPIv1();                                          // Tell Boost to initialize the module (all that boost code above).
    // ~mikep. This is sketchy: we're loading in the rpy methods from Python, adding them to the boost Location class, then deleting them. See if there's a cleaner way to do this.
    PyRun_SimpleString("from GenGisAPIv1 import *");            // Import the Boost module's contents into Python, at the global scope.
    PyRun_SimpleString("from PythonCode import plotLocation");  // Load Python functions (written in PythonCode.py).
    PyRun_SimpleString("Location.plot = plotLocation");         // Add plotting methods to Location object.
    PyRun_SimpleString("del plotLocation");                     // Delete the old methods from the global scope.
		
		// add scripts directory to Python path
		wxString scriptDirCmd = wxT("sys.path.insert(0,'") + App::Inst().GetExeDir() + wxT("scripts") + wxT("')");
		std::string str = std::string(scriptDirCmd.begin(), scriptDirCmd.end());
		str = util::StringTools::ReplaceChar(str, '\\', '/');
		PyRun_SimpleString(str.c_str());
  }
  catch(boost::python::error_already_set)
  {
    PyErr_Print();
  }
}

PythonInterface::~PythonInterface()
{
  // Tell wxPython to close and clean up.
  // wxPyEndAllowThreads(m_mainTState);     // ~mikep. had to comment this line out along with wxPyBeginAllowThreads() (which occurs later in this file)
  // Close the Python interpreter.
  Py_Finalize();
}

// Adapted from the samples in wxPython
wxPanel* PythonInterface::GetPythonPanel()
{
  wxWindow* parent = ((GenGisFrame*)GenGIS::App::Inst().GetMainWindow())->GetBottomNotebook();
  
  //wxPanel* pythonPanel = NULL;
  PyObject* result;
  wxPyBlock_t blocked = wxPyBeginBlockThreads(); // Get the Global Interpreter Lock.
  

  PyObject* globals = PyDict_New();
  PyObject* builtins = PyImport_ImportModule("__builtin__");
  PyDict_SetItemString(globals, "__builtins__", builtins);
  Py_DECREF(builtins);

  // Make a method that makes the wxPython panel and return it.
  char* pythonPanelCode = "import sys\nsys.path.insert(0,'./GenGIS.app/Contents/Resources') # Mac location \nsys.path.insert(0,'.') # Windows location \nfrom PythonCode import PythonInterpreter\n\ndef makePanel(parent):\n  pyPanel = PythonInterpreter(parent)\n  return pyPanel\n";
  result = PyRun_String(pythonPanelCode, Py_file_input, globals, globals);
  
  // Check for an exception.
  if(!result)
  {
    PyErr_Print();
    wxPyEndBlockThreads(blocked);
    return NULL;
  }
  Py_DECREF(result);

  // Get the 'makePanel' method
  PyObject* func = PyDict_GetItemString(globals, "makePanel");
  wxASSERT(PyCallable_Check(func));

  // Call makePanel
  PyObject* arg = wxPyMake_wxObject(parent, false);
  wxASSERT(arg != NULL);
  PyObject* tuple = PyTuple_New(1);
  PyTuple_SET_ITEM(tuple, 0, arg);
  result = PyEval_CallObject(func, tuple);

  // Check for an exception.
  if(!result)
  {
    PyErr_Print();
  }
  else
  {
    // Bring the wxPython wxPanel into GenGis
    bool success = wxPyConvertSwigPtr(result, (void**)&m_pythonPanel, _T("wxPanel"));

		//pythonPanel->SetWindowStyle(wxBORDER_NONE);
		//pythonPanel->Refresh();

    wxASSERT_MSG(success, _T("Returned object was not a wxPanel!"));
    Py_DECREF(result);
  }

  // Release the python objects we still have
  Py_DECREF(globals);
  Py_DECREF(tuple);

  // Release the Global Interpreter Lock.
  wxPyEndBlockThreads(blocked);

  return m_pythonPanel;
}
#endif