//=======================================================================
// Author: Donovan Parks
//
// Copyright 2009 Donovan Parks
//
// This file is part of GenGIS.
//
// GenGIS is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// GenGIS is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with GenGIS.  If not, see <http://www.gnu.org/licenses/>.
//=======================================================================

#include "../core/Precompiled.hpp"

#include "../utils/DemReader.hpp"
#include "../utils/Log.hpp"
#include "../utils/StringTools.hpp"
#include "../utils/Colour.hpp"

#include "../core/StudyController.hpp"
#include "../core/TerrainModel.hpp"
#include "../core/TerrainTexture.hpp"

#include "../gui/ProgressDlg.hpp"

using namespace GenGIS;

bool DemReader::Open(std::wstring fullPath, TerrainModelPtr terrainModel, StudyControllerPtr studyController, 
											ProgressDlgPtr progressDlg) 
{    
	GDALAllRegister();

	std::string tempStr(fullPath.begin(), fullPath.end());
	GDALDataset* gdalDataset = (GDALDataset*) GDALOpen( tempStr.c_str(), GA_ReadOnly );

	if(gdalDataset == NULL) 
	{
		error::Log::Inst().Warning("(Warning) Failed to open file at " + tempStr + ".");
		return false;	
	}

	// determine type of file based on number of raster bands	
	bool bElevationMap;
	uint numBands = gdalDataset->GetRasterCount();
	if(numBands == 1)
		bElevationMap = true;
	else if(numBands == 3 || numBands == 4)
		bElevationMap = false;
	else
	{
		error::Log::Inst().Warning("(Warning) File type not supported. Please contact the GenGIS to request support for the file format.");
	}

	double adfGeoTransform[6];
	gdalDataset->GetGeoTransform( adfGeoTransform );

	GDALRasterBand* gdalBand = gdalDataset->GetRasterBand( 1 );

	int nBlockXSize, nBlockYSize;
	gdalBand->GetBlockSize( &nBlockXSize, &nBlockYSize );

	// get original data dimensions
	FileHeader* header = terrainModel->GetHeader();
	header->nCols = gdalBand->GetXSize();
	header->nRows = gdalBand->GetYSize();
	header->noDataValue = gdalBand->GetNoDataValue();

	// Setup desired projection 
	if(!SetupProjection(gdalDataset, studyController, adfGeoTransform, header->nCols, header->nRows, bElevationMap))
	{
		GDALClose(gdalDataset);
		return false;
	}

	if(progressDlg)
	{
		if(!progressDlg->Update(0, _T("Reading map file...")))
		{
			GDALClose(gdalDataset);
			return false;
		}
	}

	// check whether it's too large to handle all the data
	float* elevations = NULL;
	if(bElevationMap)
	{
		try
		{
			elevations = new float[gdalBand->GetXSize()* gdalBand->GetYSize()];
		}
		catch(std::bad_alloc&)
		{
			error::Log::Inst().Warning("(Warning) Insufficent memory for elevation map."); 

			GDALClose(gdalDataset);
			return false;
		}
		
		gdalBand->RasterIO( GF_Read, 0, 0, gdalBand->GetXSize(), gdalBand->GetYSize(), 
												elevations, gdalBand->GetXSize(), gdalBand->GetYSize(), 
												GDT_Float32, 
												0, 0 );
	}

	// project all terrain model cells
	Point3D* gridArray;
	try
	{
		gridArray = new Point3D[header->nCols*header->nRows];
	}
	catch(std::bad_alloc&)
	{
		error::Log::Inst().Warning("(Warning) Insufficent memory to load map."); 

		GDALClose(gdalDataset);
		delete[] elevations;
		return false;
	}

	// populate grid with projected points and find extents of projected map
	Point3DArrayPtr grid( gridArray );
	double xOffset = adfGeoTransform[0];
	double zOffset = adfGeoTransform[3];
	int index = 0;
	double x, z;
	header->projExtents.x = header->projExtents.y = std::numeric_limits<float>::max();
	header->projExtents.dx = header->projExtents.dy = -std::numeric_limits<float>::max();
	ProjectionToolPtr projTool = studyController->GetProjectionTool();
	for(int m = 0; m < header->nRows ; ++m) 
	{
		xOffset = adfGeoTransform[0];
		for(int n = 0; n < header->nCols; ++n) 
		{
			x = xOffset;
			z = zOffset;
			if(studyController->GetProject())
			{			
				double elevation = (double)elevations[index];
				if(!projTool->Transform(1, &x, &z, &elevation))
				{
					error::Log::Inst().Warning("(Warning) Failed to project map.");
					GDALClose(gdalDataset);
					delete[] elevations;
					return false;
				}
			}

			xOffset += adfGeoTransform[1];

			float X = (float)x;
			float Z = (float)z;

			grid[index].x = X;
			grid[index].z = Z;
			index++;

			if(X < header->projExtents.x) header->projExtents.x = X;
			if(Z < header->projExtents.y) header->projExtents.y = Z;
			if(X > header->projExtents.dx) header->projExtents.dx = X;
			if(Z > header->projExtents.dy) header->projExtents.dy = Z;
		}

		zOffset += adfGeoTransform[5];

		if(progressDlg)
		{
			if(!progressDlg->Update(int((float(m)/header->nRows)*50)))
			{
				GDALClose(gdalDataset);
				if(elevations != NULL)
					delete[] elevations;

				return false;
			}
		}
	}

	// transform all points into unit grid space
	float minElevation = std::numeric_limits<float>::max();
	float maxElevation = -std::numeric_limits<float>::max();
	if(!TransformToGridSpace(grid, header, elevations, bElevationMap, minElevation, maxElevation, progressDlg))
	{
		GDALClose(gdalDataset);
		if(elevations != NULL)
			delete[] elevations;

		return false;
	}

	// setup terrain model
	terrainModel->SetMinElevationGridSpace(minElevation);
	terrainModel->SetMaxElevationGridSpace(maxElevation);

	terrainModel->SetMinElevation(minElevation/header->scaleFactor);
	terrainModel->SetMaxElevation(maxElevation/header->scaleFactor);

	terrainModel->SetGrid(grid);
	terrainModel->SetElevationMap(bElevationMap);

	// report information in file to user
	MetaDataInfo(gdalDataset, terrainModel);

	// build texture that will be mapped onto terrain
	if(!BuildTerrainTexture(gdalDataset, terrainModel, progressDlg))
	{
		GDALClose(gdalDataset);	
		if(elevations != NULL)
			delete[] elevations;

		return false;
	}

	// will free memory allocated to any GDALRasterBand objects
	GDALClose(gdalDataset);	
	if(elevations != NULL)
		delete[] elevations;

	return true;
}

bool DemReader::TransformToGridSpace(Point3DArrayPtr grid, FileHeader* header, float* elevations, bool bElevationMap, 
																		 float& minElevation, float& maxElevation, ProgressDlgPtr progressDlg)
{
	// model (grid) space for map should go from x = -1 to 1, and z = -aspect to aspect
	double aspect = header->projExtents.Height() / header->projExtents.Width();
	header->cellSizeX = 2.0 / header->nCols;
	header->cellSizeZ = 2.0*aspect / header->nRows;
	header->width = 2.0;
	header->height = 2.0*aspect;
	header->scaleFactor = 2.0 / header->projExtents.Width();

	// transform all points into grid space
	uint index = 0;
	minElevation = std::numeric_limits<float>::max();
	maxElevation = -std::numeric_limits<float>::max();
	for(int m = 0; m < header->nRows ; m++) 
	{
		for(int n = 0; n < header->nCols; n++) 
		{
			float X = grid[index].x;
			float Z = grid[index].z;

			grid[index].x = -1.0f + 2*((X - header->projExtents.x) / header->projExtents.Width());
			grid[index].z = (1.0f + -2*((Z - header->projExtents.y) / header->projExtents.Height()))*aspect;

			if(bElevationMap)
			{
				double elevation = elevations[index];
				if(elevation == header->noDataValue)
				{
					grid[index].y = 0;
					if(grid[index].y < minElevation)
						minElevation = grid[index].y;
				}
				else
				{
					grid[index].y = elevation * header->scaleFactor;

					if(grid[index].y > maxElevation)
						maxElevation = grid[index].y;
					else if(grid[index].y < minElevation)
						minElevation = grid[index].y;
				}	
			}
			else
			{
				grid[index].y = 0;
				minElevation = maxElevation = 0;
			}

			index++;
		}

		if(progressDlg)
		{
			if(!progressDlg->Update(int(50 + (float(m)/header->nRows)*50)))
				return false;
		}
	}

	return true;
}

bool DemReader::SetupProjection(GDALDataset* gdalDataset, StudyControllerPtr studyController, double* adfGeoTransform, uint nCols, uint nRows, bool bElevationMap)
{
	OGRSpatialReference oSource;
	ProjectionToolPtr projTool;

	// getting the string about projection system for this dataset.
	const char* currentWkt = gdalDataset->GetProjectionRef();

	if((currentWkt != NULL && oSource.importFromWkt( (char**)&currentWkt ) != OGRERR_CORRUPT_DATA)) 
	{
		studyController->SetProject(true);

		if(oSource.IsGeographic())
		{
			// lat/lon coordinates are using the geographic projection (aka, plate carre) 
			error::Log::Inst().Write("Loading map with lat/long coordinates.");

			// determine centre of map
			double longStart = adfGeoTransform[0];
			double longEnd = adfGeoTransform[0] + nCols*adfGeoTransform[1];
			studyController->SetCentreLongitude(float(longStart + fabs(longStart - longEnd) / 2.0));

			double latStart = adfGeoTransform[3] + nRows*adfGeoTransform[5];
			double latEnd = adfGeoTransform[3];
			studyController->SetCentreLatitude(float(latStart + fabs(latStart - latEnd) / 2.0));

			studyController->SetFirstStandardParallel(latStart);
			studyController->SetSecondStandardParallel(latEnd);

			projTool = studyController->GetProjectionTool(&oSource);
			if(projTool == ProjectionToolPtr())
				return false;
		}
		else if(oSource.IsProjected())
		{
			error::Log::Inst().Write("Loading map with projected coordinates.");
			studyController->SetProject(false);
		}
		else
		{
			// The user should probably be made aware of this warning.
			error::Log::Inst().Warning("(Warning) Unknown type of coordinate system.");
			return false;
		}
	}
	else 
	{
		if(bElevationMap)
		{
			// The user should probably be made aware of this warning.
			error::Log::Inst().Warning("(Warning) Well-Known text which specified coordinate system information is missing or not recognized."); 
			error::Log::Inst().Warning("(Warning) The projection information for this study is being ignored.");
			error::Log::Inst().Warning("(Warning) Location sites and leaf nodes must be specified in the same coordinate system as this map.");
			error::Log::Inst().Warning("");
		}
		else
		{
			// Right now we assume that any 3-channel map (i.e., non-elevational) should be projected using the 
			// the geographic projection (aka, plate carre) 
			error::Log::Inst().Write("Coordinate system information is not available for this map.");
			error::Log::Inst().Write("As a result, the projection information for this study is being ignored.");
			error::Log::Inst().Write("");
		}

		studyController->SetProject(false);
	}

	return true;
}

void DemReader::MetaDataInfo(GDALDataset* gdalDataset, TerrainModelPtr terrainModel)
{
	using namespace error;
	using namespace util;

	FileMetaData* header = terrainModel->GetMetaData();

	header->driverDesc = gdalDataset->GetDriver()->GetDescription();
	header->driverMetaData = gdalDataset->GetDriver()->GetMetadataItem(GDAL_DMD_LONGNAME);
	Log::Inst().Write("Driver: " + header->driverDesc + " \\ " + header->driverMetaData);

	header->xSize = gdalDataset->GetRasterXSize();
	header->ySize = gdalDataset->GetRasterYSize();
	header->bands = gdalDataset->GetRasterCount();
	Log::Inst().Write("Size (x,y): " + 
		StringTools::ToString(header->xSize) + ", " +
		StringTools::ToString(header->ySize));
	Log::Inst().Write("Bands: " + StringTools::ToString(header->bands));

	header->projection = std::string(gdalDataset->GetProjectionRef());
	Log::Inst().Write("Projection: " + header->projection);

	double adfGeoTransform[6];
	gdalDataset->GetGeoTransform( adfGeoTransform );

	header->originX = adfGeoTransform[0];
	header->originY = adfGeoTransform[3];
	Log::Inst().Write("Origin: " +
		StringTools::ToString(adfGeoTransform[0]) + ", " +
		StringTools::ToString(adfGeoTransform[3]));

	header->pixelSizeX = adfGeoTransform[1];
	header->pixelSizeY = adfGeoTransform[5];
	Log::Inst().Write("Pixel Size: " +
		StringTools::ToString(adfGeoTransform[1]) + ", " +
		StringTools::ToString(adfGeoTransform[5]));
	
	GDALRasterBand* gdalBand = gdalDataset->GetRasterBand(1);
	int nBlockXSize, nBlockYSize;
	gdalBand->GetBlockSize(&nBlockXSize, &nBlockYSize);

	header->dataType = std::string(GDALGetDataTypeName(gdalBand->GetRasterDataType()));
	Log::Inst().Write("Block, Type: " + StringTools::ToString(nBlockXSize) + ", " +
		StringTools::ToString(nBlockYSize) + ", " +
		std::string(header->dataType));
		
	header->colourInterpretation = std::string(GDALGetColorInterpretationName(gdalBand->GetColorInterpretation()));	
	Log::Inst().Write("Color Interpretation: " + header->colourInterpretation);

	header->extents.x = adfGeoTransform[0];
	header->extents.y = adfGeoTransform[3] + gdalBand->GetYSize()*adfGeoTransform[5];
	Log::Inst().Write("Lower, Left (x,y): " +
		StringTools::ToString(header->extents.x) + ", " +
		StringTools::ToString(header->extents.y));

	header->extents.dx = adfGeoTransform[0]+gdalBand->GetXSize()*adfGeoTransform[1];
	header->extents.dy = adfGeoTransform[3];
	Log::Inst().Write("Upper, Right (x,y): " +
		StringTools::ToString(header->extents.dx) + ", " +
		StringTools::ToString(header->extents.dy));

	Log::Inst().Write("");
}

bool DemReader::BuildTerrainTexture(GDALDataset* gdalDataset, TerrainModelPtr terrainModel, ProgressDlgPtr progressDlg)
{
	uint numBands = gdalDataset->GetRasterCount();

	TerrainTexturePtr texture(new TerrainTexture(terrainModel));

	// check if GDAL dataset specifies colour information for each cell
	bool bTerrain;
	if(numBands == 3 || numBands == 4)
	{
		try
		{
			GDALRasterBand* gdalBandRed = gdalDataset->GetRasterBand( 1 );
			byte* red = new byte[gdalBandRed->GetXSize() * gdalBandRed->GetYSize()];
			gdalBandRed->RasterIO( GF_Read, 0, 0, gdalBandRed->GetXSize(), gdalBandRed->GetYSize(), 
													red, gdalBandRed->GetXSize(), gdalBandRed->GetYSize(), 
													GDT_Byte, 0, 0 );

			GDALRasterBand* gdalBandGreen = gdalDataset->GetRasterBand( 2 );
			byte* green = new byte[gdalBandGreen->GetXSize() * gdalBandGreen->GetYSize()];
			gdalBandGreen->RasterIO( GF_Read, 0, 0, gdalBandGreen->GetXSize(), gdalBandGreen->GetYSize(), 
													green, gdalBandGreen->GetXSize(), gdalBandGreen->GetYSize(), 
													GDT_Byte, 0, 0 );

			GDALRasterBand* gdalBandBlue = gdalDataset->GetRasterBand( 3 );
			byte* blue = new byte[gdalBandBlue->GetXSize() * gdalBandBlue->GetYSize()];
			gdalBandBlue->RasterIO( GF_Read, 0, 0, gdalBandBlue->GetXSize(), gdalBandBlue->GetYSize(), 
													blue, gdalBandBlue->GetXSize(), gdalBandBlue->GetYSize(), 
													GDT_Byte, 0, 0 );

			if(gdalBandRed->GetXSize() != gdalBandGreen->GetXSize() || gdalBandRed->GetXSize() != gdalBandBlue->GetXSize()
						|| gdalBandRed->GetYSize() != gdalBandGreen->GetYSize() || gdalBandRed->GetYSize() != gdalBandBlue->GetYSize())
			{
				error::Log::Inst().Error("DemReader::BuildTerrainTexture(): Failed to build terrain texture.");
				delete[] red;
				delete[] green;
				delete[] blue;

				return false;
			}
	
			bTerrain = texture->BuildTerrainTexture(red, green, blue, progressDlg);

			// free memory
			delete[] red;
			delete[] green;
			delete[] blue;
			red = green = blue = NULL;
		}
		catch(std::bad_alloc&)
		{
			error::Log::Inst().Warning("(Warning) Insufficent memory for colour map."); 
			return false;
		}
	}
	else
	{
		bTerrain = texture->BuildTerrainTexture(progressDlg);
	}
	
	if(!bTerrain)
	{
		error::Log::Inst().Error("DemReader::BuildTerrainTexture(): Failed to build terrain texture.");
		return false;
	}

	terrainModel->SetTexture(texture);

	return true;
}
