//=======================================================================
// Author: Donovan Parks
//
// Copyright 2009 Donovan Parks
//
// This file is part of GenGIS.
//
// GenGIS is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// GenGIS is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with GenGIS.  If not, see <http://www.gnu.org/licenses/>.
//=======================================================================

#include "../core/Precompiled.hpp"

#include "../utils/GraphVizIO.hpp"
#include "../core/TerrainModel.hpp"

#include "../utils/Log.hpp"

using namespace GenGIS;

bool GraphVizIO::Write(const std::wstring& path, const std::wstring& filename, Tree<NodeGeoTree>::Ptr tree, TerrainModelPtr map, bool bFixLeafNodes)
{
	std::wstring fullPath = path + _T("/") + filename;
  char* tempFullPath = 0;
  sprintf(tempFullPath,"%ls",fullPath.c_str());  // Convert the wstring to a const char*
	std::wofstream output(tempFullPath, std::ios::out);
	if(!output.is_open()) 
	{
		error::Log::Inst().Warning("(Warning) Failed to create file.");	
		return false;
	}

	output << "graph G {" << std::endl;
	output << "size =\"5,5\";" << std::endl;
	output << "ratio=compress;" << std::endl;

	float mapWidth = map->GetHeader()->width;
	float mapHeight = map->GetHeader()->height;

	// traverse tree in breadth first order
	std::vector<NodeGeoTree*> curNodes;
	curNodes.push_back(tree->GetRootNode());
	while(!curNodes.empty())
	{
		std::vector<NodeGeoTree*> nextNodes;	
		for(unsigned int curIndex = 0; curIndex < curNodes.size(); ++curIndex)
		{
			std::vector<NodeGeoTree*> childNodes = curNodes.at(curIndex)->GetChildren();
			if(!curNodes.at(curIndex)->IsLeaf())
			{
				output << curNodes.at(curIndex)->GetId() << " [shape=point, color=red] " << std::endl;
			}

			for(unsigned int childIndex = 0; childIndex < childNodes.size(); ++childIndex)
			{
				NodeGeoTree* childNode = childNodes.at(childIndex);

				output << curNodes.at(curIndex)->GetId() << " -- " << childNode->GetId() << ";" << std::endl;
				if(bFixLeafNodes && childNode->IsLeaf())
				{
					Point3D coord = childNode->GetGridCoord();

					// normalize coordinates between [0, 5]
					float x = 5*((coord.x + 0.5*mapWidth) / mapWidth);
					float z = 5 - 5*((coord.z + 0.5*mapHeight) / mapHeight);

					output << childNode->GetId() << " [shape=point, color=blue, pos = \"" << x <<"," << z << "!\" ];" << std::endl;
				}

				nextNodes.push_back(childNode);
			}
		}

		curNodes = nextNodes;
	}


	output << "}" << std::endl;
	
	output.close();

	return true;
}

