//=======================================================================
// Author: Donovan Parks
//
// Copyright 2009 Donovan Parks
//
// This file is part of GenGIS.
//
// GenGIS is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// GenGIS is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with GenGIS.  If not, see <http://www.gnu.org/licenses/>.
//=======================================================================

#include "../core/Precompiled.hpp"

#include "NewickIO.hpp"

#include "../utils/Log.hpp"
#include "../utils/StringTools.hpp"
#include "../utils/NewickTokenizer.hpp"

using namespace GenGIS;
using namespace std;

bool NewickIO::Read(Tree<NodePhylo>::Ptr tree, istream & in)
{	
	m_numElement = 0;
	m_processedElement = 0;

	// Checking the existence of specified input stream
	if(!in)
	{
		error::Log::Inst().Warning("(Warning) Failed to open Newick file.");
		return false;
	}	
	
	// We concatenate all line in file till we reach the ending semi colon:
	string temp, description;

	// Read entire input stream
	while (!in.eof())
  {
		getline(in, temp, '\n');  // Copy current line in temporary string
    string::size_type index = temp.find(";");
		if(index != string::npos)
    {
			description += temp.substr(0, index + 1);
			break;
		}
    else description += temp;
	}

	description = util::StringTools::RemoveSubstrings(description, '[', ']');

	return ParseTree(tree, description);
}

bool NewickIO::Read(Tree<NodePhylo>::Ptr tree, const string& filename)
{
	// Set name of tree to the filename
	string file = util::StringTools::ReplaceChar(filename, '\\', '/');
	string::size_type slashPos = file.rfind('/');
	string::size_type dotPos = file.rfind('.');
	if(dotPos > slashPos)
		tree->SetName(util::StringTools::ToStringW(file.substr(slashPos+1, dotPos-slashPos-1)));
	else
		tree->SetName(_T(""));

	// Parse Newick file
	ifstream input(file.c_str(), std::ios::in);
	bool bLoaded = Read(tree, input);
	input.close();

	return bLoaded;
}

void NewickIO::Write(Tree<NodePhylo>::Ptr tree, ostream& out) const
{
	// Checking the existence of specified file, and possibility to open it in write mode
	assert(out != NULL);

	out << "(";

	NodePhylo* root = tree->GetRootNode();
	
  if(tree->GetNumberOfLeaves() == 0)
  {
    out << "'" << root->GetName().c_str() << "'";

		float dist = root->GetDistanceToParent();
		if(dist != NodePhylo::NO_DISTANCE)
			out << " " << dist;
  }
  else
  {
		WriteNodes(tree, out, root);
  }
	out << ")";
	
	// Output the name of the root if it has one
	if(!util::StringTools::IsEmpty(root->GetName()))
		out << "'" << root->GetName().c_str() << "'";
	
	out << ";" << endl;
}

void NewickIO::WriteNodes(Tree<NodePhylo>::Ptr tree, ostream& out, NodePhylo* parent) const
{
	vector<unsigned int> childrenId = parent->GetChildrenIds();
	assert(childrenId.size() > 0);

	WriteElements(tree, out, parent, parent->GetChild(0));
	for(uint i = 1; i < childrenId.size(); i++)
	{
		out << ",";
		WriteElements(tree, out, parent, parent->GetChild(i));
	}
}

void NewickIO::WriteElements(Tree<NodePhylo>::Ptr tree, ostream& out, NodePhylo* parent, NodePhylo* child) const
{	
  if(child->GetNumberOfChildren() != 0)
  {
    out << "(";
		WriteNodes(tree, out, child); 
    out << ")";
  }

	if(!util::StringTools::IsEmpty(child->GetName()))
		out << "'" << child->GetName().c_str() << "'";
	
	if(child->GetBootstrapToParent() != NodePhylo::NO_DISTANCE)
		out << " " << ((NodePhylo*)child)->GetBootstrapToParent();

	if(child->GetDistanceToParent() != NodePhylo::NO_DISTANCE)
		out << ":" << child->GetDistanceToParent();
}

NewickIO::Element NewickIO::Elements(const string& description)
{
  Element element;
	element.node			= "";
	element.name			= "";
  element.length    = "";
  element.bootstrap = "";

	string desc = description;

	// Check if description is a node or leaf
	string::size_type lastP = desc.rfind(')');
  string::size_type firstP = desc.find('(');
	if(firstP != string::npos)
	{
		//This is a node:
		if(lastP < firstP)
		{
			wxMessageBox(_T("Incorrectly formated Newick string"), _T("Failed Newick String."), wxOK | wxICON_ERROR);
			return element;
		}


		element.node = desc.substr(firstP + 1, lastP - firstP - 1);
		desc = util::StringTools::RemoveSurroundingWhiteSpaces(desc.substr(lastP + 1));
	}
  
	// Check if this element has length
	string::size_type colon = desc.rfind(':');
  if(colon != string::npos)
  {
    element.length = util::StringTools::RemoveSurroundingWhiteSpaces(desc.substr(colon + 1));
		desc = util::StringTools::RemoveSurroundingWhiteSpaces(desc.substr(0, colon));
  }

	// Check for name and bootstrap value
	lastP = desc.rfind('\'');
	firstP = desc.find('\'');
	if(firstP != string::npos)
	{
		// Quoted name:
		element.name = desc.substr(firstP+1, lastP-firstP-1);
		element.name = util::StringTools::ReplaceChar(element.name, '_', ' ');
		element.bootstrap = util::StringTools::RemoveSurroundingWhiteSpaces(desc.substr(lastP+1));
	}
	else
	{
		// No quotations:
		firstP = desc.find(' ');
		if(firstP != string::npos)
		{
			// Parse the name and bootstrap value
			element.name = desc.substr(0, firstP-1);		
			element.bootstrap = util::StringTools::RemoveSurroundingWhiteSpaces(desc.substr(firstP+1));
		}
		else
		{
			// If the remaining description is a decimal number it is assumed it is a bootstrap
			// value, otherwise it is assumed it is a name
			if(util::StringTools::IsDecimalNumber(desc))
			{
				element.bootstrap = desc;
			}
			else
			{
				element.name = desc;
			}
		}
	}
 
  return element;
}  

bool NewickIO::ParseNode(Tree<NodePhylo>::Ptr tree, const string& description, NodePhylo* parent)
{
	Element element = Elements(description);
	m_processedElement++;

	// update loading progress and check if user has cancelled the load
	if(m_conLoadingProgress.connected())
	{
		bool bTemp;
		if(!m_sigLoadingProgress((int)(float(m_processedElement*100)/m_numElement + 0.5), "", &bTemp))
			return false;
	}

	// New node:
	NodePhylo* node(new NodePhylo(m_processedElement));
	parent->AddChild(node);

	node->SetName(_T(""));
	if(!util::StringTools::IsEmpty(element.name))
	{
		node->SetName(util::StringTools::ToStringW(element.name));
	}

	node->SetDistanceToParent(NodePhylo::NO_DISTANCE);
	if(!util::StringTools::IsEmpty(element.length))
	{
		node->SetDistanceToParent(util::StringTools::ToDouble(element.length));
	}

	node->SetBootstrapToParent(NodePhylo::NO_DISTANCE);
	if(!util::StringTools::IsEmpty(element.bootstrap))
	{
		node->SetBootstrapToParent(util::StringTools::ToDouble(element.bootstrap));
	}
  
	// Parse rest of tree
	NewickTokenizer nt(element.node);

	if(nt.NumTokens() != 1)
	{
		for(uint i = 0; i < nt.NumTokens(); i++)
		{
			if(!ParseNode(tree, nt.Token(i), node))
				return false;
		}
	}

	return true;
}

bool NewickIO::ParseTree(Tree<NodePhylo>::Ptr tree, const string& description)
{
  string::size_type lastP  = description.rfind(')');
	if(lastP == string::npos)
	{
		wxMessageBox(_T("Missing closing parenthesis."), _T("Failed Newick String."), wxOK | wxICON_ERROR);
		return false;
	}

  string::size_type firstP = description.find('(');
	if(firstP == string::npos)
	{
		wxMessageBox(_T("Missing opening parenthesis."), _T("Failed Newick String."), wxOK | wxICON_ERROR);
		return false;
	}

  string::size_type semi = description.rfind(';');
	if(semi == string::npos)
	{
		wxMessageBox(_T("Missing closing semicolon."), _T("Failed Newick String."), wxOK | wxICON_ERROR);
		return false;
	}
	
	assert(lastP > firstP);

	string content = description.substr(firstP + 1, lastP - firstP - 1);
  string root_elements = description.substr(lastP + 1, semi - lastP - 1);

	// Quickly determine number of elements in the file.
	foreach(char ch, content)
	{
		if(ch == ',' || ch == ')')
			m_numElement++;
	}
	m_numElement++;

	// Track number of elements that have been processed
	m_processedElement = 0;

  // Create root node:
	NodePhylo* root(new NodePhylo(m_processedElement));
	tree->SetRootNode(root);
	root->SetDistanceToParent(Node::NO_DISTANCE);

	// Parse rest of tree
	NewickTokenizer nt(content);

	// The check for a parenthese is necessary as it is allowable to have
	// a tree rooted on a leaf in which case it should not be treated as a leaf.
	if(nt.NumTokens() == 1 && content.rfind("(") == string::npos)
  {    
		std::string name = util::StringTools::RemoveSurroundingWhiteSpaces(nt.Token(0));
		root->SetName(util::StringTools::ToStringW(name));
  }
  else
  {
		// This is a node.
		Element elements = Elements(root_elements);
		root->SetName(_T(""));
		if(!util::StringTools::IsEmpty(elements.name))
		{
			root->SetName(util::StringTools::ToStringW(elements.name));
		}

    for(uint i = 0; i < nt.NumTokens(); i++)
    {
      if(!ParseNode(tree, nt.Token(i), root))
				return false;
    }
	}

	return true;
}
