//=======================================================================
// Author: Donovan Parks
//
// Copyright 2009 Donovan Parks
//
// This file is part of GenGIS.
//
// GenGIS is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// GenGIS is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with GenGIS.  If not, see <http://www.gnu.org/licenses/>.
//=======================================================================

#include "../core/Precompiled.hpp"

#include "../widgets/MiniMapView2D.hpp"

#include "../core/App.hpp"
#include "../core/Camera.hpp"
#include "../core/LayerTree.hpp"
#include "../core/LocationLayer.hpp"
#include "../core/LocationSetLayer.hpp"
#include "../core/MapLayer.hpp"
#include "../core/StudyLayer.hpp"
#include "../core/TerrainController.hpp"
#include "../core/TerrainTexture.hpp"
#include "../core/TerrainView.hpp"
#include "../core/OrthoLayers.hpp"

#include "../core/Viewport.hpp"

#include "../utils/Geometry.hpp"

using namespace GenGIS;

MiniMapView2D::MiniMapView2D() : ViewWidget() 
{
	m_origin.x = 0;
	m_origin.y = 0;
}

GeoCoord MiniMapView2D::MiniMapCoordToGeoCoord( Point3D& miniMapCoord ) 
{
	// This code has not been verified. Use with caution.

  GeoCoord coord;
  
	FileHeader* header = m_terrainController->GetHeader();

  coord.easting = ( -( ( miniMapCoord.x + 1 ) * ( header->projExtents.Width() ) /
     2 + header->projExtents.x ));
  
  coord.northing = (-( ( miniMapCoord.y + 1 ) * ( header->projExtents.Height()) /
      2 + header->projExtents.y ) );
  
  return coord;
}

void MiniMapView2D::MiniMapCoordToWorldCoord( Point3D& miniMapCoord, Point3D& worldCoord ) 
{
  FileHeader* header = m_terrainController->GetHeader();

	float spanX = (float)(header->nCols * header->cellSizeX);
	float spanY = (float)(header->nRows * header->cellSizeZ);
  
  worldCoord.x = ( miniMapCoord.x + 1 ) * ( header->nCols * header->cellSizeX ) / 2 - spanX*0.5f;
  worldCoord.y = ( miniMapCoord.y + 1 ) * ( header->nRows * header->cellSizeZ ) / 2 - spanY*0.5f;
  worldCoord.z = 0;
}

Point3D MiniMapView2D::GeoCoordToMiniMapCoord( GeoCoord& geoCoord ) 
{
  Point3D coord;

  FileHeader* header = m_terrainController->GetHeader();
  
  coord.x = 2*(geoCoord.easting - header->projExtents.x) /
							(header->projExtents.Width()) - 1;
  
  coord.y = 2*(geoCoord.northing - header->projExtents.y) /
							(header->projExtents.Height()) - 1;
  
  coord.z = 0;
  
  return coord;
}


Point3D MiniMapView2D::WorldCoordToMiniMapCoord( Point3D& worldCoord ) 
{
	// Note: this is almost certainly incorrect. Needs to take into account span of map. 
	// See miniMapCoordToWorldCoord.

  FileHeader* header = m_terrainController->GetHeader();
  Point3D mmCoord;
  
  mmCoord.x = -2 * worldCoord.x / ( header->nCols * header->cellSizeX ) + 1;
  mmCoord.y = 2 * worldCoord.z / ( header->nRows * header->cellSizeZ ) - 1;
  
  return mmCoord;
}

void MiniMapView2D::DrawTerrain()
{
	if(m_terrainController && m_terrainController->GetTerrainView()->GetTerrainModel()->GetTexture())
	{
		glPushMatrix();
			glTranslated(0.0, 0.0, -WIDGETS_LAYER );
			glScaled(1,1,1);

			glDisable( GL_LIGHTING );

			glColor4f(1, 1, 1, 1);
			glBindTexture(GL_TEXTURE_2D, m_terrainController->GetTerrainView()->GetTerrainModel()->GetTexture()->GetTextureId());
			glEnable(GL_TEXTURE_2D);

			glBegin(GL_POLYGON);
				glTexCoord2f(0, 0);
				glVertex2f(+1, -1);
			  
				glTexCoord2f(0, 1);
				glVertex2f(+1, +1);
			  
				glTexCoord2f(1, 1);   
				glVertex2f(-1, +1);
			  
				glTexCoord2f(1, 0);    
				glVertex2f(-1, -1);
			glEnd();

			glDisable(GL_TEXTURE_2D);

			glEnable( GL_LIGHTING );
		glPopMatrix();
	}
}

void MiniMapView2D::Render() 
{
	const float LOCATION_MARKER_SIZE = 3.0;

	if(m_bVisible && !App::Inst().IsSavingHiResImage())
	{
		glPushMatrix();  

			glTranslated( m_screenCoordinates.x, m_screenCoordinates.y, -(WIDGETS_LAYER+1) );
			glScaled( m_scale.x, m_scale.y, m_scale.z );  
			glRotated( 180, 0, 0, 1 );       

			DrawTerrain();

			// draw location markers
			if(m_layerTree)
			{
				for(unsigned int studyIndex = 0; studyIndex < m_layerTree->GetNumStudies(); studyIndex++)
				{
					StudyLayerPtr study = m_layerTree->GetStudy(studyIndex);
					if(study->IsVisible())
					{
						for(unsigned int mapIndex = 0; mapIndex < study->GetNumMaps(); ++mapIndex)
						{
							MapLayerPtr map = study->GetMap(mapIndex);
							if(map->IsVisible())
							{
								for(unsigned int locationSetIndex = 0; locationSetIndex < map->GetNumLocationSets(); ++locationSetIndex)
								{
									LocationSetLayerPtr locationSet = map->GetLocationSet(locationSetIndex);
									if(locationSet->IsVisible())
									{
										for(unsigned int locationIndex = 0; locationIndex < locationSet->GetNumLocations(); ++locationIndex)
										{
											LocationLayerPtr location = locationSet->GetLocation(locationIndex);
											if(location->IsVisible())
											{
												GeoCoord temp = location->GetLocationController()->GetProjection();
												Point3D mapCoord = GeoCoordToMiniMapCoord(temp);

												if(fabs(mapCoord.x) <= 1.0 && fabs(mapCoord.y) <= 1.0)
												{
													glDisable( GL_LIGHTING );
													glColor3f(1.0f, 0.0f, 0.0f);
													glPointSize(LOCATION_MARKER_SIZE);
													glBegin(GL_POINTS);	
														glVertex3f(-mapCoord.x, -mapCoord.y, 1.0);
													glEnd();
													glEnable( GL_LIGHTING );  
												}
											}
										}
									}
								}
							}
						}
					}
				}
			}

		glPopMatrix();
	}
}

bool MiniMapView2D::IsClicked(const Point2D& mousePt) 
{
  if (!IsVisible())
      return false;
  
  Box2D box = GetBoundingBox();
  
  return (mousePt.x >= box.x && mousePt.x <= box.x + box.dx && 
          mousePt.y >= box.y && mousePt.y <= box.y + box.dy );
}

bool MiniMapView2D::MouseLeftDown(const Point2D& mousePt)
{
	if(!IsClicked(mousePt))
		return false;

	// convert screen cords to minimap grid coordinates
	Box2D box = GetBoundingBox();

	float u = mousePt.x;
	float v = mousePt.y;

	u -= box.x + box.dx / 2;
	v -= box.y + box.dy / 2;

	u = 2 * u / box.dx;
	v = 2 * v / box.dy;

	Point3D gridCoord;
	gridCoord.x = u;
	gridCoord.y = v;
  
	Point3D newWorld;
	MiniMapCoordToWorldCoord( gridCoord, newWorld );

	Point3D rotatedWorldCoord;
	float theta = App::Inst().GetViewport()->GetCamera()->GetYaw();
	rotatedWorldCoord.x = sin(PI_BY_2 - theta * DEG_TO_RAD ) * newWorld.x - sin( theta * DEG_TO_RAD ) * newWorld.y;
	rotatedWorldCoord.y = sin( theta * DEG_TO_RAD ) * newWorld.x + sin(PI_BY_2 - theta * DEG_TO_RAD ) * newWorld.y;

	float height = App::Inst().GetViewport()->GetCamera()->GetHeight();
	App::Inst().GetViewport()->GetCamera()->SetPosition(rotatedWorldCoord.x, rotatedWorldCoord.y, height);

	return true;
}
